#ifndef lint
#endif
#define YYBYACC 1
#define YYMAJOR 1
#define YYMINOR 9
#define yyclearin (yychar=(-1))
#define yyerrok (yyerrflag=0)
#define YYRECOVERING (yyerrflag!=0)
#define YYPREFIX "yy"
#line 3 "gram.y"

/* Copyright (c) 1995, 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003, 2004 by Arkkra Enterprises */
/* All rights reserved */

/* parser for music publication program. Uses lex to get tokens. */


#include "rational.h"
#include "globals.h"
#include "defines.h"
#include "structs.h"

#ifdef DEBUG
#define YYDEBUG 1
#endif


/* how many items to allocate at a time for
 * "with" list, and location lists, etc */
#define ITEMS	(4)


static struct MAINLL *Currstruct_p;	/* main list struct being filled in */
static struct BLOCKHEAD *Currblock_p;	/* current block, if non-zero */

static struct INPCOORD *Curr_loc_info_p; /* location info being collected */


static short Gotheader, Gotfooter;	/* boolean: to make sure each is defined only once */
static short Gothead2, Gotfoot2;
static short Getting_tup_dur;		/* YES if currently collection tuplet
					 * duration rather than GRPSYL dur */

static struct GRPSYL *Curr_grpsyl_p;	/* current GRPSYL struct being filled in */
static struct GRPSYL *Last_grpsyl_p;	/* previous GRPSYL struct filled in (contains
					 * defaults for the current one */
static struct GRPSYL *Lyrics_p;		/* pointer to current list of lyrics */
static char **Curr_marklist;		/* current "with" list */
static struct PRINTDATA **Next_print_link_p_p;	/* points to where to attach next
						 * print command that we get */
static int Item_count;			/* items in current list */
static int Max_items;			/* current amount of space in lists
					 * "with" list, curve locations, etc) */

static int Plus_minus;			/* if 1 add, if -1 subtract offset */
static int Defining_multiple = NO;	/* if defining multiple voices at once */
static int Chord_at_a_time = NO;	/* YES for chord-at-a-time,
					 * NO for voice-at-a-time */
static short Endingloc;			/* if current bar starts or ends an
					 * ending, etc (STARTITEM, ENDITEM, etc) */
static short Got_ending;		/* YES if got ending on this bar */

static int Til_bars;			/* number of bars in "til" clause */
static int Dist;			/* dist for a specific STUFF */
static int Dist_usage;			/* meaning of Dist field */
static short User_meas_time;		/* YES if user specified a time for
					 * an 'm' item. Only valid on mr */
static int Order_place;			/* PL_* for markorder in SSV */
static int Order_prio;			/* value in markorder in SSV */
static int Gracebackup;			/* how many grace notes to back up
					 * before starting a "stuff" */
static float Til_offset;		/* beat count in "til" clause */
static int Slurstyle;			/* L_DOTTED, L_DASHED, or L_NORMAL */

static char *String1, *String2;		/* to temporarily keep track
					 * of labels and other temporary
					 * strings */
static char Tmpbuff[20];		/* temporary buffer for converting
					 * float numbers */
static char Stringbuff[100];		/* temporary string storage */
static int Titlefont, Titlesize;	/* font/size for title command */
static double Extra;			/* extra vertical space for prints */
static short Good_till_canceled = NO;	/* if [] things should be continued
					 * until turned off */
static short Curr_paratype = J_JUSTPARA; /* current paragraph type */
static struct TIMEDSSV *Curr_tssv_p;	/* current mid-measure param change */



/* static functions */
static char *stripquotes P((char *string));	/* rmv double quotes
						 * from string */
static void chkdup_headfoot P((int flag, char *which));
static void end_prev_context P((void));
static void var_valid P((void));	 /* check if okay to set location var */
static void proc_printcmd P((int justifytype, struct INPCOORD *inpc_p,
		char *str, int font, int size, int got_nl,
		int isPostScript, double extra));
static double extra_needed P((int font, int size, char *string));
static void keyword_notes P((char *str));


#line 99 "gram.y"
typedef union
{
int	intval;		/* for tokens/nonterminals that return int values */
char	*stringval;	/* for tokens/nonterminals that return strings */
float	floatval;	/* for nonterminals that return floats */
RATIONAL ratval;	/* for nonterminals that return rational numbers */
struct INPCOORD *inpcoord_p;	/* for nonterminals that return coord info */
} YYSTYPE;
#line 114 "y.tab.c"
#define T_SCORE 257
#define T_STAFF 258
#define T_VOICE 259
#define T_GRIDS 260
#define T_BLOCKHEAD 261
#define T_NUMBER 262
#define T_EQUAL 263
#define T_SEMICOLON 264
#define T_NEWLINE 265
#define T_COLON 266
#define T_COMMA 267
#define T_LBRACKET 268
#define T_RBRACKET 269
#define T_LBRACE 270
#define T_RBRACE 271
#define T_VARIABLE 272
#define T_DASH 273
#define T_PLUS 274
#define T_PLACE 275
#define T_ALL 276
#define T_WITH 277
#define T_CUE 278
#define T_GRACE 279
#define T_XNOTE 280
#define T_DIAM 281
#define T_DOT 282
#define T_MULTIWHOLE 283
#define T_LET_M 284
#define T_LET_R 285
#define T_LET_S 286
#define T_LET_U 287
#define T_PITCH 288
#define T_SHARP 289
#define T_AMPERSAND 290
#define T_LET_N 291
#define T_LET_X 292
#define T_DBLFLAT 293
#define T_LETTER 294
#define T_LYRICS 295
#define T_PAD 296
#define T_LPAREN 297
#define T_RPAREN 298
#define T_LVAR 299
#define T_MUSIC 300
#define T_HEADSHAPES 301
#define T_HS 302
#define T_SWINGUNIT 303
#define T_POSTSCRIPT 304
#define T_SCOREFEED 305
#define T_BARTYPE 306
#define T_HAT 307
#define T_TILDE 308
#define T_QUESTION 309
#define T_LINETYPE 310
#define T_LINE 311
#define T_TO 312
#define T_STAR 313
#define T_BULGE 314
#define T_CURVE 315
#define T_TIE 316
#define T_PRINTTYPE 317
#define T_NL 318
#define T_STAFFLINES 319
#define T_FFAMILY 320
#define T_PARAGRAPH 321
#define T_PARATYPE 322
#define T_NUMVAR 323
#define T_FNUMVAR 324
#define T_2FNUMVAR 325
#define T_RATNUMLISTVAR 326
#define T_RANGELISTVAR 327
#define T_TIMEUNIT 328
#define T_FONTVAR 329
#define T_STRVAR 330
#define T_CLEFVAR 331
#define T_VVAR 332
#define T_BARSTLISTVAR 333
#define T_SCORESEP 334
#define T_FONT 335
#define T_SLASH 336
#define T_KEY 337
#define T_TIME 338
#define T_CUT 339
#define T_COMMON 340
#define T_CLEF 341
#define T_MULTIREST 342
#define T_YESNOVAR 343
#define T_BM 344
#define T_ESBM 345
#define T_EBM 346
#define T_INVISBAR 347
#define T_OCTAVE 348
#define T_ROLL 349
#define T_VISVAR 350
#define T_WHEREUSED 351
#define T_UNSET 352
#define T_WORD 353
#define T_UP 354
#define T_DOWN 355
#define T_ENDING 356
#define T_ENDENDING 357
#define T_FAMILY 358
#define T_ENDSTYLE 359
#define T_ENDTYPE 360
#define T_SLASHMARK 361
#define T_ALT 362
#define T_TITLE 363
#define T_REHEARSAL 364
#define T_RNUM 365
#define T_LET 366
#define T_MNUM 367
#define T_TRANSPOSE 368
#define T_XPOS_INT 369
#define T_PEDAL 370
#define T_PHRASE 371
#define T_MUSSYM 372
#define T_L_ANGLE 373
#define T_MODIFIER 374
#define T_R_ANGLE 375
#define T_TIL 376
#define T_MIDI 377
#define T_PEDSTYLE 378
#define T_PEDSTAR 379
#define T_HIDECHANGES 380
#define T_SLUR 381
#define T_UNITS 382
#define T_CENTIMETERS 383
#define T_UNITTYPE 384
#define T_TAB 385
#define T_TICKS 386
#define T_NOWHERE_SLIDE 387
#define T_LEN 388
#define T_EXCLAM 389
#define T_OTHERTEXT 390
#define T_ORDER 391
#define T_GRIDPLACE 392
#define T_GRPLVALUE 393
#define T_DRUM 394
#define T_RPT 395
#define T_STRING 396
#define T_HO 397
#define T_REH_STYLE 398
#define T_REHTYPE 399
#define T_L_DBLANGLE 400
#define T_R_DBLANGLE 401
#define YYERRCODE 256
short yylhs[] = {                                        -1,
    0,    0,   82,   82,   82,   82,   82,   82,   82,   82,
   82,   82,   83,   83,   83,   83,   83,   92,   92,   92,
   98,   97,   93,   96,   94,   89,   95,    3,    4,    1,
   84,  100,  100,   99,   99,   99,   99,   99,   99,   99,
   99,   99,   99,   99,   99,   99,   99,   99,   99,   99,
   99,   99,   99,   99,   99,   99,   99,   99,   99,   99,
   99,  112,  112,  112,    2,    2,    2,  101,  102,  102,
  115,  115,  115,  115,  117,  116,   76,   76,   79,   79,
   46,   46,  103,  104,  104,  104,  119,    8,    8,  120,
  120,  121,  121,   61,  105,  106,  106,  106,  122,  107,
  108,  108,  108,   45,   44,   44,   44,   44,   44,   44,
   44,   44,  110,  110,  110,  110,    5,    5,  109,  109,
  111,  111,  111,  111,   28,   28,  123,   29,   29,   27,
   27,    6,    6,  113,  114,  114,   57,   57,   40,   40,
   53,   53,   37,   37,   37,   41,   85,  124,   39,   39,
   38,   38,  125,  125,  126,  126,  126,   86,   58,   58,
   56,   56,   56,   56,   56,   56,   56,   56,   56,   56,
   56,   56,   56,   56,   56,   56,   56,   56,   56,   56,
   56,   56,   56,   56,   56,   56,   60,   60,   87,   87,
   87,  127,  128,  128,  133,  133,  134,  135,  137,  136,
  136,  132,  132,  139,  138,  138,  129,  140,  140,  141,
    9,    9,   47,   47,  142,  142,  130,  130,  143,  143,
  145,  145,  149,  149,  149,  150,  150,  151,  151,  146,
  152,  152,  152,  156,  156,  157,  157,  154,  155,  155,
  158,  158,  159,  159,  159,  159,  159,  159,  159,  159,
  159,  159,   55,   55,   55,   54,   54,  148,  148,  165,
  165,  166,  166,  166,  166,  160,   21,   21,  161,  161,
  168,   62,   62,   67,   67,   67,   67,  163,  163,  163,
  163,   20,   20,  167,  167,  167,  153,  153,   24,   24,
   77,   77,  118,  118,    7,    7,  164,   74,   12,   12,
  147,  147,  169,  170,  172,  173,  173,  173,  173,  173,
  173,  173,  171,  171,   35,   35,  175,  175,  175,  175,
  175,  175,  175,  175,  175,  175,  175,  175,  176,  176,
  176,  176,  176,  176,  176,  176,  176,  176,  176,  176,
   30,   30,   13,   18,   18,   19,   19,   23,   23,   23,
   23,   23,   14,   14,   14,   15,   15,   16,   16,  174,
  177,  178,  178,  178,  179,  179,  180,  162,   65,   65,
   65,   65,   65,   65,   65,   65,   65,  144,   25,   25,
   25,  181,  182,   22,   22,   78,   78,  131,  183,  185,
   51,   51,  186,  186,  186,  187,   10,   10,   10,  188,
  188,  189,  189,  190,   11,   11,  191,  184,  192,  192,
  194,  194,  195,  196,  196,  198,  197,  197,  193,  193,
  199,  199,   90,   90,   90,  205,   90,   90,  204,  204,
  207,  208,  208,   91,   91,   88,   88,  200,   59,   59,
  201,  201,   33,   33,   33,   33,   33,   33,   33,   33,
   68,   68,  211,   69,   69,   69,   69,   80,   81,  213,
  215,  215,  215,   31,   31,   31,  216,  216,  218,   17,
   17,  217,  217,  214,  219,  219,   32,   32,   32,  220,
  220,  221,   73,   72,   72,   72,   63,   63,   64,   66,
  202,  222,   36,   36,   71,   71,  203,  223,  225,  225,
  226,  226,  227,  227,  228,  224,  209,  229,  229,  229,
  232,  233,  233,  231,   48,   48,   48,   48,   48,   48,
   49,   49,   50,   50,  212,  212,   43,   43,  234,  234,
  230,  230,  235,   75,  237,  237,   26,   26,   70,   70,
   70,  236,  236,  238,   52,   52,  239,  239,  210,  240,
  240,  241,  243,  243,  244,  244,  244,  242,  242,  206,
  206,  245,   34,   42,   42,   42,
};
short yylen[] = {                                         2,
    1,    2,    1,    1,    1,    1,    2,    2,    1,    3,
    1,    2,    1,    1,    1,    1,    2,    2,    3,    4,
    1,    0,    1,    1,    1,    3,    1,    1,    1,    1,
    2,    1,    1,    4,    2,    3,    5,    4,    4,    5,
    4,    4,    3,    3,    3,    3,    3,    3,    3,    3,
    3,    2,    3,    3,    4,    3,    3,    3,    5,    3,
    3,    1,    1,    1,    0,    1,    1,    0,    0,    1,
    1,    1,    3,    3,    0,    4,    3,    1,    0,    1,
    0,    1,    0,    0,    1,    3,    3,    0,    2,    0,
    4,    0,    2,    1,    0,    0,    1,    3,    2,    1,
    1,    3,    3,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    2,    4,    1,    1,    0,    1,    3,    3,
    2,    2,    1,    4,    1,    2,    4,    0,    1,    0,
    1,    1,    1,    1,    0,    2,    1,    1,    1,    1,
    0,    1,    1,    2,    1,    1,    6,    2,    0,    1,
    0,    1,    0,    3,    1,    2,    3,    7,    0,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    3,    4,    3,
    1,    0,    1,    1,    1,    2,    3,    1,    1,    1,
    3,    1,    3,    1,    1,    2,    1,    1,    3,    2,
    0,    2,    2,    4,    0,    2,    1,    2,    1,    5,
    0,    4,    1,    1,    1,    1,    3,    3,    3,    2,
    1,    4,    6,    0,    3,    0,    4,    0,    0,    1,
    1,    3,    3,    1,    1,    1,    1,    3,    2,    2,
    2,    2,    1,    1,    2,    0,    1,    0,    1,    1,
    3,    1,    2,    2,    2,    0,    0,    1,    1,    3,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    2,    1,    1,    0,    1,    0,    3,
    1,    1,    0,    1,    0,    2,    1,    3,    0,    1,
    1,    1,    2,    2,    0,    1,    2,    2,    3,    2,
    2,    1,    0,    2,    0,    3,    1,    1,    1,    2,
    1,    1,    1,    1,    1,    1,    1,    3,    2,    5,
    5,    3,    5,    7,    4,    4,    6,    4,    5,    4,
    0,    1,    1,    0,    1,    0,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    2,    1,    2,    4,
    1,    0,    1,    1,    1,    2,    2,    2,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    8,    0,    1,
    1,    1,    1,    0,    1,    0,    2,    4,    2,    1,
    0,    1,    1,    1,    3,    2,    0,    2,    2,    1,
    1,    1,    3,    2,    0,    2,    0,    2,    0,    1,
    1,    2,    2,    3,    1,    5,    0,    1,    1,    2,
    2,    5,    2,    1,    1,    0,    3,    1,    0,    2,
    3,    0,    1,    0,    1,    1,    1,    3,    0,    3,
    0,    2,    1,    2,    1,    1,    3,    3,    3,    7,
    2,    1,    0,    1,    1,    1,    1,    6,    0,    2,
    1,    3,    1,    1,    1,    1,    0,    2,    2,    1,
    1,    1,    2,    2,    1,    3,    1,    1,    1,    0,
    2,    2,    1,    1,    3,    2,    0,    1,    1,    1,
    7,    0,    0,    1,    0,    4,    7,    0,    2,    2,
    1,    3,    1,    3,    2,    1,    3,    6,    6,    2,
    1,    1,    1,    1,    2,    1,    1,    1,    1,    1,
    0,    3,    0,    1,    0,    2,    2,    3,    0,    1,
    1,    2,    5,    2,    0,    4,    0,    4,    0,    1,
    1,    0,    2,    2,    0,    2,    0,    2,    5,    0,
    1,    2,    1,    2,    0,    2,    3,    2,    3,    3,
    3,    1,    1,    0,    1,    1,
};
short yydefred[] = {                                      0,
    0,   22,    0,    0,   25,   23,   11,    0,    0,   24,
   27,    0,  562,  426,  494,  563,    0,  152,  160,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,  520,    0,    0,    0,    0,
    0,    0,    0,  518,  519,  516,  517,  511,    0,    0,
  100,  187,    0,    0,    0,    0,    0,  514,    0,    0,
    1,    3,    4,    5,    6,    0,    0,    9,    0,   13,
   14,   15,   16,   22,    0,    0,    0,  191,  441,  424,
  425,  428,  436,  437,    0,    0,    0,    0,   12,   21,
   18,   30,   28,   22,    0,  392,  390,    0,    0,    0,
    0,  429,    0,    0,    0,    0,   68,   83,    0,    0,
    0,    0,    0,   95,    0,    0,    0,    0,    0,  551,
    0,    0,  167,  166,  164,  161,  169,  170,  181,  176,
  165,  184,  183,  186,  175,  168,  163,  178,  179,  162,
  182,  185,  172,  180,  174,  171,  177,  173,   52,    0,
    0,    0,    0,    0,    0,    0,    0,    2,  459,  565,
    0,  566,    0,  492,  498,    0,    0,  515,    0,    0,
    0,    7,    8,  435,    0,   17,   32,   33,   31,    0,
  198,    0,    0,    0,    0,  195,    0,  202,    0,  208,
    0,    0,    0,    0,  513,    0,  510,    0,   19,   29,
   22,    0,  394,    0,    0,  393,  300,    0,  291,   78,
  292,   80,  295,   54,    0,    0,    0,   46,    0,    0,
    0,   36,    0,    0,    0,   53,   47,    0,   49,    0,
    0,  134,    0,  343,    0,    0,   44,    0,  115,  116,
    0,   45,  190,  133,  132,   50,    0,    0,    0,   64,
   63,   62,   51,   48,  138,  137,   56,  150,  148,    0,
    0,  139,  140,    0,  143,    0,  145,   58,   60,   57,
    0,    0,    0,    0,    0,    0,  438,    0,  188,   26,
   10,  109,  106,  110,  111,  112,  105,  108,  107,  104,
  101,    0,    0,  210,  207,    0,  204,    0,  196,    0,
    0,    0,    0,  206,    0,    0,    0,  452,  453,    0,
    0,    0,  446,  442,  445,  443,    0,    0,    0,  531,
    0,  524,    0,    0,   20,  382,    0,  388,    0,    0,
    0,  411,    0,  415,    0,    0,  396,    0,   38,    0,
    0,  430,    0,  118,  122,  121,   34,  489,  486,    0,
    0,   75,   71,    0,    0,   72,    0,    0,   85,   66,
   67,   39,    0,    0,   97,    0,   55,  348,  349,  350,
  351,  352,    0,    0,  345,  113,    0,  554,    0,    0,
  552,    0,    0,    0,  146,    0,  144,  471,  470,  371,
  372,  373,  374,  370,  375,  376,  369,  377,    0,    0,
  490,    0,  472,    0,  467,  461,  463,  440,    0,  506,
    0,    0,    0,    0,    0,  212,    0,    0,  217,  219,
    0,    0,  203,    0,  199,  197,  209,    0,    0,  213,
  444,    0,    0,    0,    0,    0,    0,    0,  534,    0,
    0,  532,  522,    0,    0,  288,    0,    0,    0,    0,
  419,  412,  413,  398,  399,  395,  296,  294,   77,  433,
    0,    0,    0,  125,  485,  488,   37,    0,   82,   40,
    0,    0,    0,    0,   99,    0,  136,  142,  119,  120,
    0,    0,    0,    0,  154,    0,    0,   59,  473,  469,
    0,    0,    0,    0,    0,    0,    0,  102,  103,  223,
  224,  225,    0,  218,    0,    0,    0,  231,    0,    0,
  216,    0,  298,    0,  449,    0,  447,    0,    0,  540,
    0,    0,    0,  383,    0,  418,  414,  401,    0,    0,
    0,  402,  421,  420,  431,  129,    0,  124,  126,    0,
   73,   74,   89,    0,   87,   86,   98,  114,  558,    0,
  557,    0,  147,  466,  465,  464,  462,    0,  483,  475,
    0,  480,  468,    0,    0,    0,  509,    0,    0,    0,
    0,  226,    0,    0,    0,    0,  302,  313,    0,  230,
    0,  214,    0,    0,    0,    0,    0,  508,    0,    0,
  404,    0,    0,  131,    0,   76,    0,    0,  559,    0,
  458,    0,    0,  491,    0,    0,  497,    0,    0,  526,
  158,    0,    0,    0,  222,  238,    0,  266,  278,  279,
  281,  280,    0,    0,  282,  283,    0,    0,    0,  247,
  297,    0,  240,  241,  244,  245,  246,  323,  318,  319,
    0,    0,  322,  325,  326,  327,  324,  321,    0,  313,
  317,    0,    0,  285,  286,    0,    0,    0,    0,  260,
  262,    0,    0,    0,    0,    0,  306,  361,  304,  312,
    0,  385,    0,    0,  536,  538,    0,  543,  533,    0,
  416,  406,    0,  403,  356,  358,  353,  347,    0,    0,
  127,    0,    0,  478,  477,  479,  476,    0,  481,    0,
  501,    0,    0,    0,  503,    0,  530,  527,  229,  228,
  227,  232,  368,    0,    0,    0,  252,  251,    0,    0,
  257,    0,  249,    0,    0,  320,    0,    0,    0,    0,
    0,  329,    0,    0,  268,  265,    0,  284,  263,  264,
  220,    0,  314,  308,    0,  310,    0,  307,  363,    0,
    0,    0,  365,    0,  456,  457,  455,    0,    0,    0,
  544,  387,  422,  357,  359,   93,   91,  482,    0,    0,
  505,    0,  528,  275,  274,  277,  276,    0,  271,  272,
    0,  269,  248,  255,  242,    0,  236,  328,  332,    0,
    0,    0,    0,    0,    0,    0,    0,    0,  261,  309,
  367,    0,  366,  380,  381,    0,  450,    0,  546,    0,
  502,  504,    0,    0,    0,    0,    0,    0,  336,  340,
    0,  338,  335,    0,    0,  290,  360,    0,  548,  270,
  235,    0,  331,  330,    0,  339,    0,  333,  378,    0,
  337,    0,  237,  334,
};
short yydgoto[] = {                                      54,
  221,  362,  182,  303,  346,  246,  340,  473,  294,  337,
  591,  703,  462,  688,  689,  690,  400,  374,  691,  735,
  736,  673,  375,  738,  806,  441,  595,  463,  537,  732,
  557,  697,  314,   55,  575,   56,  268,  152,  259,  264,
  386,  161,  610,  290,  291,  470,  304,   58,  194,  323,
   97,  761,  479,  722,  723,  149,  257,   59,  167,  733,
  598,  779,  465,  349,  401,  402,  780,  315,  759,  522,
  604,  317,  560,  631,  318,  353,  213,  681,  214,  410,
  271,   61,   62,   63,   64,   65,   66,   67,   68,   69,
  175,   70,   71,   72,   73,   74,   90,   91,   75,  179,
  224,  354,  225,  358,  231,  364,   76,  292,  237,  242,
  218,  253,  233,  367,  355,  356,  468,  459,  359,  545,
  693,  365,  464,  153,  261,  384,   77,  183,  296,  418,
   78,  184,  185,  186,  187,  301,  426,  188,  298,  189,
  190,  430,  419,  420,  421,  506,  576,  658,  503,  571,
  572,  507,  447,  508,  632,  787,  815,  633,  634,  714,
  781,  635,  636,  637,  659,  660,  661,  782,  577,  578,
  662,  579,  743,  670,  650,  651,  671,  751,  752,  753,
  327,  525,   98,  328,   99,  205,  206,  530,  531,  532,
  329,  330,  450,  331,  332,  333,  527,  334,  451,   79,
  191,   80,   81,  215,  102,   82,  342,  461,   83,   84,
  434,  567,  404,  561,  405,  493,  406,  407,  562,  602,
  699,  274,  275,  411,  607,  702,  704,  705,   85,  319,
   86,   87,  197,  708,  320,  587,  439,  678,  809,  121,
  248,  483,  249,  381,   88,
};
short yysindex[] = {                                    438,
 -202,    0, -153, -153,    0,    0,    0, -133,  -69,    0,
    0,  -62,    0,    0,    0,    0,   -4,    0,    0,    5,
   37,   73,  157,  173,  221,  267,  275,  283,  299,  319,
  328,  362,  416, -153,  425,    0,  288,  429, 2073,  437,
  490,  436,  496,    0,    0,    0,    0,    0,  521,  536,
    0,    0,  541,  438,   19,  -41,  472,    0,  470, -246,
    0,    0,    0,    0,    0,  544,  547,    0,  555,    0,
    0,    0,    0,    0,  117,  559,  315,    0,    0,    0,
    0,    0,    0,    0,  569,  540,   44,   19,    0,    0,
    0,    0,    0,    0, -153,    0,    0,  582,   52,  586,
   88,    0, -216,  586,   27,   27,    0,    0,   88,  497,
  459,  534, -153,    0, -153,    4,  -28,  555,  322,    0,
 -153, -162,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  571,
 -220,  561,  616,  107, -100,  548,  529,    0,    0,    0,
  459,    0,  628,    0,    0,  540,  630,    0,  436,  519,
  259,    0,    0,    0,  672,    0,    0,    0,    0,  311,
    0,  665,  682,  663,  686,    0, -153,    0,  399,    0,
  272,   27, -153,  581,    0,  663,    0,  459,    0,    0,
    0,  687,    0,  182,  691,    0,    0,   27,    0,    0,
    0,    0,    0,    0,  635,  667, -218,    0, -153,  703,
  684,    0,  700,   95, -153,    0,    0,  694,    0,  333,
 -153,    0,  702,    0,  535,  535,    0,  680,    0,    0,
  636,    0,    0,    0,    0,    0, -153,  707,  662,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, -153,
  459,    0,    0,  606,    0,  599,    0,    0,    0,    0,
 2162,  694,  679,  683,  683, -133,    0,  616,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, -215, -153,    0,    0, -234,    0, -153,    0,  663,
  247, -153,  359,    0,  598,  586,  459,    0,    0,  716,
  719,  720,    0,    0,    0,    0,  688,  722,   27,    0,
  693,    0, -133,  694,    0,    0,  687,    0,   88, -233,
  687,    0,  729,    0, -153, -153,    0, -153,    0,   -7,
  721,    0,  706,    0,    0,    0,    0,    0,    0,  703,
   27,    0,    0,  711,  730,    0,  728,  735,    0,    0,
    0,    0,  728,  737,    0, -153,    0,    0,    0,    0,
    0,    0,  644,  644,    0,    0, -153,    0,   27, -153,
    0,  708, -246,  555,    0, -153,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,   27,    3,
    0,  723,    0,  748,    0,    0,    0,    0,  705,    0,
  710,   52,  459,  311,  311,    0,  -33, -234,    0,    0,
  750, -234,    0, -153,    0,    0,    0, -153, -153,    0,
    0,   27,  694,  436, -153,  459, -153,  746,    0,  296,
 -183,    0,    0,   52,  364,    0,  734,   20,  228, -233,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
   27,  535,   78,    0,    0,    0,    0,   95,    0,    0,
   95, -153,  726, -153,    0, -153,    0,    0,    0,    0,
  680,  760,   27, -153,    0, -246,  763,    0,    0,    0,
  326,  424,  296,  683,  683,  -74,  249,    0,    0,    0,
    0,    0,  135,    0,  752,    0,   88,    0, -212,  663,
    0,  767,    0,  561,    0,  694,    0, -153,   27,    0,
  694,  660,  -74,    0, -153,    0,    0,    0,  764,  769,
  772,    0,    0,    0,    0,    0,  656,    0,    0, -213,
    0,    0,    0,  459,    0,    0,    0,    0,    0,  777,
    0,  694,    0,    0,    0,    0,    0,  761,    0,    0,
  747,    0,    0,  770,  366,  155,    0,  779,  785,  789,
 -237,    0,  784, -201, 2048,   69,    0,    0,  -66,    0,
  783,    0,  540,  756,  790,   27,  796,    0,  795, -153,
    0,  459, -153,    0,   51,    0,  694,  797,    0,  295,
    0,  296,  436,    0,  683,  586,    0, -153,  674,    0,
    0, -153,  725,  135,    0,    0,  598,    0,    0,    0,
    0,    0,  459,  586,    0,    0, -153,   27,  330,    0,
    0,  404,    0,    0,    0,    0,    0,    0,    0,    0,
  781,  706,    0,    0,    0,    0,    0,    0, -146,    0,
    0,  382,  794,    0,    0, -153,  382,  809,  807,    0,
    0,  -66,  598,  459,  -73,  382,    0,    0,    0,    0,
 -248,    0, -153, -136,    0,    0,  791,    0,    0,   88,
    0,    0,  268,    0,    0,    0,    0,    0,  810,  808,
    0,  459,  787,    0,    0,    0,    0,   27,    0,  616,
    0,  775,   27,  821,    0,  674,    0,    0,    0,    0,
    0,    0,    0, -192,  694, -153,    0,    0,    0,    0,
    0,   27,    0, -201,  811,    0,  798,  576,  220,  365,
 -177,    0,  694,  -66,    0,    0,  831,    0,    0,    0,
    0,   69,    0,    0,  694,    0,   51,    0,    0,   51,
  717,  706,    0,  378,    0,    0,    0,  694,  768,  820,
    0,    0,    0,    0,    0,    0,    0,    0,  459,  683,
    0,  586,    0,    0,    0,    0,    0,  694,    0,    0,
  836,    0,    0,    0,    0,  816,    0,    0,    0,  806,
  812,   51,  459,  439,  459,  706,  278,  832,    0,    0,
    0,  382,    0,    0,    0,  795,    0,   27,    0,  694,
    0,    0, -192,  824,  840,  459,   51,  813,    0,    0,
  814,    0,    0,   51,  459,    0,    0,  849,    0,    0,
    0, -201,    0,    0,  459,    0,  818,    0,    0,  407,
    0,  459,    0,    0,
};
short yyrindex[] = {                                    244,
    0,    0,    0,    0,    0,    0,    0,  202,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,  852,    0,    0,    0,
    0, -115,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,  244,  727,  815,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,  853,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, -193,    0,  727,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,   60,
  456,    0,    0,  380,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,  853,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, -198,  731,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,  233,    0,    0, -115,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  332,    0,  851,  856,    0,    0,    0,  101,    0,
  509,    0,    0,  269,    0,  863,    0,    0,    0,    0,
    0, -197,    0,  -29,  865,    0,    0,    0,    0,    0,
    0,    0,    0,    0,  513,  566,  589,    0,    0,    0,
 1208,    0,    0,    8,  487,    0,    0,  593,    0,  607,
  518,    0,  609,    0,    0, -157,    0,  589,    0,    0,
    0,    0,    0,    0,    0,    0, -224,    0,  866,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  613,    0,    0,    0,  202,    0,  731,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  615,    0,    0,    0,  793,    0,    0,    0,  418,
    0,    0,  307,    0,    0,   60,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, -211, -196,  868,    0,
    0,    0,  202,  623,    0,    0,  243,    0, -160,    0,
 -127,    0,    0,    0,    0,    0,    0,    0,    0, 1348,
  102,    0,    0,    0,    0,    0,    0,    0,    0, 1273,
    0,    0,    0,  629,  192,    0,   54,  637,    0,    0,
    0,    0,  528,  639,    0,    0,    0,    0,    0,    0,
    0,    0,  641,  641,    0,    0,    0,    0,    0,    0,
    0,    0,  643,  853,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,  724,    0,    0,
  931,  793,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,   38,   34,    0,    0,    0,    0,    0,    0,
 -221,    0,    0,    0,  243,    0,  864,    0,    0,  870,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  -71,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,  538,    0,    0,    0,    0,    0,    0,    0,
  589,    0,  871,  874,    0,  645,    0,    0,    0,    0,
    0,    0,  875,    0,    0,  877,  853,    0,    0,    0,
    0,    0,    0,    0, -230, 1624, 1693,    0,  793,  450,
    0,  307,    0,   -6,    0,  123,    0,    0,    0,    0,
 -203,  880,  877,    0,    0,    0,    0,    0,  531,    0,
  882,    0,    0,    0,    0,    0,  165,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  647,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,  655,    0,    0,    0,    0,    0,    0,
    0,    0,    0,  453, -167,  888,    0,    0,    0,    0,
  892,    0,   41,    0,    0,    0,    0,    0,  893,    0,
    0,    0,    0,    0,  290,    0, -164,  858,    0,    0,
    0,  860, -115,    0,    0,   60,    0,    0,  -12,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,  898,    0,    0,    0,    0,  162,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1762,    0,
    0,  426,  440,    0,    0,    0,  426,    0,  897,    0,
    0,  -58,    0,    0,    0, 1831,    0,    0,    0,    0,
  788,    0,    0,    0,    0,    0,  901,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,  554,  654,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  731,
    0,  657,    0,  659,    0,  -12,    0,    0,    0,    0,
    0,    0,    0,    0,  471,    0,    0,    0,  185,  348,
    0,    0,    0,    0,  862,    0,    0,    0, 1069, 1139,
  212,    0, 1417, 2033,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1900,    0, 1969,    0,    0, -165,
    0,  792,    0,  482,    0,    0,    0,  -18,  284,  908,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,   60,    0,    0,    0,    0,    0,  396,    0,    0,
  481,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1486,    0, 1486,    0, 1069,    0,    0,    0,
    0, 1831,    0,    0,    0,  893,    0,    0,    0,  661,
    0,    0,    0,    0, 1000, 1486, 1555,    0,    0,    0,
    0,    0,    0,    0, 1486,    0,    0,    0,    0,    0,
    0,  453,    0,    0, 1486,    0,    0,    0,    0,    0,
    0, 1486,    0,    0,
};
short yygindex[] = {                                      0,
   -3,    0,   17,  -86, -214,    0,    0,  799,    0,    0,
    0,  -88, -105, -617,    0,    0, -410,  443, -186, -552,
 -628,    0, -227,    0,    0,    0,  444,    0,    0, -383,
    0,    0,    0,    0,    0,    0,    0,    6,  666,    0,
    0, 1088,    0,    0,  516,    0,  749,    0, -148,    0,
 -232,    0,  805,    0,    0,    0,    0,    0,    0,   10,
  489,    0,    0,  833, -290,  690,    0,    0,    0,    0,
    0, -103,  488,  994,    0,  -96,    0,  381,    0,  -51,
    0, 1136,    0,    0,    0,    0,    0,    0,    0,    0,
 -101,    0,    0,    0,    0,    0, 1117,  -39,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,  732,  733,    0,    0,  718,    0,
    0,  736,  738, -155, -271,    0,    0,    0,    0,  774,
    0,  -68,    0, 1012,    0,    0,    0,  900,    0,    0,
  903,  697, -373,    0,    0,    0,    0,    0,    0,    0,
  585,    0,  695, -411,  379,    0,    0,    0,  499,    0,
    0,    0,    0,    0,    0,  478,    0,  400,    0,    0,
  575,    0,  642,    0,    0,  584,    0,    0,    0,  475,
 -270,  739,    0,    0,    0, -180,  891,    0,    0,  638,
    0,    0,    0,  905, -275,    0,    0,    0,  780,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, -498,    0,    0,    0,    0,  834,  743,    0,    0,
    0,    0,    0, -472,    0,    0,    0,  469,    0,    0,
    0,    0,    0,  539,  923,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,
};
#define YYTABLESIZE 2500
short yytable[] = {                                      93,
   93,  222,  223,  162,  212,   57,  413,  373,  201,   60,
  236,  208,  226,  278,  431,  219,  243,  276,  196,   94,
   95,  630,  565,  376,  588,  422,  614,  170,  740,  519,
  118,  730,  238,  238,  448,  326,  162,  748,  238,  234,
  255,  553,  539,  412,  504,   92,  238,  238,  238,  238,
  238,  414,  535,  471,  199,  452,  535,  326,  524,   57,
  541,  617,   89,   60,  407,  238,  407,  537,  521,  171,
  409,  238,  344,   93,  415,  618,  619,  620,  621,  622,
  774,  521,  521,   93,  596,  407,  407,  553,  407,  775,
  444,  200,  238,  574,  306,  204,  301,  211,  149,  217,
  623,  535,   94,  287,  339,  211,  344,  344,   92,  230,
  790,  232,  235,  241,  776,   92,  537,   93,  300,  796,
  228,  624,  346,  238,  238,  287,  685,  686,  250,  520,
  238,  251,  701,   94,  407,  504,  407,  247,  749,  256,
  410,   96,  368,  369,  370,  371,  372,  422,  301,   52,
  728,  422,  625,  626,  539,  407,  407,  238,  407,  627,
  378,  325,   52,  615,  535,  417,  238,  403,  216,  452,
  272,  630,  541,  827,  818,  345,  301,  301,  301,  537,
  521,  151,  777,   93,  535,  200,  628,  417,  252,  321,
  128,  698,  338,  100,  301,  629,  663,  149,  409,  537,
  101,  128,  128,   52,  712,  303,  837,  324,  594,  346,
  265,  344,   52,  301,  234,  347,  128,  432,  721,  151,
  211,  357,  409,  500,  501,  502,  128,  363,  755,  756,
  757,  496,  446,   92,  536,  664,  397,  397,  422,  729,
  665,  666,  667,  200,  454,  454,  454,  467,  566,   52,
  529,  529,  529,  529,  238,  163,  382,  303,  103,   52,
  807,  266,  746,  523,   92,   92,  548,  104,  410,  164,
  383,   69,   69,  165,  457,  482,  458,  454,  267,  630,
  151,   92,  487,  529,  220,  303,  303,  303,   92,   93,
  149,  234,   69,  397,   93,  489,  403,  811,   93,  105,
  451,  451,  451,  303,  454,   92,  668,  528,  220,  416,
  239,  240,   92,   92,  128,  159,  433,   88,   88,  195,
   88,  299,  303,  685,  686,  211,  713,  203,  513,  166,
  151,  454,  455,  451,  204,  106,  160,  454,  454,  449,
  399,  299,  512,  529,  529,  454,  454,  454,  454,   92,
   88,  529,  529,  529,  529,  510,   92,  535,  149,  149,
  149,  454,  477,  432,  205,  234,  205,  529,  151,  205,
  209,  210,  744,  481,  541,  538,   93,  209,  210,  550,
  177,  178,  488,  432,  652,  448,  448,  448,  559,  149,
  205,  352,  486,  451,  451,  568,  484,  551,  151,  151,
  151,  451,  451,  451,  451,  521,  521,  521,  204,  819,
  446,  822,  653,  654,  655,  585,   92,  451,  448,  107,
   93,  731,  497,  256,  200,  200,  130,  608,  769,  151,
  656,  515,  833,  517,  674,  108,  521,  130,  130,  514,
  204,  838,  564,  256,  529,  516,  471,  700,  254,  657,
  521,  841,  130,  254,  335,   70,   70,  569,  844,  449,
  262,  263,  130,  391,  211,  570,  471,  211,  543,  649,
  357,  336,  363,  130,  130,  130,   70,  391,  448,  448,
  200,   92,  677,  109,  130,  130,  448,  448,  448,  448,
  130,  533,  685,  686,  521,  552,  130,  130,  130,  130,
  791,  170,  448,  211,  407,  192,  407,  521,  521,  130,
  424,  192,  174,  130,  584,  425,  792,  130,  130,  130,
  130,  589,  170,  280,  718,  407,  407,  130,  407,  110,
  523,  763,  170,  130,  305,  716,  649,  111,  130,   92,
  493,  170,  793,  523,  523,  112,  525,  525,  525,  493,
  685,  686,  130,  597,  493,  130,  130,  130,  493,  747,
  800,  113,  609,  801,  159,  750,  821,  306,  388,  389,
  215,  130,  215,  130,  824,  215,   92,  346,  151,  525,
  694,  114,  181,  762,  130,  695,  682,  346,  696,  529,
  115,  687,  130,  211,  559,  211,  215,  211,  211,  771,
  211,  683,  719,  720,  706,  282,  130,  130,  709,  470,
  825,  253,  244,  554,  493,  245,  253,  555,  784,  556,
  360,  211,  789,  717,  116,  428,  361,  307,  308,  470,
  834,  429,  715,  326,  524,  309,  310,  311,  312,  525,
  525,  299,  120,   35,   35,  687,  750,  525,  525,  525,
  525,  313,  739,  368,  369,  370,  371,  372,  283,  273,
   92,  794,  273,  525,  273,  302,  284,  724,  804,  754,
  724,  805,  725,  745,  285,  843,  211,  605,  117,  606,
  286,  200,  287,  758,  288,   92,  200,  119,  820,  267,
  823,  122,  267,    1,    2,    3,    4,    5,    6,  150,
  289,  597,    7,  289,  829,  220,  289,  390,  391,  392,
  393,  394,  783,  201,  395,  396,  239,  397,  201,   79,
   79,  239,  398,  778,  687,  687,  234,  368,  369,  370,
  371,  372,    8,    9,  250,  625,  626,   10,   11,  250,
   12,   13,   14,  687,  243,  379,  687,   15,  379,  243,
   84,   84,  151,   84,   16,   18,   17,   18,  154,   19,
   20,   21,   22,   23,   24,   25,   26,   27,   28,   29,
   30,   31,  423,  423,   32,   33,  427,  427,  810,   34,
   35,   96,   96,  155,   96,   36,   37,   38,  687,   39,
  169,   88,   88,  687,   88,   40,   41,  405,  156,  405,
   42,   90,   90,  157,   90,   43,  168,   44,  172,   45,
   46,  173,   47,  687,   48,   49,  354,  354,  174,   50,
  687,  180,  778,  368,  369,  370,  371,  372,   51,  123,
  123,  227,  354,   52,  192,   53,  193,   92,  354,  354,
  354,  354,  354,  354,  354,  354,  354,  202,  685,  686,
  354,  354,  117,  117,   52,  354,   61,   61,  207,  354,
  354,  354,  354,  234,  368,  369,  370,  371,  372,  354,
   65,   65,  135,  135,  229,  354,  560,  560,   43,   43,
  354,  390,  391,  392,  393,  394,  561,  561,  395,  396,
  254,  397,   81,   81,  354,  258,  398,  354,  354,  354,
   41,   41,   42,   42,  141,  141,  155,  155,  156,  156,
  157,  157,  260,  354,  279,  354,  355,  355,  495,  495,
  499,  499,  500,  500,  496,  496,  354,  270,  354,  498,
  499,  269,  355,  273,  354,  277,  281,  293,  355,  355,
  355,  355,  355,  355,  355,  355,  355,  295,  354,  354,
  355,  355,  297,  181,  322,  355,  326,  338,  341,  355,
  355,  355,  355,  343,  348,  350,  351,  170,  366,  355,
  344,  377,  379,  380,  385,  355,  408,  387,  435,  159,
  355,  436,  437,  460,  438,  221,  221,  221,  189,  440,
  443,  221,  453,  234,  355,  469,  471,  355,  355,  355,
  472,  474,  221,  476,  491,  485,  221,  221,  221,  221,
  221,  221,  478,  355,  492,  355,  494,  505,  518,  526,
  221,  495,  544,  549,  573,  221,  355,  553,  355,  221,
  221,  221,  221,  428,  355,  586,  590,  592,  593,  221,
  599,  594,  600,  611,  601,  221,  603,  612,  355,  355,
  221,  613,  616,  675,  221,  221,  221,  672,  676,  679,
  221,  680,  707,  692,  221,  710,  726,  221,  221,  221,
  737,  221,  741,  742,  760,  221,  221,  221,  221,  221,
  221,  765,  764,  221,  767,  221,  770,  772,  798,  221,
  566,  802,  786,  808,  221,  788,  221,  814,  221,  221,
  221,  221,  813,  816,  221,  831,  826,  832,  221,  817,
  835,  836,  839,  550,  221,  842,  193,  434,  221,  221,
  439,  194,  564,  234,  234,  234,  153,  417,  512,  234,
  389,  555,  507,  221,  408,  549,  221,  221,  221,  556,
  234,  460,  525,  542,  234,  234,  234,  234,  234,  234,
  400,  258,  221,  384,  221,   92,  386,  474,  234,  299,
  259,  475,  362,  234,  545,  221,  364,  234,  234,  234,
  234,  547,  795,  221,  797,  198,  511,  234,  480,  583,
  766,  558,  466,  234,  316,  768,  828,  221,  234,  158,
  176,  546,  238,  238,  238,  509,  299,  423,  711,  540,
  539,  580,  234,  542,  427,  234,  234,  234,  582,  238,
  840,  547,  830,  238,  238,  238,  238,  238,  238,  799,
  669,  234,  785,  234,  734,  727,  803,  238,  456,  534,
  684,  445,  238,  490,  234,  563,  238,  238,  238,  238,
  812,  442,  234,    0,  773,    0,  238,  581,    0,    0,
    0,    0,  238,    0,    0,    0,  234,  238,    0,    0,
    0,  233,  233,  233,    0,    0,    0,    0,    0,    0,
    0,  238,    0,    0,  238,  238,  238,    0,  233,    0,
    0,    0,  233,  233,  233,  233,  233,  233,    0,    0,
  238,    0,  238,    0,    0,    0,  233,    0,    0,    0,
    0,  233,    0,  238,    0,  233,  233,  233,  233,    0,
    0,  238,    0,    0,    0,  233,    0,    0,    0,    0,
    0,  233,    0,    0,    0,  238,  233,    0,    0,    0,
    0,  346,  346,    0,    0,    0,    0,    0,    0,    0,
  233,    0,    0,  233,  233,  233,    0,  346,    0,    0,
    0,    0,    0,  346,  346,  346,  346,    0,    0,  233,
    0,  233,    0,    0,    0,    0,  346,    0,    0,    0,
  346,    0,  233,    0,  346,  346,  346,  346,    0,    0,
  233,    0,    0,    0,  346,    0,    0,    0,    0,    0,
  346,    0,    0,    0,  233,  346,    0,    0,    0,    0,
    0,  344,  344,    0,    0,    0,    0,    0,    0,  346,
    0,    0,  346,  346,  346,    0,    0,  344,    0,    0,
    0,    0,    0,  344,  344,  344,  344,    0,  346,    0,
  346,    0,    0,    0,    0,    0,  344,    0,    0,    0,
  344,  346,    0,    0,  344,  344,  344,  344,    0,  346,
    0,    0,    0,    0,  344,    0,    0,    0,    0,    0,
  344,    0,    0,  346,  346,  344,    0,    0,    0,    0,
  484,  484,  484,    0,  484,  484,  484,    0,    0,  344,
  484,  484,  344,  344,  344,    0,    0,    0,    0,    0,
    0,  484,    0,    0,    0,    0,    0,    0,  344,    0,
  344,    0,    0,  484,  484,  484,    0,    0,    0,    0,
    0,  344,    0,    0,    0,    0,    0,    0,    0,  344,
  484,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  484,    0,  344,  344,  487,  487,  487,    0,  487,
  487,  487,    0,    0,    0,  487,  487,    0,    0,    0,
    0,    0,    0,    0,    0,    0,  487,    0,    0,    0,
    0,    0,    0,  484,  484,    0,    0,    0,  487,  487,
  487,  484,  484,  484,  484,    0,    0,    0,    0,    0,
    0,    0,    0,  484,    0,  487,    0,  484,    0,    0,
    0,    0,    0,    0,    0,    0,  487,    0,    0,    0,
    0,    0,    0,  484,    0,    0,    0,    0,    0,    0,
  293,  293,  293,    0,  293,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,  293,    0,  487,  487,
    0,    0,  293,  293,  293,  293,  487,  487,  487,  487,
    0,    0,    0,    0,  293,  293,    0,    0,  487,  293,
    0,    0,  487,  293,  293,  293,  293,    0,    0,    0,
    0,    0,    0,  293,    0,    0,    0,    0,  487,  293,
    0,    0,    0,    0,  293,    0,    0,    0,    0,  342,
  342,    0,    0,    0,    0,    0,    0,    0,  293,    0,
    0,  293,  293,  293,    0,  342,    0,    0,    0,    0,
    0,  342,  342,  342,  342,    0,    0,  293,    0,  293,
    0,    0,    0,  342,  342,    0,    0,    0,  342,    0,
  293,    0,  342,  342,  342,  342,    0,    0,  293,    0,
    0,    0,  342,    0,    0,    0,    0,    0,  342,    0,
    0,    0,  293,  342,    0,    0,    0,    0,  341,  341,
    0,    0,    0,    0,    0,    0,    0,  342,    0,    0,
  342,  342,  342,    0,  341,    0,    0,    0,    0,    0,
  341,  341,  341,  341,    0,    0,  342,    0,  342,    0,
    0,    0,  341,  341,    0,    0,    0,  341,    0,  342,
    0,  341,  341,  341,  341,    0,    0,  342,    0,    0,
    0,  341,    0,    0,    0,    0,    0,  341,    0,    0,
    0,  342,  341,    0,    0,    0,    0,  346,  346,    0,
    0,    0,    0,    0,    0,    0,  341,    0,    0,  341,
  341,  341,    0,  346,    0,    0,    0,    0,    0,  346,
  346,  346,  346,    0,    0,  341,    0,  341,    0,    0,
    0,  346,  346,    0,    0,    0,  346,    0,  341,    0,
  346,  346,  346,  346,    0,    0,  341,    0,    0,    0,
  346,    0,    0,    0,    0,    0,  346,    0,    0,    0,
  341,  346,    0,    0,    0,    0,  305,  315,    0,    0,
    0,    0,    0,    0,    0,  346,    0,    0,  346,  346,
  346,    0,  315,    0,    0,    0,    0,    0,  315,  315,
  315,  315,    0,    0,  346,    0,  346,    0,    0,    0,
  315,    0,    0,    0,    0,  305,    0,  346,    0,  315,
  305,  305,  305,    0,    0,  346,    0,    0,    0,  315,
    0,    0,    0,    0,    0,  315,    0,    0,    0,  346,
  315,    0,    0,    0,    0,  287,  287,    0,    0,    0,
    0,    0,    0,    0,  315,    0,    0,  315,  315,  315,
    0,  287,    0,    0,    0,    0,    0,  287,  287,  287,
  287,    0,    0,  315,    0,  315,    0,    0,    0,  287,
    0,    0,    0,    0,  287,    0,  305,    0,  287,  287,
  287,  287,    0,    0,  315,    0,    0,    0,  287,    0,
    0,    0,    0,    0,  287,    0,    0,    0,  315,  287,
    0,    0,    0,    0,  341,  341,    0,    0,    0,    0,
    0,    0,    0,  287,    0,    0,  287,  287,  287,    0,
  341,    0,    0,    0,    0,    0,  341,  341,  341,  341,
    0,    0,  287,    0,  287,    0,    0,    0,    0,  341,
    0,    0,    0,  341,    0,  287,    0,  341,  341,  341,
  341,    0,    0,  287,    0,    0,    0,  341,    0,    0,
    0,    0,    0,  341,    0,    0,    0,  287,  341,    0,
    0,    0,    0,  267,  267,    0,    0,    0,    0,    0,
    0,    0,  341,    0,    0,  341,  341,  341,    0,  267,
    0,    0,    0,    0,    0,  267,  267,  267,  267,    0,
    0,  341,    0,  341,    0,    0,    0,  267,    0,    0,
    0,    0,  267,    0,  341,    0,  267,  267,  267,  267,
    0,    0,  341,    0,    0,    0,  267,    0,    0,    0,
    0,    0,  267,    0,    0,    0,  341,  267,    0,    0,
    0,    0,  311,  311,    0,    0,    0,    0,    0,    0,
    0,  267,    0,    0,  267,  267,  267,    0,  311,    0,
    0,    0,    0,    0,  311,  311,  311,  311,    0,    0,
  267,    0,  267,    0,    0,    0,  311,    0,    0,    0,
    0,  311,    0,  267,    0,  311,  311,  311,  311,    0,
    0,  267,    0,    0,    0,  311,    0,    0,    0,    0,
    0,  311,    0,    0,    0,  267,  311,    0,    0,    0,
    0,  346,  346,    0,    0,    0,    0,    0,    0,    0,
  311,    0,    0,  311,  311,  311,    0,  346,    0,    0,
    0,    0,    0,  346,  346,  346,  346,    0,    0,  311,
    0,  311,    0,    0,    0,  346,    0,    0,    0,    0,
  346,    0,  311,    0,  346,  346,  346,  346,    0,    0,
  311,    0,    0,    0,  346,    0,    0,    0,    0,    0,
  346,    0,    0,    0,  311,  346,  316,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  346,
    0,  316,  346,  346,  346,    0,    0,  316,  316,  316,
  316,    0,    0,    0,    0,    0,  638,    0,  346,  316,
  346,    0,  639,  640,  641,  234,    0,    0,  316,    0,
    0,  346,    0,    0,  642,    0,    0,    0,  316,  346,
    0,    0,    0,  643,  316,    0,    0,    0,    0,  316,
    0,    0,    0,  346,    0,    0,    0,    0,  123,  644,
    0,    0,    0,  316,  645,  124,  316,  316,  316,    0,
    0,    0,    0,    0,    0,    0,    0,    0,  646,    0,
    0,  125,  316,    0,  316,  126,  127,  128,  129,  130,
  131,  132,  133,  134,  135,  136,  137,  647,    0,  138,
  139,    0,    0,  316,    0,  140,    0,    0,    0,    0,
    0,    0,  141,   92,    0,    0,    0,  316,    0,    0,
  142,  143,    0,    0,  388,  389,    0,    0,    0,    0,
  144,    0,  648,  220,    0,  390,  391,  392,  393,  394,
  145,    0,  395,  396,  146,  397,    0,    0,    0,    0,
  398,    0,    0,  147,    0,    0,    0,    0,    0,    0,
  148,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  399,
};
short yycheck[] = {                                       3,
    4,  105,  106,   55,  101,    0,  278,  235,   95,    0,
  116,  100,  109,  169,  305,  104,  118,  166,   87,    3,
    4,  574,  495,  238,  523,  296,  264,  274,  657,  440,
   34,  649,  263,  264,  268,  270,   88,  666,  269,  288,
  261,  266,  264,  276,  418,  262,  277,  278,  279,  280,
  281,  267,  264,  267,   94,  331,  268,  270,  271,   54,
  264,  263,  265,   54,  262,  296,  264,  264,  262,   60,
  268,  302,  291,   77,  290,  277,  278,  279,  280,  281,
  273,  275,  276,   87,  298,  283,  284,  312,  286,  282,
  323,   95,  323,  505,  296,   99,  264,  101,  297,  103,
  302,  313,  267,  264,  208,  109,  264,  265,  262,  113,
  728,  115,  116,  117,  307,  262,  313,  121,  187,  297,
  111,  323,  288,  354,  355,  286,  273,  274,  291,  313,
  361,  294,  605,  298,  262,  509,  264,  121,  387,  360,
  268,  275,  289,  290,  291,  292,  293,  418,  316,  396,
  297,  422,  354,  355,  376,  283,  284,  388,  286,  361,
  247,  201,  396,  401,  376,  400,  397,  271,  385,  445,
  161,  724,  376,  802,  792,  394,  344,  345,  346,  376,
  374,  297,  375,  187,  396,  189,  388,  400,  351,  193,
  262,  602,  267,  263,  362,  397,  263,  396,  396,  396,
  263,  273,  274,  396,  616,  264,  824,  198,  386,  375,
  311,  369,  396,  381,  288,  219,  288,  306,  629,  335,
  224,  225,  274,  257,  258,  259,  298,  231,  365,  366,
  367,  412,  329,  262,  462,  302,  266,  267,  509,  386,
  307,  308,  309,  247,  263,  264,  265,  351,  323,  396,
  263,  264,  265,  266,  283,  297,  260,  316,  263,  396,
  759,  362,  336,  444,  262,  262,  481,  263,  396,  311,
  261,  264,  265,  315,  282,  379,  284,  296,  379,  832,
  396,  262,  384,  296,  282,  344,  345,  346,  262,  293,
  297,  288,  285,  323,  298,  399,  400,  770,  302,  263,
  263,  264,  265,  362,  323,  262,  373,  288,  282,  293,
  339,  340,  262,  262,  386,  297,  307,  264,  265,  276,
  267,  262,  381,  273,  274,  329,  617,  276,  432,  371,
  297,  335,  336,  296,  338,  263,  318,  356,  357,  330,
  338,  282,  429,  356,  357,  364,  365,  366,  367,  262,
  297,  364,  365,  366,  367,  424,  262,  461,  365,  366,
  367,  380,  366,  262,  264,  288,  266,  380,  335,  269,
  283,  284,  663,  377,  471,  298,  380,  283,  284,  483,
  264,  265,  386,  282,  316,  263,  264,  265,  492,  396,
  290,  297,  383,  356,  357,  497,  380,  484,  365,  366,
  367,  364,  365,  366,  367,  365,  366,  367,  412,  793,
  507,  795,  344,  345,  346,  519,  262,  380,  296,  263,
  424,  649,  413,  262,  428,  429,  262,  273,  700,  396,
  362,  435,  816,  437,  583,  263,  396,  273,  274,  434,
  444,  825,  494,  282,  448,  436,  262,  603,  264,  381,
  441,  835,  288,  269,  273,  264,  265,  323,  842,  450,
  354,  355,  298,  262,  468,  331,  282,  471,  472,  575,
  474,  290,  476,  262,  263,  264,  285,  276,  356,  357,
  484,  262,  586,  263,  273,  274,  364,  365,  366,  367,
  279,  264,  273,  274,  262,  486,  285,  286,  287,  288,
  728,  274,  380,  507,  262,  262,  264,  275,  276,  298,
  264,  268,  264,  302,  518,  269,  297,  306,  307,  308,
  309,  525,  274,  265,  628,  283,  284,  316,  286,  263,
  262,  264,  274,  322,  263,  624,  642,  263,  327,  262,
  297,  274,  729,  275,  276,  263,  263,  264,  265,  306,
  273,  274,  341,  544,  311,  344,  345,  346,  315,  665,
  747,  263,  566,  750,  321,  671,  794,  296,  273,  274,
  264,  360,  266,  362,  297,  269,  262,  288,  335,  296,
  286,  263,  268,  680,  373,  291,  590,  298,  294,  593,
  263,  595,  381,  262,  698,  264,  290,  266,  267,  703,
  269,  592,  273,  274,  608,  295,  395,  396,  612,  262,
  797,  264,  291,  288,  371,  294,  269,  292,  722,  294,
  288,  290,  728,  627,  263,  267,  294,  356,  357,  282,
  817,  273,  623,  270,  271,  364,  365,  366,  367,  356,
  357,  262,  355,  264,  265,  649,  752,  364,  365,  366,
  367,  380,  656,  289,  290,  291,  292,  293,  348,  264,
  262,  297,  267,  380,  269,  267,  356,  264,  291,  673,
  264,  294,  269,  664,  364,  269,  680,  312,  263,  314,
  370,  264,  372,  674,  374,  262,  269,  263,  794,  264,
  796,  263,  267,  256,  257,  258,  259,  260,  261,  263,
  390,  692,  265,  264,  808,  282,  267,  284,  285,  286,
  287,  288,  716,  264,  291,  292,  264,  294,  269,  264,
  265,  269,  299,  714,  728,  729,  288,  289,  290,  291,
  292,  293,  295,  296,  264,  354,  355,  300,  301,  269,
  303,  304,  305,  747,  264,  264,  750,  310,  267,  269,
  264,  265,  263,  267,  317,  320,  319,  320,  263,  322,
  323,  324,  325,  326,  327,  328,  329,  330,  331,  332,
  333,  334,  264,  265,  337,  338,  264,  265,  769,  342,
  343,  264,  265,  263,  267,  348,  349,  350,  792,  352,
  321,  264,  265,  797,  267,  358,  359,  267,  263,  269,
  363,  264,  265,  263,  267,  368,  335,  370,  265,  372,
  373,  265,  375,  817,  377,  378,  263,  264,  264,  382,
  824,  263,  813,  289,  290,  291,  292,  293,  391,  264,
  265,  335,  279,  396,  266,  398,  297,  262,  285,  286,
  287,  288,  289,  290,  291,  292,  293,  266,  273,  274,
  297,  298,  264,  265,  396,  302,  264,  265,  273,  306,
  307,  308,  309,  288,  289,  290,  291,  292,  293,  316,
  264,  265,  264,  265,  341,  322,  264,  265,  264,  265,
  327,  284,  285,  286,  287,  288,  264,  265,  291,  292,
  320,  294,  264,  265,  341,  335,  299,  344,  345,  346,
  264,  265,  264,  265,  264,  265,  264,  265,  264,  265,
  264,  265,  297,  360,  396,  362,  263,  264,  264,  265,
  264,  265,  264,  265,  264,  265,  373,  399,  375,  414,
  415,  384,  279,  306,  381,  306,  265,  273,  285,  286,
  287,  288,  289,  290,  291,  292,  293,  266,  395,  396,
  297,  298,  290,  268,  374,  302,  270,  267,  324,  306,
  307,  308,  309,  297,  262,  282,  267,  274,  267,  316,
  291,  336,  266,  312,  369,  322,  298,  379,  263,  297,
  327,  263,  263,  263,  297,  262,  263,  264,  265,  268,
  298,  268,  264,  288,  341,  285,  267,  344,  345,  346,
  273,  267,  279,  267,  282,  298,  283,  284,  285,  286,
  287,  288,  369,  360,  267,  362,  312,  268,  273,  286,
  297,  312,  297,  264,  273,  302,  373,  265,  375,  306,
  307,  308,  309,  267,  381,  376,  273,  269,  267,  316,
  264,  386,  282,  265,  298,  322,  277,  263,  395,  396,
  327,  263,  269,  298,  262,  263,  264,  275,  269,  264,
  268,  267,  389,  267,  341,  341,  286,  344,  345,  346,
  277,  279,  264,  267,  284,  283,  284,  285,  286,  287,
  288,  274,  273,  360,  298,  362,  312,  267,  258,  297,
  323,  375,  282,  274,  302,  298,  373,  282,  306,  307,
  308,  309,  267,  298,  381,  282,  275,  268,  316,  298,
  298,  298,  264,  262,  322,  298,  266,  265,  395,  327,
  306,  266,  396,  262,  263,  264,  396,  264,  266,  268,
  266,  266,  265,  341,  265,  265,  344,  345,  346,  266,
  279,  267,  266,  264,  283,  284,  285,  286,  287,  288,
  269,  264,  360,  262,  362,  298,  264,  298,  297,  262,
  264,  363,  375,  302,  264,  373,  375,  306,  307,  308,
  309,  264,  730,  381,  731,   88,  428,  316,  374,  514,
  692,  492,  350,  322,  191,  698,  806,  395,  327,   54,
   74,  474,  262,  263,  264,  422,  185,  298,  614,  468,
  463,  507,  341,  471,  302,  344,  345,  346,  512,  279,
  832,  476,  813,  283,  284,  285,  286,  287,  288,  742,
  579,  360,  724,  362,  650,  642,  752,  297,  338,  450,
  593,  327,  302,  400,  373,  493,  306,  307,  308,  309,
  772,  319,  381,   -1,  706,   -1,  316,  509,   -1,   -1,
   -1,   -1,  322,   -1,   -1,   -1,  395,  327,   -1,   -1,
   -1,  262,  263,  264,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  341,   -1,   -1,  344,  345,  346,   -1,  279,   -1,
   -1,   -1,  283,  284,  285,  286,  287,  288,   -1,   -1,
  360,   -1,  362,   -1,   -1,   -1,  297,   -1,   -1,   -1,
   -1,  302,   -1,  373,   -1,  306,  307,  308,  309,   -1,
   -1,  381,   -1,   -1,   -1,  316,   -1,   -1,   -1,   -1,
   -1,  322,   -1,   -1,   -1,  395,  327,   -1,   -1,   -1,
   -1,  263,  264,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  341,   -1,   -1,  344,  345,  346,   -1,  279,   -1,   -1,
   -1,   -1,   -1,  285,  286,  287,  288,   -1,   -1,  360,
   -1,  362,   -1,   -1,   -1,   -1,  298,   -1,   -1,   -1,
  302,   -1,  373,   -1,  306,  307,  308,  309,   -1,   -1,
  381,   -1,   -1,   -1,  316,   -1,   -1,   -1,   -1,   -1,
  322,   -1,   -1,   -1,  395,  327,   -1,   -1,   -1,   -1,
   -1,  263,  264,   -1,   -1,   -1,   -1,   -1,   -1,  341,
   -1,   -1,  344,  345,  346,   -1,   -1,  279,   -1,   -1,
   -1,   -1,   -1,  285,  286,  287,  288,   -1,  360,   -1,
  362,   -1,   -1,   -1,   -1,   -1,  298,   -1,   -1,   -1,
  302,  373,   -1,   -1,  306,  307,  308,  309,   -1,  381,
   -1,   -1,   -1,   -1,  316,   -1,   -1,   -1,   -1,   -1,
  322,   -1,   -1,  395,  396,  327,   -1,   -1,   -1,   -1,
  263,  264,  265,   -1,  267,  268,  269,   -1,   -1,  341,
  273,  274,  344,  345,  346,   -1,   -1,   -1,   -1,   -1,
   -1,  284,   -1,   -1,   -1,   -1,   -1,   -1,  360,   -1,
  362,   -1,   -1,  296,  297,  298,   -1,   -1,   -1,   -1,
   -1,  373,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  381,
  313,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  324,   -1,  395,  396,  263,  264,  265,   -1,  267,
  268,  269,   -1,   -1,   -1,  273,  274,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  284,   -1,   -1,   -1,
   -1,   -1,   -1,  356,  357,   -1,   -1,   -1,  296,  297,
  298,  364,  365,  366,  367,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  376,   -1,  313,   -1,  380,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  324,   -1,   -1,   -1,
   -1,   -1,   -1,  396,   -1,   -1,   -1,   -1,   -1,   -1,
  263,  264,  265,   -1,  267,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  279,   -1,  356,  357,
   -1,   -1,  285,  286,  287,  288,  364,  365,  366,  367,
   -1,   -1,   -1,   -1,  297,  298,   -1,   -1,  376,  302,
   -1,   -1,  380,  306,  307,  308,  309,   -1,   -1,   -1,
   -1,   -1,   -1,  316,   -1,   -1,   -1,   -1,  396,  322,
   -1,   -1,   -1,   -1,  327,   -1,   -1,   -1,   -1,  263,
  264,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  341,   -1,
   -1,  344,  345,  346,   -1,  279,   -1,   -1,   -1,   -1,
   -1,  285,  286,  287,  288,   -1,   -1,  360,   -1,  362,
   -1,   -1,   -1,  297,  298,   -1,   -1,   -1,  302,   -1,
  373,   -1,  306,  307,  308,  309,   -1,   -1,  381,   -1,
   -1,   -1,  316,   -1,   -1,   -1,   -1,   -1,  322,   -1,
   -1,   -1,  395,  327,   -1,   -1,   -1,   -1,  263,  264,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  341,   -1,   -1,
  344,  345,  346,   -1,  279,   -1,   -1,   -1,   -1,   -1,
  285,  286,  287,  288,   -1,   -1,  360,   -1,  362,   -1,
   -1,   -1,  297,  298,   -1,   -1,   -1,  302,   -1,  373,
   -1,  306,  307,  308,  309,   -1,   -1,  381,   -1,   -1,
   -1,  316,   -1,   -1,   -1,   -1,   -1,  322,   -1,   -1,
   -1,  395,  327,   -1,   -1,   -1,   -1,  263,  264,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  341,   -1,   -1,  344,
  345,  346,   -1,  279,   -1,   -1,   -1,   -1,   -1,  285,
  286,  287,  288,   -1,   -1,  360,   -1,  362,   -1,   -1,
   -1,  297,  298,   -1,   -1,   -1,  302,   -1,  373,   -1,
  306,  307,  308,  309,   -1,   -1,  381,   -1,   -1,   -1,
  316,   -1,   -1,   -1,   -1,   -1,  322,   -1,   -1,   -1,
  395,  327,   -1,   -1,   -1,   -1,  263,  264,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  341,   -1,   -1,  344,  345,
  346,   -1,  279,   -1,   -1,   -1,   -1,   -1,  285,  286,
  287,  288,   -1,   -1,  360,   -1,  362,   -1,   -1,   -1,
  297,   -1,   -1,   -1,   -1,  302,   -1,  373,   -1,  306,
  307,  308,  309,   -1,   -1,  381,   -1,   -1,   -1,  316,
   -1,   -1,   -1,   -1,   -1,  322,   -1,   -1,   -1,  395,
  327,   -1,   -1,   -1,   -1,  263,  264,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  341,   -1,   -1,  344,  345,  346,
   -1,  279,   -1,   -1,   -1,   -1,   -1,  285,  286,  287,
  288,   -1,   -1,  360,   -1,  362,   -1,   -1,   -1,  297,
   -1,   -1,   -1,   -1,  302,   -1,  373,   -1,  306,  307,
  308,  309,   -1,   -1,  381,   -1,   -1,   -1,  316,   -1,
   -1,   -1,   -1,   -1,  322,   -1,   -1,   -1,  395,  327,
   -1,   -1,   -1,   -1,  263,  264,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  341,   -1,   -1,  344,  345,  346,   -1,
  279,   -1,   -1,   -1,   -1,   -1,  285,  286,  287,  288,
   -1,   -1,  360,   -1,  362,   -1,   -1,   -1,   -1,  298,
   -1,   -1,   -1,  302,   -1,  373,   -1,  306,  307,  308,
  309,   -1,   -1,  381,   -1,   -1,   -1,  316,   -1,   -1,
   -1,   -1,   -1,  322,   -1,   -1,   -1,  395,  327,   -1,
   -1,   -1,   -1,  263,  264,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  341,   -1,   -1,  344,  345,  346,   -1,  279,
   -1,   -1,   -1,   -1,   -1,  285,  286,  287,  288,   -1,
   -1,  360,   -1,  362,   -1,   -1,   -1,  297,   -1,   -1,
   -1,   -1,  302,   -1,  373,   -1,  306,  307,  308,  309,
   -1,   -1,  381,   -1,   -1,   -1,  316,   -1,   -1,   -1,
   -1,   -1,  322,   -1,   -1,   -1,  395,  327,   -1,   -1,
   -1,   -1,  263,  264,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  341,   -1,   -1,  344,  345,  346,   -1,  279,   -1,
   -1,   -1,   -1,   -1,  285,  286,  287,  288,   -1,   -1,
  360,   -1,  362,   -1,   -1,   -1,  297,   -1,   -1,   -1,
   -1,  302,   -1,  373,   -1,  306,  307,  308,  309,   -1,
   -1,  381,   -1,   -1,   -1,  316,   -1,   -1,   -1,   -1,
   -1,  322,   -1,   -1,   -1,  395,  327,   -1,   -1,   -1,
   -1,  263,  264,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  341,   -1,   -1,  344,  345,  346,   -1,  279,   -1,   -1,
   -1,   -1,   -1,  285,  286,  287,  288,   -1,   -1,  360,
   -1,  362,   -1,   -1,   -1,  297,   -1,   -1,   -1,   -1,
  302,   -1,  373,   -1,  306,  307,  308,  309,   -1,   -1,
  381,   -1,   -1,   -1,  316,   -1,   -1,   -1,   -1,   -1,
  322,   -1,   -1,   -1,  395,  327,  264,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  341,
   -1,  279,  344,  345,  346,   -1,   -1,  285,  286,  287,
  288,   -1,   -1,   -1,   -1,   -1,  279,   -1,  360,  297,
  362,   -1,  285,  286,  287,  288,   -1,   -1,  306,   -1,
   -1,  373,   -1,   -1,  297,   -1,   -1,   -1,  316,  381,
   -1,   -1,   -1,  306,  322,   -1,   -1,   -1,   -1,  327,
   -1,   -1,   -1,  395,   -1,   -1,   -1,   -1,  296,  322,
   -1,   -1,   -1,  341,  327,  303,  344,  345,  346,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  341,   -1,
   -1,  319,  360,   -1,  362,  323,  324,  325,  326,  327,
  328,  329,  330,  331,  332,  333,  334,  360,   -1,  337,
  338,   -1,   -1,  381,   -1,  343,   -1,   -1,   -1,   -1,
   -1,   -1,  350,  262,   -1,   -1,   -1,  395,   -1,   -1,
  358,  359,   -1,   -1,  273,  274,   -1,   -1,   -1,   -1,
  368,   -1,  395,  282,   -1,  284,  285,  286,  287,  288,
  378,   -1,  291,  292,  382,  294,   -1,   -1,   -1,   -1,
  299,   -1,   -1,  391,   -1,   -1,   -1,   -1,   -1,   -1,
  398,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  338,
};
#define YYFINAL 54
#ifndef YYDEBUG
#define YYDEBUG 0
#endif
#define YYMAXTOKEN 401
#if YYDEBUG
char *yyname[] = {
"end-of-file",0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,"T_SCORE","T_STAFF","T_VOICE",
"T_GRIDS","T_BLOCKHEAD","T_NUMBER","T_EQUAL","T_SEMICOLON","T_NEWLINE",
"T_COLON","T_COMMA","T_LBRACKET","T_RBRACKET","T_LBRACE","T_RBRACE",
"T_VARIABLE","T_DASH","T_PLUS","T_PLACE","T_ALL","T_WITH","T_CUE","T_GRACE",
"T_XNOTE","T_DIAM","T_DOT","T_MULTIWHOLE","T_LET_M","T_LET_R","T_LET_S",
"T_LET_U","T_PITCH","T_SHARP","T_AMPERSAND","T_LET_N","T_LET_X","T_DBLFLAT",
"T_LETTER","T_LYRICS","T_PAD","T_LPAREN","T_RPAREN","T_LVAR","T_MUSIC",
"T_HEADSHAPES","T_HS","T_SWINGUNIT","T_POSTSCRIPT","T_SCOREFEED","T_BARTYPE",
"T_HAT","T_TILDE","T_QUESTION","T_LINETYPE","T_LINE","T_TO","T_STAR","T_BULGE",
"T_CURVE","T_TIE","T_PRINTTYPE","T_NL","T_STAFFLINES","T_FFAMILY","T_PARAGRAPH",
"T_PARATYPE","T_NUMVAR","T_FNUMVAR","T_2FNUMVAR","T_RATNUMLISTVAR",
"T_RANGELISTVAR","T_TIMEUNIT","T_FONTVAR","T_STRVAR","T_CLEFVAR","T_VVAR",
"T_BARSTLISTVAR","T_SCORESEP","T_FONT","T_SLASH","T_KEY","T_TIME","T_CUT",
"T_COMMON","T_CLEF","T_MULTIREST","T_YESNOVAR","T_BM","T_ESBM","T_EBM",
"T_INVISBAR","T_OCTAVE","T_ROLL","T_VISVAR","T_WHEREUSED","T_UNSET","T_WORD",
"T_UP","T_DOWN","T_ENDING","T_ENDENDING","T_FAMILY","T_ENDSTYLE","T_ENDTYPE",
"T_SLASHMARK","T_ALT","T_TITLE","T_REHEARSAL","T_RNUM","T_LET","T_MNUM",
"T_TRANSPOSE","T_XPOS_INT","T_PEDAL","T_PHRASE","T_MUSSYM","T_L_ANGLE",
"T_MODIFIER","T_R_ANGLE","T_TIL","T_MIDI","T_PEDSTYLE","T_PEDSTAR",
"T_HIDECHANGES","T_SLUR","T_UNITS","T_CENTIMETERS","T_UNITTYPE","T_TAB",
"T_TICKS","T_NOWHERE_SLIDE","T_LEN","T_EXCLAM","T_OTHERTEXT","T_ORDER",
"T_GRIDPLACE","T_GRPLVALUE","T_DRUM","T_RPT","T_STRING","T_HO","T_REH_STYLE",
"T_REHTYPE","T_L_DBLANGLE","T_R_DBLANGLE",
};
char *yyrule[] = {
"$accept : song",
"song : item",
"song : song item",
"item : context",
"item : assignment",
"item : title",
"item : paragraph",
"item : noteinfo T_NEWLINE",
"item : directive T_NEWLINE",
"item : stringpair",
"item : os_directive opt_semi T_NEWLINE",
"item : T_NEWLINE",
"item : error T_NEWLINE",
"context : ssv_context",
"context : blockhead_context",
"context : grid_context",
"context : headshape_context",
"context : music_context new_context",
"ssv_context : T_SCORE alloc_ssv",
"ssv_context : T_STAFF staffnum alloc_ssv",
"ssv_context : T_VOICE staffnum voicenum alloc_ssv",
"alloc_ssv : new_context",
"new_context :",
"blockhead_context : T_BLOCKHEAD",
"music_context : T_MUSIC",
"grid_context : T_GRIDS",
"stringpair : string string T_NEWLINE",
"headshape_context : T_HEADSHAPES",
"staffnum : num",
"voicenum : num",
"num : T_NUMBER",
"assignment : assign terminator",
"terminator : T_SEMICOLON",
"terminator : T_NEWLINE",
"assign : T_NUMVAR T_EQUAL opt_minus num",
"assign : T_NUMVAR T_EQUAL",
"assign : T_FNUMVAR T_EQUAL floatnum",
"assign : T_2FNUMVAR T_EQUAL floatnum T_COMMA floatnum",
"assign : T_PAD T_EQUAL opt_minus floatnum",
"assign : T_VVAR T_EQUAL num vschemetype",
"assign : T_RATNUMLISTVAR T_EQUAL allocbeamlist ratnumlist opt_r",
"assign : T_RANGELISTVAR T_EQUAL alloc_rangelist rangelist",
"assign : T_BARSTLISTVAR T_EQUAL alloc_barstlist barstlist",
"assign : mark_order T_EQUAL orderlist",
"assign : T_KEY T_EQUAL keyspec",
"assign : T_TIME T_EQUAL timesig",
"assign : T_STAFFLINES T_EQUAL stafflinedef",
"assign : T_FONTVAR T_EQUAL T_FONT",
"assign : T_FAMILY T_EQUAL T_FFAMILY",
"assign : T_CLEFVAR T_EQUAL T_CLEF",
"assign : T_YESNOVAR T_EQUAL flagval",
"assign : T_VISVAR T_EQUAL visval",
"assign : T_UNSET paramname",
"assign : T_TIMEUNIT T_EQUAL time_item",
"assign : T_SWINGUNIT T_EQUAL swing_time",
"assign : T_SCORESEP T_EQUAL minscsep maxscsep",
"assign : T_ENDSTYLE T_EQUAL end_type",
"assign : T_REH_STYLE T_EQUAL T_REHTYPE",
"assign : T_PEDSTYLE T_EQUAL pedstyle",
"assign : T_TRANSPOSE T_EQUAL updown transpose_interval num",
"assign : T_UNITS T_EQUAL T_UNITTYPE",
"assign : T_STRVAR T_EQUAL string",
"visval : T_WHEREUSED",
"visval : T_LETTER",
"visval : T_LET_N",
"vschemetype :",
"vschemetype : T_PITCH",
"vschemetype : T_LETTER",
"allocbeamlist :",
"ratnumlist :",
"ratnumlist : nonempty_ratnumlist",
"nonempty_ratnumlist : time_item",
"nonempty_ratnumlist : subbeam",
"nonempty_ratnumlist : nonempty_ratnumlist T_COMMA time_item",
"nonempty_ratnumlist : nonempty_ratnumlist T_COMMA subbeam",
"$$1 :",
"subbeam : T_LPAREN $$1 nonempty_ratnumlist T_RPAREN",
"time_item : basic_time_val dots opt_m",
"time_item : T_LET_M",
"swing_time :",
"swing_time : time_item",
"opt_r :",
"opt_r : T_LET_R",
"alloc_rangelist :",
"rangelist :",
"rangelist : rangeitem",
"rangelist : rangelist T_COMMA rangeitem",
"rangeitem : num rangetail grplabel",
"rangetail :",
"rangetail : T_DASH num",
"grplabel :",
"grplabel : T_LPAREN glabel optlabel2 T_RPAREN",
"optlabel2 :",
"optlabel2 : T_COMMA glabel",
"glabel : string",
"alloc_barstlist :",
"barstlist :",
"barstlist : barst_item",
"barstlist : barstlist T_COMMA barst_item",
"barst_item : num rangetail",
"mark_order : T_ORDER",
"orderlist : orderitem",
"orderlist : orderlist T_COMMA orderitem",
"orderlist : orderlist T_AMPERSAND orderitem",
"orderitem : marktype",
"marktype : T_MUSSYM",
"marktype : T_OCTAVE",
"marktype : T_OTHERTEXT",
"marktype : T_MODIFIER",
"marktype : T_LYRICS",
"marktype : T_ENDING",
"marktype : T_REHEARSAL",
"marktype : T_PEDAL",
"timesig : T_MULTIWHOLE opt_n",
"timesig : num T_SLASH num opt_n",
"timesig : T_CUT",
"timesig : T_COMMON",
"opt_n :",
"opt_n : T_LET_N",
"keyspec : num acc_symbol majmin",
"keyspec : pitch opt_accidental majmin",
"stafflinedef : num opt_n",
"stafflinedef : num T_DRUM",
"stafflinedef : T_TAB",
"stafflinedef : T_TAB T_LPAREN tab_string_list T_RPAREN",
"tab_string_list : tab_item",
"tab_string_list : tab_string_list tab_item",
"tab_item : pitch opt_str_acc opt_ticks opt_octave",
"opt_str_acc :",
"opt_str_acc : acc_symbol",
"opt_ticks :",
"opt_ticks : T_TICKS",
"flagval : T_LETTER",
"flagval : T_LET_N",
"minscsep : num",
"maxscsep :",
"maxscsep : T_COMMA num",
"end_type : T_ENDTYPE",
"end_type : T_BLOCKHEAD",
"updown : T_UP",
"updown : T_DOWN",
"majmin :",
"majmin : T_XPOS_INT",
"pedstyle : T_LINE",
"pedstyle : T_ALT T_PEDSTAR",
"pedstyle : T_PEDSTAR",
"transpose_interval : T_XPOS_INT",
"title : T_TITLE tfont title_size titleA opt_semi T_NEWLINE",
"tfont : font_family title_font",
"title_font :",
"title_font : T_FONT",
"font_family :",
"font_family : T_FFAMILY",
"title_size :",
"title_size : T_LPAREN num T_RPAREN",
"titleA : string",
"titleA : string string",
"titleA : string string string",
"paragraph : opt_paratype T_PARAGRAPH tfont title_size string opt_semi T_NEWLINE",
"opt_paratype :",
"opt_paratype : T_PARATYPE",
"paramname : T_NUMVAR",
"paramname : T_YESNOVAR",
"paramname : T_SCORESEP",
"paramname : T_STAFFLINES",
"paramname : T_TIMEUNIT",
"paramname : T_SWINGUNIT",
"paramname : T_PAD",
"paramname : T_BARSTLISTVAR",
"paramname : T_FNUMVAR",
"paramname : T_2FNUMVAR",
"paramname : T_UNITS",
"paramname : T_ENDSTYLE",
"paramname : T_REH_STYLE",
"paramname : T_PEDSTYLE",
"paramname : T_VVAR",
"paramname : T_RANGELISTVAR",
"paramname : T_ORDER",
"paramname : T_KEY",
"paramname : T_TIME",
"paramname : T_TRANSPOSE",
"paramname : T_RATNUMLISTVAR",
"paramname : T_VISVAR",
"paramname : T_STRVAR",
"paramname : T_FONTVAR",
"paramname : T_FAMILY",
"paramname : T_CLEFVAR",
"string : T_STRING",
"string : string T_PLUS T_STRING",
"noteinfo : init_staff sv_spec notecolon groupinfo",
"noteinfo : T_MULTIREST num opt_semi",
"noteinfo : lyricinfo",
"init_staff :",
"sv_spec : voice_id",
"sv_spec : sv_map_list",
"sv_map_list : sv_map_entry",
"sv_map_list : sv_map_list sv_map_entry",
"sv_map_entry : map_start sv_map_item_list map_finish",
"map_start : T_LBRACKET",
"map_finish : T_RBRACKET",
"sv_map_item_list : voice_id",
"sv_map_item_list : sv_map_item_list T_SEMICOLON voice_id",
"voice_id : sv_item",
"voice_id : voice_id sv_and sv_item",
"sv_and : T_AMPERSAND",
"sv_item : grp_stafflist",
"sv_item : grp_stafflist voice_spec",
"notecolon : T_COLON",
"grp_stafflist : grp_staff_item",
"grp_stafflist : grp_stafflist T_COMMA grp_staff_item",
"grp_staff_item : staffnum grp_staff_range",
"grp_staff_range :",
"grp_staff_range : T_DASH staffnum",
"voice_spec : voicenum voice_tail",
"voice_spec : voicenum T_DASH voicenum voice_tail",
"voice_tail :",
"voice_tail : T_COMMA voice_spec",
"groupinfo : groupdata",
"groupinfo : groupinfo groupdata",
"groupdata : ntuplet",
"groupdata : mid_meas_param grp_attributes group other_attr T_SEMICOLON",
"mid_meas_param :",
"mid_meas_param : T_L_DBLANGLE mm_context mm_params T_R_DBLANGLE",
"mm_context : T_SCORE",
"mm_context : T_STAFF",
"mm_context : T_VOICE",
"mm_params : mm_param_item",
"mm_params : mm_params T_SEMICOLON mm_param_item",
"mm_param_item : T_CLEFVAR T_EQUAL T_CLEF",
"mm_param_item : T_NUMVAR T_EQUAL num",
"grp_attributes : styleinfo time_val",
"styleinfo : alloc_grpsyl",
"styleinfo : T_LBRACKET T_DASH T_RBRACKET alloc_grpsyl",
"styleinfo : T_LBRACKET alloc_grpsyl marklist T_RBRACKET gtc extra_styleinfo",
"gtc :",
"gtc : T_DOT T_DOT T_DOT",
"extra_styleinfo :",
"extra_styleinfo : extra_styleinfo T_LBRACKET marklist T_RBRACKET",
"alloc_grpsyl :",
"marklist :",
"marklist : marklistA",
"marklistA : markitem",
"marklistA : marklist T_SEMICOLON markitem",
"markitem : T_WITH alloc_marklist marks",
"markitem : grp_locsave",
"markitem : notetype",
"markitem : padding",
"markitem : dir",
"markitem : T_NUMVAR opt_minus num",
"markitem : T_HO hor_offset",
"markitem : T_HS string",
"markitem : T_LEN floatnum",
"markitem : T_SLASHMARK num",
"hor_offset : T_PLUS",
"hor_offset : T_DASH",
"hor_offset : opt_plus_or_minus floatnum",
"opt_plus_or_minus :",
"opt_plus_or_minus : plus_or_minus",
"other_attr :",
"other_attr : attr_list",
"attr_list : attr_item",
"attr_list : attr_list T_COMMA attr_item",
"attr_item : beammark",
"attr_item : T_ALT num",
"attr_item : T_SLUR opt_dir",
"attr_item : T_TIE opt_dir",
"alloc_marklist :",
"opt_dir :",
"opt_dir : dir",
"marks : mark",
"marks : marks T_COMMA mark",
"mark : mark_item",
"mark_item : shorthand_mark",
"mark_item : string",
"shorthand_mark : T_DOT",
"shorthand_mark : T_DASH",
"shorthand_mark : T_R_ANGLE",
"shorthand_mark : T_HAT",
"notetype : T_CUE",
"notetype : T_GRACE",
"notetype : T_DIAM",
"notetype : T_XNOTE",
"dir : T_UP",
"dir : T_DOWN",
"beammark : T_BM crossbeam",
"beammark : T_ESBM",
"beammark : T_EBM",
"time_val :",
"time_val : time_item",
"crossbeam :",
"crossbeam : T_WITH T_STAFF T_PLACE",
"basic_time_val : T_MULTIWHOLE",
"basic_time_val : num",
"opt_m :",
"opt_m : T_LET_M",
"dots :",
"dots : dots T_DOT",
"padding : padspec",
"padspec : T_PAD opt_minus floatnum",
"opt_minus :",
"opt_minus : T_DASH",
"group : notelist",
"group : special_group",
"special_group : extra_attr note_opts",
"extra_attr : spec_grp_alloc extra_item",
"spec_grp_alloc :",
"extra_item : T_QUESTION",
"extra_item : T_TILDE opt_dir",
"extra_item : T_EQUAL loc_variable",
"extra_item : T_HAT pitch opt_octave",
"extra_item : T_HAT T_SLASH",
"extra_item : T_HS string",
"extra_item : slur_item",
"note_opts :",
"note_opts : note_opts extra_item",
"notelist :",
"notelist : notelist notedata note_opts",
"notedata : notedata1",
"notedata : T_LET_R",
"notedata : T_LET_S",
"notedata : T_LET_U T_LET_S",
"notedata : T_RPT",
"notedata : T_BARTYPE",
"notedata : T_GRACE",
"notedata : T_ENDTYPE",
"notedata : T_PARATYPE",
"notedata : T_RANGELISTVAR",
"notedata : T_CLEF",
"notedata : T_LPAREN notedata1 T_RPAREN",
"notedata1 : pitch opt_bend",
"notedata1 : pitch T_LPAREN acc_symbol T_RPAREN opt_octave",
"notedata1 : pitch T_LPAREN octave T_RPAREN opt_bend",
"notedata1 : pitch T_LPAREN pitch",
"notedata1 : pitch acc_symbol opt_ticks opt_octave opt_bend",
"notedata1 : pitch acc_symbol opt_ticks T_LPAREN octave T_RPAREN opt_bend",
"notedata1 : pitch acc_symbol T_LPAREN pitch",
"notedata1 : pitch T_TICKS opt_octave opt_bend",
"notedata1 : pitch T_TICKS T_LPAREN octave T_RPAREN opt_bend",
"notedata1 : pitch octave opt_accidental opt_bend",
"notedata1 : pitch octave T_LPAREN acc_symbol T_RPAREN",
"notedata1 : pitch octave T_LPAREN pitch",
"opt_bend :",
"opt_bend : string",
"pitch : T_PITCH",
"opt_accidental :",
"opt_accidental : acc_symbol",
"opt_octave :",
"opt_octave : octave",
"acc_symbol : T_SHARP",
"acc_symbol : T_AMPERSAND",
"acc_symbol : T_LET_N",
"acc_symbol : T_LET_X",
"acc_symbol : T_DBLFLAT",
"octave : num",
"octave : minuslist",
"octave : pluslist",
"minuslist : T_DASH",
"minuslist : minuslist T_DASH",
"pluslist : T_PLUS",
"pluslist : pluslist T_PLUS",
"slur_item : start_slur slurlist T_R_ANGLE opt_dir",
"start_slur : T_L_ANGLE",
"slurlist :",
"slurlist : T_NOWHERE_SLIDE",
"slurlist : fullslurlist",
"fullslurlist : slurnote",
"fullslurlist : fullslurlist slurnote",
"slurnote : pitch opt_octave",
"grp_locsave : T_EQUAL loc_variable",
"loc_variable : T_LETTER",
"loc_variable : T_PITCH",
"loc_variable : T_LET_M",
"loc_variable : T_LET_R",
"loc_variable : T_LET_S",
"loc_variable : T_LET_U",
"loc_variable : T_LET_N",
"loc_variable : T_LET_X",
"loc_variable : T_LVAR",
"ntuplet : start_tuplet groupinfo rbrace opt_side num printtup opt_time T_SEMICOLON",
"printtup :",
"printtup : T_LET_N",
"printtup : T_LETTER",
"start_tuplet : T_LBRACE",
"rbrace : T_RBRACE",
"opt_side :",
"opt_side : T_PLACE",
"opt_time :",
"opt_time : T_COMMA time_item",
"lyricinfo : T_LYRICS lyric_id T_COLON lyricdata",
"lyric_id : lyr_opt_place staff_list",
"lyr_opt_place : opt_place",
"opt_place :",
"opt_place : T_PLACE",
"staff_list : staff_item",
"staff_list : T_ALL",
"staff_list : staff_list T_COMMA staff_item",
"staff_item : num staff_range",
"staff_range :",
"staff_range : T_DASH num",
"staff_range : T_AMPERSAND num",
"verse_spec : verselist",
"verse_spec : T_PITCH",
"verselist : verse_item",
"verselist : verselist T_COMMA verse_item",
"verse_item : num verse_range",
"verse_range :",
"verse_range : T_DASH num",
"alloc_lyr_grpsyl :",
"lyricdata : lyr_tlist lyr_verse",
"lyr_tlist :",
"lyr_tlist : lyrics_timelist",
"lyrics_timelist : lyr_time_item",
"lyrics_timelist : lyrics_timelist lyr_time_item",
"lyr_time_item : lyr_tm T_SEMICOLON",
"lyr_tm : alloc_lyr_grpsyl time_val opt_space",
"lyr_tm : lyrtuplet",
"lyrtuplet : start_tuplet lyrics_timelist rbrace num opt_time",
"opt_space :",
"opt_space : T_LET_S",
"lyr_verse : verse_string",
"lyr_verse : lyr_verse verse_string",
"verse_string : string T_SEMICOLON",
"verse_string : T_LBRACKET verse_spec T_RBRACKET string T_SEMICOLON",
"os_directive : barinfo bar_items",
"os_directive : line",
"os_directive : curve",
"$$2 :",
"os_directive : T_SCOREFEED $$2 opt_temp_margin",
"os_directive : printcmd",
"opt_temp_margin :",
"opt_temp_margin : opt_temp_margin temp_margin",
"temp_margin : T_FNUMVAR opt_equals floatnum",
"opt_equals :",
"opt_equals : T_EQUAL",
"opt_semi :",
"opt_semi : T_SEMICOLON",
"directive : stuff_info",
"directive : roll_cmd",
"barinfo : linetype precbartype T_BARTYPE",
"precbartype :",
"precbartype : T_LPAREN T_BARTYPE T_RPAREN",
"bar_items :",
"bar_items : bar_items bar_opt",
"bar_opt : padspec",
"bar_opt : T_EQUAL loc_variable",
"bar_opt : ending_info",
"bar_opt : T_HIDECHANGES",
"bar_opt : T_MNUM T_EQUAL num",
"bar_opt : T_LET T_EQUAL string",
"bar_opt : T_RNUM T_EQUAL num",
"bar_opt : T_REHEARSAL chk_dup_reh font_family title_font opt_size rehearsal_mark opt_dist",
"ending_info : T_ENDING string",
"ending_info : T_ENDENDING",
"chk_dup_reh :",
"rehearsal_mark : string",
"rehearsal_mark : T_MNUM",
"rehearsal_mark : T_RNUM",
"rehearsal_mark : T_LET",
"location : T_LPAREN alloc_loc xcoord T_COMMA ycoord T_RPAREN",
"alloc_loc :",
"xcoord : xterm xoffsetlist",
"xterm : xoffsetval",
"xterm : variable T_DOT east_or_west",
"xterm : xoffset",
"east_or_west : T_LETTER",
"east_or_west : T_LET_X",
"east_or_west : T_PITCH",
"xoffsetlist :",
"xoffsetlist : xoffsetlist xoffset",
"xoffset : plus_or_minus xoffsetval",
"plus_or_minus : T_PLUS",
"plus_or_minus : T_DASH",
"xoffsetval : floatnum",
"xoffsetval : T_TIME floatnum",
"ycoord : yterm yoffsetlist",
"yterm : yoffsetval",
"yterm : variable T_DOT north_or_south",
"north_or_south : T_LET_N",
"north_or_south : T_LET_S",
"north_or_south : T_LETTER",
"yoffsetlist :",
"yoffsetlist : yoffsetlist yoffset",
"yoffset : plus_or_minus yoffsetval",
"yoffsetval : floatnum",
"floatnum : num",
"floatnum : num T_DOT opt_decimal_part",
"floatnum : T_DOT numstr",
"opt_decimal_part :",
"opt_decimal_part : numstr",
"numstr : T_NUMBER",
"variable : loc_variable",
"line : linetype T_LINE line_alloc location T_TO location opt_line_str",
"line_alloc :",
"linetype :",
"linetype : T_LINETYPE",
"opt_line_str :",
"opt_line_str : T_WITH tfont title_size string",
"curve : linetype T_CURVE curve_alloc curve_loc T_TO curve_loc curve_tail",
"curve_alloc :",
"curve_tail : T_TO loc_list",
"curve_tail : T_BULGE bulgelist",
"loc_list : curve_loc",
"loc_list : loc_list T_TO curve_loc",
"bulgelist : bulge_value",
"bulgelist : bulgelist T_COMMA bulge_value",
"bulge_value : opt_minus floatnum",
"curve_loc : location",
"stuff_info : stuff_header T_COLON stufflist",
"stuff_header : stuff_id opt_size opt_modifier opt_place staff_list opt_dist",
"stuff_header : linetype T_PHRASE opt_size opt_place staff_list opt_dist",
"stuff_header : midi midi_staff",
"midi : T_MIDI",
"midi_staff : voice_id",
"midi_staff : T_ALL",
"stuff_id : stuff_type",
"stuff_type : font_family T_FONT",
"stuff_type : T_L_ANGLE",
"stuff_type : T_R_ANGLE",
"stuff_type : T_PEDAL",
"stuff_type : T_MUSSYM",
"stuff_type : T_OCTAVE",
"opt_size :",
"opt_size : T_LPAREN num T_RPAREN",
"opt_modifier :",
"opt_modifier : T_MODIFIER",
"opt_dist :",
"opt_dist : T_NUMVAR stuff_dist",
"stuff_dist : num opt_exclam",
"stuff_dist : T_DASH num opt_exclam",
"opt_exclam :",
"opt_exclam : T_EXCLAM",
"stufflist : stuff_item",
"stufflist : stufflist stuff_item",
"stuff_item : beat_offset steps_offset opt_string til_clause T_SEMICOLON",
"beat_offset : floatnum gracebackup",
"gracebackup :",
"gracebackup : T_LPAREN T_DASH num T_RPAREN",
"steps_offset :",
"steps_offset : T_LBRACKET plus_or_minus floatnum T_RBRACKET",
"opt_string :",
"opt_string : T_STAR",
"opt_string : string",
"til_clause :",
"til_clause : T_TIL til_val",
"til_val : floatnum til_suffix",
"til_suffix :",
"til_suffix : T_LET_M til_offset",
"til_offset :",
"til_offset : T_PLUS floatnum",
"roll_cmd : T_ROLL opt_down roll_extent T_COLON roll_offsetlist",
"opt_down :",
"opt_down : T_DOWN",
"roll_extent : roll_start roll_end",
"roll_start : staffnum",
"roll_start : staffnum voicenum",
"roll_end :",
"roll_end : T_TO staffnum",
"roll_end : T_TO staffnum voicenum",
"roll_offsetlist : floatnum T_SEMICOLON",
"roll_offsetlist : roll_offsetlist floatnum T_SEMICOLON",
"printcmd : printtype opt_loc string",
"printcmd : postscript opt_loc string",
"postscript : T_POSTSCRIPT",
"printtype : T_PRINTTYPE",
"opt_loc :",
"opt_loc : T_NL",
"opt_loc : location",
};
#endif
#ifdef YYSTACKSIZE
#undef YYMAXDEPTH
#define YYMAXDEPTH YYSTACKSIZE
#else
#ifdef YYMAXDEPTH
#define YYSTACKSIZE YYMAXDEPTH
#else
#define YYSTACKSIZE 500
#define YYMAXDEPTH 500
#endif
#endif
int yydebug;
int yynerrs;
int yyerrflag;
int yychar;
short *yyssp;
YYSTYPE *yyvsp;
YYSTYPE yyval;
YYSTYPE yylval;
short yyss[YYSTACKSIZE];
YYSTYPE yyvs[YYSTACKSIZE];
#define yystacksize YYSTACKSIZE
#line 4754 "gram.y"

#include <stdio.h>


/* print error message along with current filename and line number */

int
yyerror(msg)

char *msg;

{
	if (strcmp(msg, "syntax error") == 0) {
		/* beef up yacc's default syntax error message */
		l_yyerror(Curr_filename, yylineno,
				"error detected at '%s'",
				*yytext == '\n' ? "<newline>" : yytext);
	}
	else {
		l_yyerror(Curr_filename, yylineno, "%s", msg);
	}

	/* return something to keep lint happy */
	return(0);
}


/* if argument is surrounded by double quotes, remove them and return the
 * unquoted string. Actually just NULL out the end quote and return a
 * pointer to beyond the leading quote */

static char *
stripquotes(string)

char *string;

{
	char *p;

	p = string + strlen(string) - 1;
	if (*p == '"') {
		*p = '\0';
	}
	return( *string == '"' ? string + 1 : string);
}


/* make sure each header/footer type specified only once */

static void
chkdup_headfoot(flag, which)

int flag;	/* if non-zero, had gotten this head/foot before */
char *which;	/* header, footer, header2 or footer2 */

{
	if (flag != 0) {
		l_yyerror(Curr_filename, yylineno,
				"%s context specified more than once", which);
	}
}


/* when entering new context, need to save away everything about current one */

static void
end_prev_context()

{
	struct MAINLL *place_p;		/* where to insert SSV in main list */


	/* assume we will place at the end of main list. Later we'll set
	 * the right value if that assumption is wrong */
	place_p = Mainlltc_p;

	/* If current main list item hasn't been added to list yet,
	 * do that now. */
	if (Currstruct_p != (struct MAINLL *) 0 && Currstruct_p != place_p) {

		if (Currstruct_p->str == S_SSV) {

			if (List_of_staffs_p != (struct MAINLL *) 0) {
				yyerror("can't change context inside a measure");
				Currstruct_p = (struct MAINLL *) 0;
				return;
			}

			/* if is an SSV struct, we also need to set the
			 * correct values at this point, because the
			 * parser needs to know current values of some
			 * things (like font and point size) */
			asgnssv(Currstruct_p->u.ssv_p);

			/* if this SSV is after a FEED, we have to move it
			 * before the FEED in order to follow the rules of
			 * what comes when on the main list */
			for (   ; place_p != (struct MAINLL *) 0 &&
					place_p->str == S_FEED;
					place_p = place_p->prev) {
			}
		}

		insertMAINLL(Currstruct_p, place_p);
	}
	Curr_grpsyl_p = (struct GRPSYL *) 0;
	Currblock_p = (struct BLOCKHEAD *) 0;
	set_win_coord(0);
}


/* if defining more than one voice at the moment, not valid to set
 * a location variable. Print message and reset flag so we don't
 * give any more of the same message this measure */

static void
var_valid()

{
	if (Defining_multiple == YES) {
		yyerror("location tag is only allowed when defining a single voice");
		Defining_multiple = NO;
	}
	else if (Chord_at_a_time == YES) {
		yyerror("location tag not allowed with chord-at-a-time input");
	}
}


/* do a print command like left, right, or center. Allocate a PRINTDATA
 * struct and fill in all the information about what to print and where
 * to print it. */

static void
proc_printcmd(justifytype, inpc_p, str, font, size, got_nl, isPostScript, extra)

int justifytype;	/* J_LEFT, etc */
struct INPCOORD *inpc_p;	/* where to print */
char *str;		/* the string to print */
int font;
int size;
int got_nl;	/* YES if should go to next line before printing */
int isPostScript;	/* YES if is raw PostScript rather than normal print */
double extra;	/* how much extra vetical padding to add */

{
	struct PRINTDATA *curr_print_p;


	(void) contextcheck(C_MUSIC | C_BLOCKHEAD, "print");

	/* save all the info in a PRINTDATA struct */
	MALLOC(PRINTDATA, curr_print_p, 1);

	if (inpc_p == (struct INPCOORD *) 0) {

		/* print at default location */
		curr_print_p->location.vert_p = symval("_cur");

		switch (justifytype) {
		case J_LEFT:
		case J_RAGPARA:
		case J_JUSTPARA:
			/* default x is relative to _win.w */
			curr_print_p->location.hor_p = symval("_win");
			curr_print_p->location.htype = AW;
			break;
		case J_RIGHT:
			/* default x is relative to _win.e */
			curr_print_p->location.hor_p = symval("_win");
			curr_print_p->location.htype = AE;
			break;
		case J_CENTER:
			/* default x is relative to _win.x */
			curr_print_p->location.hor_p = symval("_win");
			curr_print_p->location.htype = AX;
			break;
		case J_NONE:
		default:
			/* default is at current location */
			curr_print_p->location.hor_p = symval("_cur");
			curr_print_p->location.htype = AX;
			break;
		}

		curr_print_p->location.hsteps = 0.0;
		curr_print_p->location.counts = 0.0;
		curr_print_p->location.vtype = AY;
		curr_print_p->location.vsteps = -extra;

		/* If this is the first print command in a block,
		 * we implicitly go down by the font ascent. */
		if (isPostScript == NO && Currblock_p != 0
					&& Next_print_link_p_p
					== &(Currblock_p->printdata_p) ) {
			curr_print_p->location.vsteps -= fontascent(font, size)
						/ STEPSIZE;
			got_nl = NO;
		}

	}
	else {
		curr_print_p->location = *(inpc_p);
		FREE(inpc_p);
	}

	curr_print_p->isPostScript = isPostScript;
	if (isPostScript == YES) {
		curr_print_p->width = 0.0;
	}
	else {
		curr_print_p->width = strwidth(str);
	}
	curr_print_p->justifytype = (short) justifytype;
	curr_print_p->string = str;
	curr_print_p->inputfile = Curr_filename;
	curr_print_p->inputlineno = (short) yylineno;

	/* special case of user asking that this be printed
	 * on the next line. In this case we go down by the
	 * current point size */
	if (got_nl) {
		curr_print_p->location.vsteps -=
					fontheight(font, size) / STEPSIZE;
	}

	/* Now link onto proper list -- could be in BLOCKHEAD
	 * context or doing a PRHEAD. If we are in BLOCKHEAD context, we
	 * definitely have something in the main list to attach to.
	 * If in C_MUSIC, we may or may not depending on whether
	 * this is the first in a series of print commands. */
	if ( ((Context & C_BLOCKHEAD) != 0)
		|| ((Currstruct_p != (struct MAINLL *) 0) &&
		(Currstruct_p->str == S_PRHEAD) ) ) {
			if (Next_print_link_p_p == (struct PRINTDATA **) 0) {
				pfatal("Next_print_link_p_p is null");
			}
			/* this means the value of Next_print_link_p_p
			 * is valid and will tell us where to link */
			*Next_print_link_p_p = curr_print_p;
	}
	else {
		/* We must be in music context, but the last
		 * thing we saw wasn't a print command, so
		 * we have to allocate a PRHEAD and put it
		 * in the main list and attach to that. */
		/* allocate a new struct, put on main list and attach
		 * the print command to it */
		Currstruct_p = newMAINLLstruct(S_PRHEAD, yylineno);
		insertMAINLL(Currstruct_p, Mainlltc_p);
		Currstruct_p->u.prhead_p->printdata_p = curr_print_p;
		Currstruct_p = (struct MAINLL *) 0;
	}

	/* in any case, if we get another print command right away, it
	 * should be linked onto this one, so remember where we are */
	Next_print_link_p_p = &(curr_print_p->next);
	curr_print_p->next = (struct PRINTDATA *) 0;
}


/* if the given string has a higher ascent than the given font/size, return
 * the amount is it higher, in stepsizes */

static double
extra_needed(font, size, string)

int font;	/* default font for this string */
int size;	/* default size for this string */
char *string;	/* string, which might contain \s(xx) to make it bigger
		 * than normal */

{
	double usual_ascent;
	double this_string_ascent;


	/* get expected ascent for this font/size and actual for string */
	usual_ascent = fontascent(font, size);
	(void) fix_string(string, font, size, Curr_filename, yylineno);
	this_string_ascent = strascent(string);

	/* if this string is too tall, return by how much it is too tall */
	if (this_string_ascent > usual_ascent) {
		return (this_string_ascent - usual_ascent) / STEPSIZE;
	}
	else {
		return(0.0);
	}
}


/* Sometimes lexer will recognize something as a keyword, but it's
 * really a bunch of note letters, due to context. This function turns
 * the mistaken keyword into notes. It handles a through g and r and s, but
 * not n and u which would also be troublesome.
 */

static void
keyword_notes(str)

char *str;

{
	for (  ; *str != '\0'; str++) {
		/* On the last letter, if it's a true pitch, have to
		 * push it back into input, because it could be followed
		 * by an accidental or octave mark or tick or something. */
		if (*(str+1) == '\0' && *str >= 'a' && *str <= 'g') {
			pushback(*str);
		}
		else {
			add_note(Curr_grpsyl_p, (int) *str, 0, USE_DFLT_OCTAVE,
							0, NO, (char *) 0);
		}
	}
}
#line 2146 "y.tab.c"
#define YYABORT goto yyabort
#define YYREJECT goto yyabort
#define YYACCEPT goto yyaccept
int
yyparse()
{
    register int yym, yyn, yystate;
#if YYDEBUG
    register char *yys;
    extern char *getenv();

    if (yys = getenv("YYDEBUG"))
    {
        yyn = *yys;
        if (yyn >= '0' && yyn <= '9')
            yydebug = yyn - '0';
    }
#endif

    yynerrs = 0;
    yyerrflag = 0;
    yychar = (-1);

    yyssp = yyss;
    yyvsp = yyvs;
    *yyssp = yystate = 0;

yyloop:
    if (yyn = yydefred[yystate]) goto yyreduce;
    if (yychar < 0)
    {
        if ((yychar = yylex()) < 0) yychar = 0;
#if YYDEBUG
        if (yydebug)
        {
            yys = 0;
            if (yychar <= YYMAXTOKEN) yys = yyname[yychar];
            if (!yys) yys = "illegal-symbol";
            fprintf(stderr, "%sdebug: state %d, reading %d (%s)\n",
                    YYPREFIX, yystate, yychar, yys);
        }
#endif
    }
    if ((yyn = yysindex[yystate]) && (yyn += yychar) >= 0 &&
            yyn <= YYTABLESIZE && yycheck[yyn] == yychar)
    {
#if YYDEBUG
        if (yydebug)
            fprintf(stderr, "%sdebug: state %d, shifting to state %d\n",
                    YYPREFIX, yystate, yytable[yyn]);
#endif
        if (yyssp >= yyss + yystacksize - 1)
        {
            goto yyoverflow;
        }
        *++yyssp = yystate = yytable[yyn];
        *++yyvsp = yylval;
        yychar = (-1);
        if (yyerrflag > 0)  --yyerrflag;
        goto yyloop;
    }
    if ((yyn = yyrindex[yystate]) && (yyn += yychar) >= 0 &&
            yyn <= YYTABLESIZE && yycheck[yyn] == yychar)
    {
        yyn = yytable[yyn];
        goto yyreduce;
    }
    if (yyerrflag) goto yyinrecovery;
#ifdef lint
#endif
    yyerror("syntax error");
#ifdef lint
#endif
    ++yynerrs;
yyinrecovery:
    if (yyerrflag < 3)
    {
        yyerrflag = 3;
        for (;;)
        {
            if ((yyn = yysindex[*yyssp]) && (yyn += YYERRCODE) >= 0 &&
                    yyn <= YYTABLESIZE && yycheck[yyn] == YYERRCODE)
            {
#if YYDEBUG
                if (yydebug)
                    fprintf(stderr, "%sdebug: state %d, error recovery shifting\
 to state %d\n", YYPREFIX, *yyssp, yytable[yyn]);
#endif
                if (yyssp >= yyss + yystacksize - 1)
                {
                    goto yyoverflow;
                }
                *++yyssp = yystate = yytable[yyn];
                *++yyvsp = yylval;
                goto yyloop;
            }
            else
            {
#if YYDEBUG
                if (yydebug)
                    fprintf(stderr, "%sdebug: error recovery discarding state %d\n",
                            YYPREFIX, *yyssp);
#endif
                if (yyssp <= yyss) goto yyabort;
                --yyssp;
                --yyvsp;
            }
        }
    }
    else
    {
        if (yychar == 0) goto yyabort;
#if YYDEBUG
        if (yydebug)
        {
            yys = 0;
            if (yychar <= YYMAXTOKEN) yys = yyname[yychar];
            if (!yys) yys = "illegal-symbol";
            fprintf(stderr, "%sdebug: state %d, error recovery discards token %d (%s)\n",
                    YYPREFIX, yystate, yychar, yys);
        }
#endif
        yychar = (-1);
        goto yyloop;
    }
yyreduce:
#if YYDEBUG
    if (yydebug)
        fprintf(stderr, "%sdebug: state %d, reducing by rule %d (%s)\n",
                YYPREFIX, yystate, yyn, yyrule[yyn]);
#endif
    yym = yylen[yyn];
    yyval = yyvsp[1-yym];
    switch (yyn)
    {
case 7:
#line 202 "gram.y"
{
		(void) contextcheck(C_MUSIC, "music information");
		tssv_line();
		/* reset flags for next voice */
		Defining_multiple = NO;
		Curr_grpsyl_p = (struct GRPSYL *) 0;
		Good_till_canceled = NO;
	}
break;
case 11:
#line 222 "gram.y"
{
		/* for blank lines -- nothing to do */
	}
break;
case 12:
#line 228 "gram.y"
{
		/* to keep going after stumbling on an input error */
		Getting_tup_dur = NO;
		Good_till_canceled = NO;
		Defining_multiple = NO;
		Curr_grpsyl_p = (struct GRPSYL *) 0;
	}
break;
case 13:
#line 237 "gram.y"
{
		/* fill in context field of the SSV struct */
		Currstruct_p->u.ssv_p->context = (short) Context;
	}
break;
case 14:
#line 244 "gram.y"
{
		Curr_family = Score.fontfamily;
	}
break;
case 15:
#line 250 "gram.y"
{}
break;
case 16:
#line 254 "gram.y"
{}
break;
case 17:
#line 258 "gram.y"
{
		/* have lots of separate structs in music context, so
		 * mark that we have no current struct */
		Currstruct_p = (struct MAINLL *) 0;

		/* make sure at least 1 staff is visible. Not good to
		 * check when user sets visible, because it would be
		 * nice to first set everything invisible and later
		 * turn one on. So if we wait till now to check,
		 * user doesn't have to switch back and forth from
		 * score to staff to get 1 voice visible. */
		check_at_least1visible();
	}
break;
case 18:
#line 273 "gram.y"
{
		Context = C_SCORE;
	}
break;
case 19:
#line 279 "gram.y"
{
		(void) rangecheck(yyvsp[-1].intval, MINSTAFFS, Score.staffs, "staff number");
		Context = C_STAFF;
		Currstruct_p->u.ssv_p->staffno = yyvsp[-1].intval;
	}
break;
case 20:
#line 287 "gram.y"
{
		(void) rangecheck(yyvsp[-2].intval, MINSTAFFS, Score.staffs, "staff number");
		(void) rangecheck(yyvsp[-1].intval, MINVOICES, MAXVOICES, "voice number");
		Context = C_VOICE;
		Currstruct_p->u.ssv_p->staffno = yyvsp[-2].intval;
		Currstruct_p->u.ssv_p->voiceno = yyvsp[-1].intval;
	}
break;
case 21:
#line 296 "gram.y"
{
		/* allocate an SSV struct */
		Currstruct_p = newMAINLLstruct(S_SSV, yylineno);
	}
break;
case 22:
#line 302 "gram.y"
{
		/* entering new context, if we have a current struct,
		 * we were in a different context, so add its info to the list */
		end_prev_context();
	}
break;
case 23:
#line 309 "gram.y"
{
		struct FEED *feed_p;		/* where to attach blocks */
		int calloced;			/* YES if dynamically alloced */

		end_prev_context();
		Currstruct_p = (struct MAINLL *) 0;
		Context = yyvsp[0].intval;
		calloced = NO;

		switch (yyvsp[0].intval) {
		case C_HEADER:
			/* We only allow one of these */
			chkdup_headfoot(Gotheader++, "header");
			Currblock_p = &Header;
			break;
		case C_HEAD2:
			chkdup_headfoot(Gothead2++, "header2");
			Currblock_p = &Header2;
			break;
		case C_FOOTER:
			chkdup_headfoot(Gotfooter++, "footer");
			Currblock_p = &Footer;
			break;
		case C_FOOT2:
			chkdup_headfoot(Gotfoot2++, "foot2");
			Currblock_p = &Footer2;
			break;
			break;
		case C_TOP:
		case C_TOP2:
		case C_BOT:
		case C_BOT2:
		case C_BLOCK:
			CALLOC(BLOCKHEAD, Currblock_p, 1);
			calloced = YES;
			break;
		default:
			pfatal("Unknown block-type context");
			/*NOTREACHED*/
			break;
		}
		set_win_coord(Currblock_p->c);
		/* Remember where to start attaching "print" commands */
		Next_print_link_p_p = &(Currblock_p->printdata_p);

		/* The dynamically allocated blocks
		 * (not the static head/foot blocks)
		 * need extra processing to populate
		 * the main list properly with FEEDs.
		 */
		if (calloced == YES) {
			if (Mainlltc_p == 0 || Mainlltc_p->str != S_FEED) {
				/* User had not specified a feed
				 * right before this block, so we need
				 * to add one for them implicitly.
				 * Create a new FEED and add it
				 * to the end of the main list.
				 */
				insertMAINLL(newMAINLLstruct(S_FEED, yylineno),
							Mainlltc_p);
			}
			feed_p = Mainlltc_p->u.feed_p;

			/* Normally, top/top2/bot/bot2 result in a pagefeed.
			 * But we have a special case for backward
			 * compatiblity and for flexibility.
			 * User can get a separate
			 * "title" page if they really want to,
			 * by using an explicit newpage at the beginning.
			 * But in the normal case, they'd want to declare
			 * top/bot that will go on the first page without
			 * that causing a newpage. So if there hasn't been
			 * any music or block yet, we won't set newpage=YES
			 * for top/top2/bot/bot2, but otherwise we will.
			 */
			if (Context & (C_TOP|C_TOP2|C_BOT|C_BOT2)) {
				struct MAINLL *m_p;
				for (m_p = Mainlltc_p->prev; m_p != 0;
							m_p = m_p->prev) {
					if (m_p->str == S_BAR ||
							m_p->str == S_BLOCKHEAD) {
						/* There was music or block,
						 * so this is not
						 * at very beginning,
						 * so no special case:
						 * there will be new page.
						 */
						feed_p->pagefeed = YES;
						break;
					}
				}
			}

			/* Now check for redefined blocks and
			 * attach the BLOCKHEAD onto the FEED or the
			 * main list, as appropriate.
			 */
			switch (Context) {
			case C_TOP:
				if (feed_p->top_p != 0) {
					warning("top redefined before being used");
				}
				feed_p->top_p = Currblock_p;
				break;
			case C_TOP2:
				if (feed_p->top2_p != 0) {
					warning("top2 redefined before being used");
				}
				feed_p->top2_p = Currblock_p;
				break;
			case C_BOT:
				if (feed_p->bot_p != 0) {
					warning("bot redefined before being used");
				}
				feed_p->bot_p = Currblock_p;
				break;
			case C_BOT2:
				if (feed_p->bot2_p != 0) {
					warning("bot2 redefined before being used");
				}
				feed_p->bot2_p = Currblock_p;
				break;
			case C_BLOCK:
				insertMAINLL(newMAINLLstruct(S_BLOCKHEAD, yylineno),
						Mainlltc_p);
				Mainlltc_p->u.blockhead_p = Currblock_p;
				/* These blocks count like "music" for purposes
				 * of things that can only be set
				 * before any "music"
				 */
				Got_some_data = YES;
				break;
			case C_HEADER:
			case C_HEAD2:
			case C_FOOTER:
			case C_FOOT2:
				/* These are static, not in main list,
				 * so nothing needs to be done with them.
				 */
				break;
			default:
				pfatal("unexpected context (0x%x) for block", Context);
				/*NOTREACHED*/
				break;
			}

			/* We need a feed after a block.
			 * Use -1 as lineno to mark it as internally generated.
			 * That way if user puts their own explicit feed
			 * next in the input, we can know we can
			 * discard this internally generated one.
			 */
			if (Context & C_BLOCK) {
				insertMAINLL(newMAINLLstruct(S_FEED, -1), Mainlltc_p);
			}
		}
	}
break;
case 24:
#line 468 "gram.y"
{
		Context = C_MUSIC;
	}
break;
case 25:
#line 473 "gram.y"
{
		end_prev_context();
		Currstruct_p = (struct MAINLL *) 0;
		Context = C_GRIDS;
	}
break;
case 26:
#line 481 "gram.y"
{
		if (Context == C_GRIDS) {
			add_grid(yyvsp[-2].stringval, yyvsp[-1].stringval);
		}
		else if (Context == C_HEADSHAPES) {
			/* skip past the font/size bytes */
			add_shape(yyvsp[-2].stringval + 2, yyvsp[-1].stringval + 2);
		}
		else {
			yyerror("unexpected strings");
		}
	}
break;
case 27:
#line 496 "gram.y"
{
		end_prev_context();
		Currstruct_p = (struct MAINLL *) 0;
		Context = C_HEADSHAPES;
	}
break;
case 29:
#line 507 "gram.y"
{
		yyval.intval = yyvsp[0].intval;
		/* later on we check for being within current range */
	}
break;
case 30:
#line 513 "gram.y"
{
		yyval.intval = atoi(yytext);
	}
break;
case 34:
#line 527 "gram.y"
{
		/* Only a few things are allowed to have a negative number.
		 * Other code assumes that that restriction is enforced here,
		 * so we only honor the minus if it is legal. */
		if (yyvsp[-1].intval == -1) {
			if (yyvsp[-3].intval != SCOREPAD && yyvsp[-3].intval != STAFFPAD
							&& yyvsp[-3].intval != SYLPOSITION) {
				yyerror("negative value not allowed");
			}
			else {
				assign_int(yyvsp[-3].intval, -(yyvsp[0].intval), Currstruct_p);
			}
		}
		else {
			assign_int(yyvsp[-3].intval, yyvsp[0].intval, Currstruct_p);
		}
	}
break;
case 35:
#line 547 "gram.y"
{
		/* only restcombine and gridfret can be set to empty */
		if (yyvsp[-1].intval == RESTCOMBINE) {
			assign_int(yyvsp[-1].intval, NORESTCOMBINE, Currstruct_p);
		}
		else if (yyvsp[-1].intval == GRIDFRET) {
			assign_int(yyvsp[-1].intval, NOGRIDFRET, Currstruct_p);
		}
		else {
			yyerror("parameter value required");
		}
	}
break;
case 36:
#line 562 "gram.y"
{
		assign_float(yyvsp[-2].intval, yyvsp[0].floatval, Currstruct_p);
	}
break;
case 37:
#line 568 "gram.y"
{
		assign_2floats(yyvsp[-4].intval, yyvsp[-2].floatval, yyvsp[0].floatval, Currstruct_p);
	}
break;
case 38:
#line 574 "gram.y"
{
		/* specified in stepsizes, stored in inches */
		assign_float(PAD, yyvsp[-1].intval * yyvsp[0].floatval * STEPSIZE, Currstruct_p);
	}
break;
case 39:
#line 581 "gram.y"
{
		assign_vscheme (yyvsp[-1].intval, yyvsp[0].intval, Currstruct_p);
	}
break;
case 40:
#line 587 "gram.y"
{
		set_beamlist(Currstruct_p);
		if (Currstruct_p != 0 && Currstruct_p->u.ssv_p != 0) {
			Currstruct_p->u.ssv_p->beamrests = yyvsp[0].intval;
		}
	}
break;
case 41:
#line 596 "gram.y"
{
		set_staffset(yyvsp[-3].intval, Currstruct_p);
	}
break;
case 42:
#line 602 "gram.y"
{
		set_barstlist(Currstruct_p);
	}
break;
case 43:
#line 608 "gram.y"
{
		if (Currstruct_p != 0) {
			chk_order(Currstruct_p->u.ssv_p, Order_place);
		}
	}
break;
case 45:
#line 619 "gram.y"
{
		/* all the work for this is done in timesig */
	}
break;
case 46:
#line 625 "gram.y"
{
	}
break;
case 47:
#line 630 "gram.y"
{
		set_font(yyvsp[-2].intval, yyvsp[0].intval, Currstruct_p);
	}
break;
case 48:
#line 636 "gram.y"
{
		set_font(yyvsp[-2].intval, yyvsp[0].intval, Currstruct_p);
	}
break;
case 49:
#line 642 "gram.y"
{
		if (contextcheck(C_SCORE|C_STAFF,"clef parameter") == YES) {
			used_check(Currstruct_p, CLEF, "clef");
			if (Context == C_STAFF &&
					is_tab_staff(Currstruct_p->u.ssv_p->staffno)) {
				yyerror("can't set clef on tab staff");
			}
			Currstruct_p->u.ssv_p->clef = yyvsp[0].intval;
			Currstruct_p->u.ssv_p->used[CLEF] = YES;
			asgnssv(Currstruct_p->u.ssv_p);
		}
	}
break;
case 50:
#line 657 "gram.y"
{
		assign_int(yyvsp[-2].intval, yyvsp[0].intval, Currstruct_p);
	}
break;
case 52:
#line 666 "gram.y"
{
		if (contextcheck(C_STAFF|C_VOICE, "unset") == YES) {
			if (yyvsp[0].intval < 0 || yyvsp[0].intval >= NUMFLDS) {
				pfatal("invalid parameter index %d", yyvsp[0].intval);
			}
			if (Currstruct_p->u.ssv_p->used[yyvsp[0].intval] == YES) {
				warning("unsetting parameter overrides previous setting");
			}
			/* Note that if user tries to unset something that
			 * can't be set in the current context, we silently
			 * ignore that. We figure that since it's already unset,
			 * it doesn't hurt anything to unset it again,
			 * and it's easier to not bother to check.
			 */

			Currstruct_p->u.ssv_p->used[yyvsp[0].intval] = UNSET;

			/* special case: scoresep controls two fields */
			if (yyvsp[0].intval == MINSCSEP) {
				Currstruct_p->u.ssv_p->used[MAXSCSEP] = UNSET;
			}

			/* Assign, just in case there are new interactions */
			asgnssv(Currstruct_p->u.ssv_p);
		}
	}
break;
case 53:
#line 695 "gram.y"
{
		if ( contextcheck(C_SSV, "timeunit parameter") == YES ) {
			used_check(Currstruct_p, TIMEUNIT, "timeunit");
			Currstruct_p->u.ssv_p->timeunit = yyvsp[0].ratval;
			Currstruct_p->u.ssv_p->used[TIMEUNIT] = YES;
		}
	}
break;
case 54:
#line 705 "gram.y"
{
		if ( contextcheck(C_SSV, "swingunit parameter") == YES ) {
			used_check(Currstruct_p, SWINGUNIT, "swingunit");
			Currstruct_p->u.ssv_p->swingunit = yyvsp[0].ratval;
			Currstruct_p->u.ssv_p->used[SWINGUNIT] = YES;
		}
	}
break;
case 55:
#line 715 "gram.y"
{
	}
break;
case 56:
#line 720 "gram.y"
{
		if ( contextcheck(C_SCORE, "endingstyle parameter") == YES ) {
			used_check(Currstruct_p, ENDINGSTYLE, "endingstyle");
			Currstruct_p->u.ssv_p->endingstyle = yyvsp[0].intval;
			Currstruct_p->u.ssv_p->used[ENDINGSTYLE] = YES;
		}
	}
break;
case 57:
#line 730 "gram.y"
{
		if (contextcheck(C_SCORE | C_STAFF, "rehstyle parameter") == YES ) {
			used_check(Currstruct_p, REHSTYLE, "rehstyle");
			Currstruct_p->u.ssv_p->rehstyle = yyvsp[0].intval;
			Currstruct_p->u.ssv_p->used[REHSTYLE] = YES;
		}
		
	}
break;
case 58:
#line 741 "gram.y"
{
		if ( contextcheck(C_SCORE | C_STAFF, "pedstyle parameter") == YES ) {
			used_check(Currstruct_p, PEDSTYLE, "pedstyle");
			Currstruct_p->u.ssv_p->pedstyle = yyvsp[0].intval;
			Currstruct_p->u.ssv_p->used[PEDSTYLE] = YES;
		}
	}
break;
case 59:
#line 751 "gram.y"
{
		char *trans_name;
		trans_name = (yyvsp[-4].intval == TRANSPOSITION ?
			"transpose parameter" : "addtranspose parameter");
		if (contextcheck(C_SCORE|C_STAFF, trans_name) == YES) {
			used_check(Currstruct_p, yyvsp[-4].intval, trans_name);
			chk_interval(yyvsp[-1].intval, yyvsp[0].intval);
			if (yyvsp[-4].intval == TRANSPOSITION) {
				Currstruct_p->u.ssv_p->inttype = yyvsp[-1].intval;
				Currstruct_p->u.ssv_p->intnum = yyvsp[-2].intval * yyvsp[0].intval;
			}
			else {
				Currstruct_p->u.ssv_p->addinttype = yyvsp[-1].intval;
				Currstruct_p->u.ssv_p->addintnum = yyvsp[-2].intval * yyvsp[0].intval;
			}
			Currstruct_p->u.ssv_p->used[yyvsp[-4].intval] = YES;
			if (Context == C_STAFF &&
					is_tab_staff(Currstruct_p->u.ssv_p->staffno)
					== YES) {
				l_yyerror(Curr_filename, yylineno,
					"%s not allowed on tablature staff",
					trans_name);
			}
			else {
				asgnssv(Currstruct_p->u.ssv_p);
			}
		}
	}
break;
case 60:
#line 782 "gram.y"
{
		assign_unit(yyvsp[0].intval, Currstruct_p);
	}
break;
case 61:
#line 788 "gram.y"
{
		assign_string(yyvsp[-2].intval, yyvsp[0].stringval, Currstruct_p);
	}
break;
case 62:
#line 794 "gram.y"
{
		/* whereused is not allowed on voice, just score and staff */
		if (contextcheck(C_SCORE|C_STAFF, "visible=whereused") == YES) {
			assign_int(VISIBLE, YES, Currstruct_p);
			Currstruct_p->u.ssv_p->hidesilent = YES;
		}
	}
break;
case 63:
#line 804 "gram.y"
{
		/* only valid letter is y */
		if (*yytext != 'y') {
			yyerror("visible value must be y, n, or whereused");
		}
		else if (contextcheck(C_SSV, "visible parameter")
								== YES) {
			assign_int(VISIBLE, YES, Currstruct_p);
			Currstruct_p->u.ssv_p->hidesilent = NO;
		}
	}
break;
case 64:
#line 818 "gram.y"
{
		if (contextcheck(C_SSV, "visible parameter") == YES) {
			assign_int(VISIBLE, NO, Currstruct_p);
			Currstruct_p->u.ssv_p->hidesilent = NO;
		}
	}
break;
case 65:
#line 826 "gram.y"
{
		/* none specified */
		yyval.intval = V_1;
	}
break;
case 66:
#line 833 "gram.y"
{
		/* only valid value here is 'f'  */
		if (*yytext == 'f') {
			yyval.intval = V_2FREESTEM;
		}
		else {
			yyerror("invalid voice qualifier: must be f or o");
			yyval.intval = V_1;
		}
	}
break;
case 67:
#line 846 "gram.y"
{
		/* only valid value here is 'o' */
		if (*yytext == 'o') {
			yyval.intval = V_2OPSTEM;
		}
		else {
			yyerror("invalid voice qualifier: must be f or o");
			yyval.intval = (int) V_1;
		}
	}
break;
case 68:
#line 858 "gram.y"
{
		/* NULL token for allocating a new beamlist */
		new_beamlist();
	}
break;
case 71:
#line 872 "gram.y"
{
		add_beamlist(yyvsp[0].ratval);
	}
break;
case 73:
#line 881 "gram.y"
{
		add_beamlist(yyvsp[0].ratval);
	}
break;
case 75:
#line 890 "gram.y"
{ begin_subbeam(); }
break;
case 76:
#line 891 "gram.y"
{
		end_subbeam();
	}
break;
case 77:
#line 896 "gram.y"
{
		/* actual time value is (2 * basic_time) - (basic_time x 1/2 to the (dots) power) */
		yyval.ratval = rsub (rmul(Two, yyvsp[-2].ratval), rmul(yyvsp[-2].ratval, rrai(One_half, yyvsp[-1].intval)));

		/* if filling in a GRPSYL struct, need to fill in dots.
		 * Could also be here due to beamstyle
		 * or timeunit or location time offset,
		 * in which case the Curr_grpsyl_p will be NULL,
		 * or while getting tuplet duration. */
		if (Curr_grpsyl_p != (struct GRPSYL *) 0
						&& Getting_tup_dur == NO) {
			Curr_grpsyl_p->dots = yyvsp[-1].intval;
		}
	}
break;
case 78:
#line 913 "gram.y"
{
		if (Curr_grpsyl_p != (struct GRPSYL *) 0
					&& Getting_tup_dur == NO) {
			yyval.ratval.n = (svpath(1,TIME))->time.n;
			yyval.ratval.d = (svpath(1,TIME))->time.d;
			/* use whole note symbol as default */
			Curr_grpsyl_p->basictime = 1;
			Curr_grpsyl_p->is_meas = YES;
			User_meas_time = NO;
		}
		else {
			yyerror("'m' is not valid here");
		}
	}
break;
case 79:
#line 930 "gram.y"
{
		yyval.ratval = Zero;
	}
break;
case 81:
#line 940 "gram.y"
{
		yyval.intval = NO;
	}
break;
case 82:
#line 946 "gram.y"
{
		yyval.intval = YES;
	}
break;
case 83:
#line 951 "gram.y"
{
		/* null token for the purpose of allocating a new rangelist
		 * to be filled in with user data */
		new_staffset();
	}
break;
case 87:
#line 968 "gram.y"
{
		/* save information about this range in the list */
		/* if only a single number, not a range, rangetail will be
		 * 0, so use num as both beginning and end of range */
		add_staffset( yyvsp[-2].intval, (yyvsp[-1].intval > 0 ? yyvsp[-1].intval : yyvsp[-2].intval), String1, String2);
	}
break;
case 88:
#line 976 "gram.y"
{
		/* no end of range, so return 0 as flag that there was only 1 number */
		yyval.intval = 0;
	}
break;
case 89:
#line 983 "gram.y"
{
		/* make sure isn't 0, because that it illegal and would be used
		 * later as though there were no range. Other error checking
		 * on this number will be done later. */
		if (yyvsp[0].intval == 0) {
			yyerror("second number of range may not be zero");
		}
		yyval.intval = yyvsp[0].intval;
	}
break;
case 90:
#line 994 "gram.y"
{
		/* no labels */
		String1 = String2 = (char *) 0;
	}
break;
case 91:
#line 1001 "gram.y"
{
		/* save label */
		String1 = yyvsp[-2].stringval;
	}
break;
case 92:
#line 1007 "gram.y"
{
		/* no second label */
		String2 = (char *) 0;
	}
break;
case 93:
#line 1014 "gram.y"
{
		/* save label2 for later use */
		String2 = yyvsp[0].stringval;
	}
break;
case 95:
#line 1023 "gram.y"
{
		/* NULL token for allocating space for a bar style list */
		new_barstlist();
	}
break;
case 99:
#line 1040 "gram.y"
{
		/* if only 1 number, not range, rangetail will be 0, so use
		 * same number for beginning and end */
		add_barst(yyvsp[-1].intval, (yyvsp[0].intval == 0 ? yyvsp[-1].intval : yyvsp[0].intval) );
	}
break;
case 100:
#line 1047 "gram.y"
{
		Order_place = yyvsp[0].intval;
		Order_prio = 1;
		if (Currstruct_p != 0) {
			switch (yyvsp[0].intval) {
			case PL_ABOVE:
				Currstruct_p->u.ssv_p->used[ABOVEORDER] = YES;
				break;
			case PL_BELOW:
				Currstruct_p->u.ssv_p->used[BELOWORDER] = YES;
				break;
			case PL_BETWEEN:
				Currstruct_p->u.ssv_p->used[BETWEENORDER] = YES;
				break;
			}
		}
	}
break;
case 101:
#line 1067 "gram.y"
{
		if (Currstruct_p != 0) {
			Currstruct_p->u.ssv_p->markorder[Order_place][yyvsp[0].intval] = Order_prio;
		}
	}
break;
case 102:
#line 1075 "gram.y"
{
		if (Currstruct_p != 0) {
			Currstruct_p->u.ssv_p->markorder[Order_place][yyvsp[0].intval] = ++Order_prio;
		}
	}
break;
case 103:
#line 1083 "gram.y"
{
		if (Currstruct_p != 0) {
			Currstruct_p->u.ssv_p->markorder[Order_place][yyvsp[0].intval] = Order_prio;
		}
	}
break;
case 104:
#line 1090 "gram.y"
{
		if (Currstruct_p != 0) {
			if (Currstruct_p->u.ssv_p->markorder[Order_place][yyvsp[0].intval] != 0) {
				l_yyerror(Curr_filename, yylineno,
					"order item %s specified more than once", yyvsp[0].intval);
			}
		}
		yyval.intval = yyvsp[0].intval;
	}
break;
case 105:
#line 1102 "gram.y"
{
		yyval.intval = MK_MUSSYM;
	}
break;
case 106:
#line 1108 "gram.y"
{
		yyval.intval = MK_OCTAVE;
	}
break;
case 107:
#line 1114 "gram.y"
{
		yyval.intval = MK_OTHERTEXT;
	}
break;
case 108:
#line 1120 "gram.y"
{
		switch (yyvsp[0].intval) {
		case TM_CHORD:
			yyval.intval = MK_CHORD;
			break;
		case TM_DYN:
			yyval.intval = MK_DYN;
			break;
		default:
			yyerror("invalid text modifier in order list");
			/* set to something valid, since $$ will be used
			 * as an array subscript. */
			yyval.intval = MK_CHORD;
			break;
		}
	}
break;
case 109:
#line 1139 "gram.y"
{
		yyval.intval = MK_LYRICS;
	}
break;
case 110:
#line 1145 "gram.y"
{
		yyval.intval = MK_ENDING;
	}
break;
case 111:
#line 1151 "gram.y"
{
		yyval.intval = MK_REHEARSAL;
	}
break;
case 112:
#line 1157 "gram.y"
{
		yyval.intval = MK_PEDAL;
	}
break;
case 113:
#line 1162 "gram.y"
{
		/* lexer will recognize 1/2 and 1/4 as dbl and quad whole,
		 * but in this particular context,
		 * they are really a time signature */
		assign_timesig(1, yyvsp[-1].intval, (yyvsp[0].intval ? TS_INVISNUM : TS_NUMERIC), Currstruct_p);
	}
break;
case 114:
#line 1171 "gram.y"
{
		assign_timesig( yyvsp[-3].intval, yyvsp[-1].intval, (yyvsp[0].intval ? TS_INVISNUM : TS_NUMERIC), Currstruct_p);
	}
break;
case 115:
#line 1177 "gram.y"
{
		/* cut time is 2/2, but with special symbol */
		assign_timesig( 2, 2, TS_CUT, Currstruct_p);
	}
break;
case 116:
#line 1184 "gram.y"
{
		/* common time is 4/4, but with special symbol */
		assign_timesig( 4, 4, TS_COMMON, Currstruct_p);
	}
break;
case 117:
#line 1190 "gram.y"
{
		yyval.intval = NO;
	}
break;
case 118:
#line 1196 "gram.y"
{
		/* 'n' is used in various places to mean do NOT
		 * print something that would normally be printed,
		 * so YES means we do want to inhibit the normal printing. */
		yyval.intval = YES;
	}
break;
case 119:
#line 1204 "gram.y"
{
		if ( (yyvsp[-1].intval != '#') && (yyvsp[-1].intval != '&') ) {
			yyerror("key signature must include # or &");
		}
		else {
			assign_key(yyvsp[-2].intval, yyvsp[-1].intval, yyvsp[0].intval, Currstruct_p);
		}
	}
break;
case 120:
#line 1215 "gram.y"
{
		int sharps;

		/* Get number of sharps for the pitch letter */
		sharps = strchr(Circle, yyvsp[-2].intval) - Circle - 1;

		/* adjust for sharp/flat */
		if (yyvsp[-1].intval == '#') {
			sharps += 7;
		}
		else if (yyvsp[-1].intval == '&') {
			sharps -= 7;
		}
		else if (yyvsp[-1].intval != 0) {
			yyerror("key letter can only be followed by # or &");
			/* set to some ridiculous value, so will fail below */
			sharps = 999;
		}

		/* Adjust if needed for minor */
		if (yyvsp[0].intval == YES) {
			sharps -= 3;
		}

		/* disallow illegal keys, like f& */
		if (abs(sharps) > 7) {
			/* print error unless already printed one above */
			if (sharps < 100) {
				yyerror("invalid key--too many sharps or flats");
			}
		}
		else {
			assign_key(abs(sharps), (sharps >= 0 ? '#' : '&'),
							yyvsp[0].intval, Currstruct_p);
		}
	}
break;
case 121:
#line 1254 "gram.y"
{
		asgn_stafflines(yyvsp[-1].intval, yyvsp[0].intval == YES ? SS_NOTHING : SS_NORMAL, Currstruct_p);
	}
break;
case 122:
#line 1260 "gram.y"
{
		asgn_stafflines(yyvsp[-1].intval, SS_DRUM, Currstruct_p);
	}
break;
case 123:
#line 1266 "gram.y"
{
		/* use default tablature */
		if (Currstruct_p != 0) {
			Currstruct_p->u.ssv_p->strinfo = Guitar;
			asgn_stafflines(DEFTABLINES, NO, Currstruct_p);
		}
	}
break;
case 124:
#line 1276 "gram.y"
{
		/* user-specified tablature */
		asgn_stafflines(yyvsp[-1].intval, NO, Currstruct_p);
	}
break;
case 125:
#line 1283 "gram.y"
{
		yyval.intval = 1;
	}
break;
case 126:
#line 1289 "gram.y"
{
		/* count up the number of strings */
		yyval.intval = yyvsp[-1].intval + 1;
	}
break;
case 127:
#line 1295 "gram.y"
{
		/* save info about this tablature string */
		if (Currstruct_p != 0) {
			add_tab_string_info(yyvsp[-3].intval, yyvsp[-2].intval, yyvsp[-1].intval, yyvsp[0].intval, Currstruct_p->u.ssv_p);
		}
	}
break;
case 128:
#line 1303 "gram.y"
{
		yyval.intval = 0;
	}
break;
case 129:
#line 1309 "gram.y"
{
		if (yyvsp[0].intval != '&' && yyvsp[0].intval != '#') {
			yyerror("accidental on string can only be & or #");
		}
	}
break;
case 130:
#line 1316 "gram.y"
{
		/* no ticks */
		yyval.intval = 0;
	}
break;
case 131:
#line 1323 "gram.y"
{
		/* limit number of ticks. Has to match the number of bits
		 * used by TMP_NTICKS in grpsyl.c */
		if (yyvsp[0].intval > MAXTICKS) {
			l_yyerror(Curr_filename, yylineno,
				"too many ' marks (%d max)", MAXTICKS);
			yyvsp[0].intval = MAXTICKS;
		}
	}
break;
case 132:
#line 1334 "gram.y"
{
		/* only valid letter is y */
		if (*yytext != 'y') {
			yyerror("flag value must be y or n");
		}
		yyval.intval = YES;
	}
break;
case 133:
#line 1344 "gram.y"
{
		yyval.intval = NO;
	}
break;
case 134:
#line 1349 "gram.y"
{
		/* assign the value of minimum vertical space between scores. */
		assign_int(MINSCSEP, yyvsp[0].intval, Currstruct_p);
	}
break;
case 135:
#line 1355 "gram.y"
{
		if (Currstruct_p == 0) {
			yyerror("trying to set scoresep in wrong context");
		}

		else {
			/* if user set minimum greater than default maximum,
			 * adjust maximum to equal the minimum
			 * they specified. */
			if (Currstruct_p->u.ssv_p->minscsep
					> Currstruct_p->u.ssv_p->maxscsep) {
				Currstruct_p->u.ssv_p->maxscsep
					= Currstruct_p->u.ssv_p->minscsep;
			}
			else {
				/* fill in default */
				assign_int(MAXSCSEP, DEFMAXSCSEP, Currstruct_p);
			}
		}
	}
break;
case 136:
#line 1378 "gram.y"
{
		if (Currstruct_p != 0) {
			if (Currstruct_p->u.ssv_p->minscsep > yyvsp[0].intval) {
				yyerror("maximum smaller than minimum");
			}
			assign_int(MAXSCSEP, yyvsp[0].intval, Currstruct_p);
		}
	}
break;
case 137:
#line 1389 "gram.y"
{
		yyval.intval = yyvsp[0].intval;
	}
break;
case 138:
#line 1395 "gram.y"
{
		if (yyvsp[0].intval != C_TOP) {
			l_yyerror(Curr_filename, yylineno,
					"%s not valid here", yytext);
		}
		yyval.intval = ENDING_TOP;
	}
break;
case 139:
#line 1404 "gram.y"
{
		yyval.intval = 1;
	}
break;
case 140:
#line 1410 "gram.y"
{
		yyval.intval = -1;
	}
break;
case 141:
#line 1415 "gram.y"
{
		yyval.intval = NO;
	}
break;
case 142:
#line 1421 "gram.y"
{
		switch (yyvsp[0].intval) {
		case MINOR:
			yyval.intval = YES;
			break;
		case MAJOR:
			yyval.intval = NO;
			break;
		default:
			yyval.intval = NO;
			yyerror("must be major or minor");
			break;
		}
	}
break;
case 143:
#line 1437 "gram.y"
{
		yyval.intval = P_LINE;
	}
break;
case 144:
#line 1443 "gram.y"
{
		yyval.intval = P_ALTPEDSTAR;
	}
break;
case 145:
#line 1449 "gram.y"
{
		yyval.intval = P_PEDSTAR;
	}
break;
case 147:
#line 1457 "gram.y"
{
	}
break;
case 148:
#line 1461 "gram.y"
{
		if (yyvsp[0].intval == FONT_UNKNOWN) {
			yyvsp[0].intval = Curr_font;
		}
		if (yyvsp[-1].intval == FAMILY_DFLT) {
			/* user didn't specify a family. Use the current
			 * family, unless we don't have one yet, in which
			 * case look it up */
			if (Curr_family == FAMILY_DFLT) {
				Curr_family = Score.fontfamily;
			}
			Titlefont =  Curr_family + yyvsp[0].intval;
		}
		else {
			Titlefont = yyvsp[-1].intval + yyvsp[0].intval;
		}
	}
break;
case 149:
#line 1480 "gram.y"
{
		yyval.intval = FONT_UNKNOWN;
	}
break;
case 150:
#line 1486 "gram.y"
{
		yyval.intval = yyvsp[0].intval;
	}
break;
case 151:
#line 1491 "gram.y"
{
		yyval.intval = FAMILY_DFLT;
	}
break;
case 152:
#line 1496 "gram.y"
{
		yyval.intval = yyvsp[0].intval;
	}
break;
case 153:
#line 1501 "gram.y"
{
		Titlesize = Curr_size;
	}
break;
case 154:
#line 1507 "gram.y"
{
		Titlesize = yyvsp[-1].intval;
	}
break;
case 155:
#line 1513 "gram.y"
{
		Extra = extra_needed(Titlefont, Titlesize, yyvsp[0].stringval);
		proc_printcmd(J_CENTER, (struct INPCOORD *) 0, yyvsp[0].stringval,
				Titlefont, Titlesize, YES, NO, Extra);
	}
break;
case 156:
#line 1521 "gram.y"
{
		double tmp_extra;

		/* if one is taller than the default for the font/size,
		 * figure out how much more to add on */
		/* can't use MAX passing extra_needed() as arguments because
		 * it would get executed multiple times, which is a problem */
		Extra = extra_needed(Titlefont, Titlesize, yyvsp[-1].stringval);
		tmp_extra = extra_needed(Titlefont, Titlesize, yyvsp[0].stringval);
		Extra = MAX(Extra, tmp_extra);


		proc_printcmd(J_LEFT, (struct INPCOORD *) 0, yyvsp[-1].stringval,
				Titlefont, Titlesize, YES, NO, Extra);
		proc_printcmd(J_RIGHT, (struct INPCOORD *) 0, yyvsp[0].stringval,
				Titlefont, Titlesize, NO, NO, (double) 0.0);
	}
break;
case 157:
#line 1541 "gram.y"
{
		double tmp_extra;

		Extra = extra_needed(Titlefont, Titlesize, yyvsp[-2].stringval);
		tmp_extra = extra_needed(Titlefont, Titlesize, yyvsp[-1].stringval);
		Extra = MAX(Extra, tmp_extra);
		tmp_extra = extra_needed(Titlefont, Titlesize, yyvsp[0].stringval);
		Extra = MAX(Extra, tmp_extra);

		proc_printcmd(J_LEFT, (struct INPCOORD *) 0, yyvsp[-2].stringval,
				Titlefont, Titlesize, YES, NO, Extra);
		proc_printcmd(J_CENTER, (struct INPCOORD *) 0, yyvsp[-1].stringval,
				Titlefont, Titlesize, NO, NO, (double) 0.0);
		proc_printcmd(J_RIGHT, (struct INPCOORD *) 0, yyvsp[0].stringval,
				Titlefont, Titlesize, NO, NO, (double) 0.0);
	}
break;
case 158:
#line 1559 "gram.y"
{
		char *p;		/* pointer through the string */
		int backslash_count;	/* how many backslashes in a row */
		int font, size;		/* at end of \n-delimited segment */
		char *string_start;	/* where current segment begins */

		/* Convert unescaped input newlines to spaces.
		 * The +2 in the 'for' initialization is to skip font/size.
		 */
		backslash_count = 0;
		string_start = yyvsp[-2].stringval;
		/* If we are running on a file using \r or \r\n for
		 * line separator, normalize to \n instead */
		for (p = string_start + 2; *p != 0; p++) {
			if (*p == '\r') {
				if ( *(p+1) != '\n') {
					/* \r by itself. Use \n instead */
					*p = '\n';
				}
				else {
					/* delete the \r */
					char *src, *dest;
					for (src = p + 1, dest = p;  ; src++, dest++) { 
						*dest = *src;
						if (*src == '\0') {
							break;
						}
					}
				}
			}
		}

		font = Titlefont;
		size = Titlesize;
		for (p = string_start + 2; *p != 0; p++) {
			/* Real newlines not preceeded by an odd number
			 * of backslashes are converted to spaces. */
			if (*p == '\n' && (backslash_count & 0x1) == 0) {
				*p = ' ';
			}

			/* Count up number of consecutive backslashes.
			 * Also, if user explicitly uses \n, split into
			 * a new paragraph
			 */
			if (*p == '\\') {
				backslash_count++;
				if (((backslash_count & 0x1) == 1)
						&& ( *(p+1) == 'n') ) {
					/* Need to split.
					 * Terminate the current string
					 * where we are in it, and process it,
					 * then arrange to continue processing
					 * on the rest of the string.
					 */
					*p = '\0';
					(void) fix_string(string_start,
						font, size,
						Curr_filename, yylineno);
					proc_printcmd(yyvsp[-6].intval, (struct INPCOORD*) 0,
						string_start, font, size,
						YES, NO, 0.0);

					/* The user could change font/size
					 * in mid-string with \f() and/or \s().
					 * so we have to determine what they
					 * are at the end of the current segment
					 * and use that for making the copy
					 * of the remainder of the string.
					 * The original copy of the remainder
					 * of the string gets "memory leaked,"
					 * but this is a rare case, so not
					 * worth worrying about.
					 */
					end_fontsize(string_start, &font, &size);
					string_start = copy_string(p+2, font, size);
					p = string_start + 2;
				}
			}
			else {
				backslash_count = 0;
			}
		}
		(void) fix_string(string_start, font, size,
					Curr_filename, yylineno);
		proc_printcmd(yyvsp[-6].intval, (struct INPCOORD*) 0, string_start,
					font, size, YES, NO, 0.0);
	}
break;
case 159:
#line 1650 "gram.y"
{
		/* Use same paragraph time as last time (or default if
		 * this is the first paragraph ever in this song.) */
		yyval.intval = Curr_paratype;
	}
break;
case 160:
#line 1658 "gram.y"
{
		Curr_paratype = yyvsp[0].intval;
		yyval.intval = yyvsp[0].intval;
	}
break;
case 187:
#line 1743 "gram.y"
{
		/* strip the quotes from the string and make a copy for later use */
		if (Curr_family == FAMILY_DFLT) {
			Curr_family = Score.fontfamily;
		}
		yyval.stringval = copy_string(stripquotes(yytext), Curr_family + Curr_font,
								Curr_size);
	}
break;
case 188:
#line 1754 "gram.y"
{
		char *old_string;

		/* append new string part to existing part */
		old_string = yyvsp[-2].stringval;
		/* new string part has quotes, so subtract 2 from needed length,
		 * but need space for null, so overall need 1 less. */
		MALLOCA(char, yyval.stringval, strlen(old_string) + strlen(yytext) - 1);
		sprintf(yyval.stringval, "%s%s", old_string, stripquotes(yytext));
		FREE(old_string);
	}
break;
case 189:
#line 1767 "gram.y"
{
		/* Here's the deal: the staff/voice specification
		 * could be lists of staff ranges and/or voices.
		 * We need a separate copy of the groupinfo
		 * for each staff/voice. So as we parse the sv_spec,
		 * we remember which staffs and voices are specified,
		 * and keep a pointer to the first STAFF in the measure.
		 * Then after collecting the groupinfo,
		 * we make copies of the group info for each staff/voice given.
		 * It has to be copies rather than just a pointer to the
		 * same stuff, because the coordinates will be different
		 * for each staff/voice. */

		/* attach the groupinfo to each of the STAFF structs */
		link_groups();
	}
break;
case 190:
#line 1786 "gram.y"
{
		add_multirest(yyvsp[-1].intval);
		Currstruct_p = (struct MAINLL *) 0;
	}
break;
case 191:
#line 1793 "gram.y"
{
		free_staffrange();
	}
break;
case 192:
#line 1798 "gram.y"
{
		begin_range(PL_UNKNOWN);
	}
break;
case 193:
#line 1803 "gram.y"
{
		Chord_at_a_time = NO;
	}
break;
case 194:
#line 1809 "gram.y"
{
		Chord_at_a_time = YES;
	}
break;
case 198:
#line 1824 "gram.y"
{
		begin_map();
	}
break;
case 199:
#line 1829 "gram.y"
{
		end_map();
	}
break;
case 200:
#line 1834 "gram.y"
{
		map_item();
	}
break;
case 201:
#line 1840 "gram.y"
{
		map_item();
		Defining_multiple = YES;
	}
break;
case 202:
#line 1846 "gram.y"
{
		/* initialize grpsyl info */
		Curr_grpsyl_p = Last_grpsyl_p = (struct GRPSYL *) 0;
	}
break;
case 204:
#line 1856 "gram.y"
{
		/* prepare for a new set of ranges */
		begin_range(Place);
		Defining_multiple = YES;
	}
break;
case 205:
#line 1863 "gram.y"
{
		/* data is for staff with only one voice */
		save_vno_range(1, 1);
		add_to_sv_list();
	}
break;
case 206:
#line 1871 "gram.y"
{
		if (yyvsp[0].intval == YES) {
			add_to_sv_list();
		}
	}
break;
case 207:
#line 1880 "gram.y"
{
		/* set flag as to whether we are processing tab staff
		 * information or normal staff info */
		Doing_tab_staff = is_tab_range();
	}
break;
case 209:
#line 1890 "gram.y"
{
		/* several staffs have the same note data */
		Defining_multiple = YES;
	}
break;
case 210:
#line 1896 "gram.y"
{
		/* remember which staffs we are currently getting data for */
		/* if only a single number, not a range, staff_range will be
		 * 0, so use staffnum as high and low of range */
		save_staff_range(yyvsp[-1].intval, (yyvsp[0].intval == 0 ? yyvsp[-1].intval : yyvsp[0].intval));
	}
break;
case 211:
#line 1904 "gram.y"
{
		/* empty - no range, just a single number */
		yyval.intval = 0;
	}
break;
case 212:
#line 1911 "gram.y"
{
		Defining_multiple = YES;
		yyval.intval = yyvsp[0].intval;
		if (yyvsp[0].intval == 0) {
			yyerror("staff of 0 is illegal");
		}
	}
break;
case 213:
#line 1921 "gram.y"
{
		/* note data applies to a single voice */
		if (rangecheck(yyvsp[-1].intval, MINVOICES, MAXVOICES, "voice") == YES) {
			save_vno_range(yyvsp[-1].intval, yyvsp[-1].intval);
			yyval.intval = YES;
		}
		else {
			yyval.intval = NO;
		}
	}
break;
case 214:
#line 1934 "gram.y"
{
		/* note data applies to range of voices */
		if (rangecheck(yyvsp[-3].intval, MINVOICES, MAXVOICES, "voice") == YES &&
				rangecheck(yyvsp[-1].intval, MINVOICES, MAXVOICES, "voice") == YES) {
			if (yyvsp[-1].intval < yyvsp[-3].intval) {
				l_yyerror(Curr_filename, yylineno,
					"end of voice range is smaller than beginning");
				yyval.intval = NO;
			}
			else {
				save_vno_range(yyvsp[-3].intval, yyvsp[-1].intval);
				Defining_multiple = YES;
				yyval.intval = YES;
			}
		}
		else {
			yyval.intval = NO;
		}
	}
break;
case 216:
#line 1959 "gram.y"
{
		Defining_multiple = YES;
	}
break;
case 220:
#line 1974 "gram.y"
{
		/* Associate any TIMEDSSVs with this group */
		if (Curr_tssv_p != 0) {
			tssv_setgrpsyl(Curr_grpsyl_p);
			Curr_tssv_p = 0;
		}

		/* if no custom beaming indicated, fill in based on previous. */
		if ( Curr_grpsyl_p->beamloc == NOITEM) {
			setbeamloc(Curr_grpsyl_p, Last_grpsyl_p);
		}

		/* later we check that if there is an 'm', it is
		 * the only thing in the measure */

		link_notegroup(Curr_grpsyl_p, Last_grpsyl_p);
		Curr_marklist = 0;
	}
break;
case 223:
#line 2002 "gram.y"
{
		Curr_tssv_p = tssv_create(C_SCORE);
	}
break;
case 224:
#line 2008 "gram.y"
{
		Curr_tssv_p = tssv_create(C_STAFF);
	}
break;
case 225:
#line 2014 "gram.y"
{
		Curr_tssv_p = tssv_create(C_VOICE);
	}
break;
case 228:
#line 2028 "gram.y"
{
		tssv_update(Curr_tssv_p, yyvsp[-2].intval, yyvsp[0].intval);
	}
break;
case 229:
#line 2034 "gram.y"
{
		switch (yyvsp[-2].intval) {

		case DEFOCT:
		case RELEASE:
			tssv_update(Curr_tssv_p, yyvsp[-2].intval, yyvsp[0].intval);
			break;

		default:
			yyerror("parameter type cannot be changed mid-measure");
			break;
		}
	}
break;
case 231:
#line 2054 "gram.y"
{
		if (Good_till_canceled == NO) {
			/* use default attributes */
			Curr_grpsyl_p->grpvalue = GV_NORMAL;
			Curr_grpsyl_p->grpsize = GS_NORMAL;
			Curr_grpsyl_p->headshape = HS_UNKNOWN;
		}
		else {
			/* re-use previous style */
			copy_attributes(Curr_grpsyl_p, Last_grpsyl_p);
		}
	}
break;
case 232:
#line 2069 "gram.y"
{
		Good_till_canceled = NO;
	}
break;
case 233:
#line 2075 "gram.y"
{
	}
break;
case 234:
#line 2079 "gram.y"
{
		Good_till_canceled = NO;
	}
break;
case 235:
#line 2085 "gram.y"
{
		Good_till_canceled = YES;
	}
break;
case 236:
#line 2090 "gram.y"
{
	}
break;
case 237:
#line 2095 "gram.y"
{
		if (Good_till_canceled == YES) {
			yyerror("can't use ... and more than one [] on same chord");
		}
	}
break;
case 238:
#line 2102 "gram.y"
{
		/* NULL token to cause allocation of a GRPSYL struct */
		Last_grpsyl_p = Curr_grpsyl_p;
		Curr_grpsyl_p = newGRPSYL(GS_GROUP);
	}
break;
case 239:
#line 2109 "gram.y"
{
		/* an empty marklist means we should use
		 * the same attributes as the last time */
		copy_attributes(Curr_grpsyl_p, Last_grpsyl_p);
	}
break;
case 243:
#line 2126 "gram.y"
{
		Curr_grpsyl_p->nwith = (short) Item_count;
		Curr_grpsyl_p->withlist = Curr_marklist;
	}
break;
case 247:
#line 2142 "gram.y"
{
		if (Curr_grpsyl_p->stemdir != UNKNOWN &&
					Curr_grpsyl_p->stemdir != yyvsp[0].intval) {
			l_warning(Curr_filename, yylineno,
				"both stem directions specified; using last instance");
		}
		if (Doing_tab_staff == YES) {
			yyerror("stem direction specification not allowed on tab staff");
		}
		else {
			Curr_grpsyl_p->stemdir = yyvsp[0].intval;
		}
	}
break;
case 248:
#line 2158 "gram.y"
{
		if (yyvsp[-2].intval != DIST) {
			yyerror("unexpected parameter name");
		}
		else {
			if (Curr_grpsyl_p->restdist != NORESTDIST &&
					Curr_grpsyl_p->restdist != yyvsp[-1].intval * yyvsp[0].intval) {
				l_warning(Curr_filename, yylineno,
						"more than one dist value specified; using last instance");
			}
			Curr_grpsyl_p->restdist = yyvsp[-1].intval * yyvsp[0].intval;
		}
	}
break;
case 249:
#line 2174 "gram.y"
{
		if (Curr_grpsyl_p->ho_usage != HO_NONE &&
				Curr_grpsyl_p->ho_usage != yyvsp[0].intval) {
			l_warning(Curr_filename, yylineno,
				"More than one ho type specified; last instance used");
		}
		Curr_grpsyl_p->ho_usage = yyvsp[0].intval;
	}
break;
case 250:
#line 2185 "gram.y"
{
		if (Curr_grpsyl_p != 0) {
			if (Curr_grpsyl_p->headshape != HS_UNKNOWN) {
				l_warning(Curr_filename, yylineno,
					"multiple head shapes specified, using last");
			}
			/* +2 to skip past extraneous font/size */
			if (Doing_tab_staff == YES && strcmp(yyvsp[0].stringval + 2, "allx") != 0) {
				yyerror("allx is the only headshape allowed on tab staffs");
			}
			if ((Curr_grpsyl_p->headshape = get_shape_num(yyvsp[0].stringval + 2))
					== HS_UNKNOWN) {
				l_yyerror(Curr_filename, yylineno,
					"'%s' is not a valid head shape name",
					ascii_str(yyvsp[0].stringval, YES, NO, TM_NONE));
			}
		}
		else {
			pfatal("Curr_grpsyl_p was null for setting hs");
		}
		/* We don't need the name anymore; we have its corresponding
		 * number that we use internally. */
		FREE(yyvsp[0].stringval);
	}
break;
case 251:
#line 2212 "gram.y"
{
		if (Curr_grpsyl_p->stemlen != STEMLEN_UNKNOWN &&
				fabs(Curr_grpsyl_p->stemlen - yyvsp[0].floatval * STEPSIZE) > 0.0001) {
			l_warning(Curr_filename, yylineno,
				"more than one len specified; using last instance");
		}
		Curr_grpsyl_p->stemlen = yyvsp[0].floatval * STEPSIZE;
	}
break;
case 252:
#line 2223 "gram.y"
{
		if (yyvsp[0].intval <= 0) {
			yyerror("slash number must be > 0");
		}
		else if (yyvsp[0].intval > 8) {
			/* we decided 8 slashes would be at least 256th note
			 * or shorter, and 256th is the shortest note we
			 * support, 8 slashes is plenty */
			yyerror("slash number must be <= 8");
		}
		if (Curr_grpsyl_p->slash_alt > 0 &&
					Curr_grpsyl_p->slash_alt != yyvsp[0].intval) {
			l_warning(Curr_filename, yylineno,
				"more than one slash value specified; using last instance");
		}
		if (Curr_grpsyl_p->slash_alt < 0) {
			yyerror("only one slash/alt allowed per group");
		}
		Curr_grpsyl_p->slash_alt = yyvsp[0].intval;
	}
break;
case 253:
#line 2246 "gram.y"
{
		yyval.intval = HO_RIGHT;
	}
break;
case 254:
#line 2252 "gram.y"
{
		yyval.intval = HO_LEFT;
	}
break;
case 255:
#line 2258 "gram.y"
{
		/* We want to silently accept identical values specified
		 * more than once, so use fuzzy compare to deal with roundoff */
		if (Curr_grpsyl_p->ho_usage == HO_VALUE &&
				fabs(Curr_grpsyl_p->ho_value - yyvsp[-1].intval * yyvsp[0].floatval) > 0.0001) {
			l_warning(Curr_filename, yylineno,
				"More than one ho value specified; last instance used");
		}
		Curr_grpsyl_p->ho_value = yyvsp[-1].intval * yyvsp[0].floatval;
		yyval.intval = HO_VALUE;
	}
break;
case 256:
#line 2272 "gram.y"
{
		yyval.intval = 1;
	}
break;
case 257:
#line 2278 "gram.y"
{
		yyval.intval = yyvsp[0].intval;
	}
break;
case 263:
#line 2300 "gram.y"
{
		if (yyvsp[0].intval <= 0) {
			yyerror("alt number must be > 0");
		}
		if (Curr_grpsyl_p->slash_alt != 0) {
			yyerror("only one slash/alt allowed per group");
		}
		if (Doing_tab_staff == YES) {
			yyerror("alt not allowed on tablature staff");
		}
		/* Keep small enough that 1 shifted left by this still
		 * fits in a 16-bit int. Should be more than enough! */
		if (yyvsp[0].intval > 15) {
			warning("alt value too large");
			yyvsp[0].intval = 15;
		}
		Curr_grpsyl_p->slash_alt = -yyvsp[0].intval;
	}
break;
case 264:
#line 2321 "gram.y"
{
		int n;

		/* add a slur to each note in the chord. Don't know
		 * which pitch to slur to yet, just that it will be to
		 * the "matching" note in the next chord, so use
		 * special magic pitch of 'M' to convey this */
		for (n = 0; n < Curr_grpsyl_p->nnotes; n++) {
			add_slurto(Curr_grpsyl_p, 'M', USE_DFLT_OCTAVE, n, yyvsp[-1].intval);
			set_slurdir(Curr_grpsyl_p, -1, yyvsp[0].intval);
		}
	}
break;
case 265:
#line 2336 "gram.y"
{
		int n;

		Curr_grpsyl_p->tie = YES;
		for (n = 0; n < Curr_grpsyl_p->nnotes; n++) {
			Curr_grpsyl_p->notelist[n].tie = YES;
			Curr_grpsyl_p->notelist[n].tiestyle = yyvsp[-1].intval;
			Curr_grpsyl_p->notelist[n].tiedir = yyvsp[0].intval;
		}
	}
break;
case 266:
#line 2348 "gram.y"
{
		if (Curr_marklist == 0) {
			/* allocate space for a list of marks */
			MALLOCA(char *, Curr_marklist, ITEMS);
			Item_count = 0;
			Max_items = ITEMS;
		}
	}
break;
case 267:
#line 2358 "gram.y"
{
		yyval.intval = UNKNOWN;
	}
break;
case 268:
#line 2364 "gram.y"
{
		yyval.intval = yyvsp[0].intval;
	}
break;
case 271:
#line 2376 "gram.y"
{
		/* if too many items, get some more space */
		if (Item_count >= Max_items) {
			Max_items += ITEMS;
			if ((Curr_marklist = (char **) realloc(Curr_marklist,
						Max_items * sizeof(char **)))
						== (char **) 0) {
				l_no_mem(__FILE__, __LINE__);
			}
		}
		Curr_marklist[Item_count++] = yyvsp[0].stringval;
	}
break;
case 272:
#line 2390 "gram.y"
{
		MALLOCA(char, yyval.stringval, strlen(yyvsp[0].stringval) + 3);
		yyval.stringval[0] = FONT_TR;
		yyval.stringval[1] = DFLT_SIZE;
		sprintf(yyval.stringval + 2, yyvsp[0].stringval);
	}
break;
case 274:
#line 2403 "gram.y"
{
		yyval.stringval = (Curr_grpsyl_p->grpsize == GS_NORMAL
					? "\\(dot)" : "\\(smdot)");
	}
break;
case 275:
#line 2410 "gram.y"
{
		yyval.stringval = (Curr_grpsyl_p->grpsize == GS_NORMAL
					? "\\(leg)" : "\\(smleg)");
	}
break;
case 276:
#line 2417 "gram.y"
{
		yyval.stringval = (Curr_grpsyl_p->grpsize == GS_NORMAL
					? "\\(acc_gt)" : "\\(smacc_gt)");
	}
break;
case 277:
#line 2424 "gram.y"
{
		yyval.stringval = (Curr_grpsyl_p->grpsize == GS_NORMAL
					? "\\(acc_hat)" : "\\(smacc_hat)");
	}
break;
case 278:
#line 2430 "gram.y"
{
		if (Curr_grpsyl_p->grpvalue == GV_ZERO) {
			l_warning(Curr_filename, yylineno,
				"both grace and cue specified; using last instance");
		}
		Curr_grpsyl_p->grpvalue = GV_NORMAL;
		Curr_grpsyl_p->grpsize = GS_SMALL;
	}
break;
case 279:
#line 2441 "gram.y"
{
		if (Curr_grpsyl_p->grpsize == GS_SMALL
					&& Curr_grpsyl_p->grpvalue != GV_ZERO) {
			l_warning(Curr_filename, yylineno,
				"both cue and grace specified; using last instance");
		}

		Curr_grpsyl_p->grpsize = GS_SMALL;
		Curr_grpsyl_p->grpvalue = GV_ZERO;
	}
break;
case 280:
#line 2454 "gram.y"
{
		if (Curr_grpsyl_p->headshape != HS_UNKNOWN) {
			l_warning(Curr_filename, yylineno,
				"diam is overriding previous headshapes specification");
		}
		Curr_grpsyl_p->headshape = get_shape_num("diam");
	}
break;
case 281:
#line 2464 "gram.y"
{
		if (Curr_grpsyl_p->headshape != HS_UNKNOWN) {
			l_warning(Curr_filename, yylineno,
				"xnote is overriding previous headshapes specification");
		}
		Curr_grpsyl_p->headshape = get_shape_num(
					Doing_tab_staff == YES ? "allx" : "x");
	}
break;
case 282:
#line 2475 "gram.y"
{
		yyval.intval = UP;
	}
break;
case 283:
#line 2481 "gram.y"
{
		yyval.intval = DOWN;
	}
break;
case 284:
#line 2487 "gram.y"
{
		/* custom beaming cannot be nested */
		if ( (Last_grpsyl_p != (struct GRPSYL *) 0)
			&& ((Last_grpsyl_p->beamloc == STARTITEM) || (Last_grpsyl_p->beamloc == INITEM)) ) {
				yyerror("custom beaming may not be nested");
		}
		/* Non-custom beaming, if any, is done later, in do_bar() */
		else {
			/* begin custom beaming */
			Curr_grpsyl_p->beamloc = STARTITEM;
			if (Curr_grpsyl_p->basictime < 8 && yyvsp[0].intval == BT_SAME) {
				yyerror("beamed notes must be 8th or shorter");
			}
		}
		Curr_grpsyl_p->beamto = yyvsp[0].intval;
	}
break;
case 285:
#line 2506 "gram.y"
{
		Curr_grpsyl_p->breakbeam = YES;
	}
break;
case 286:
#line 2512 "gram.y"
{
		struct GRPSYL *last_nongrace_p;

		/* find the previous group, skipping grace groups */
		for (last_nongrace_p = Last_grpsyl_p;
				last_nongrace_p != (struct GRPSYL *) 0
				&& last_nongrace_p->grpvalue == GV_ZERO;
				last_nongrace_p = last_nongrace_p->prev) {
			;
		}

		/* check that a custom beam is in progress */
		if ((last_nongrace_p == (struct GRPSYL *) 0)
				|| ((last_nongrace_p != (struct GRPSYL *) 0)
				&& (last_nongrace_p->beamloc != STARTITEM)
				&& (last_nongrace_p->beamloc != INITEM) )) {
			yyerror("'ebm' unexpected: no custom beaming in progress");
		}
		else {
			Curr_grpsyl_p->beamloc = ENDITEM;
			Curr_grpsyl_p->beamto = last_nongrace_p->beamto;
			if (Curr_grpsyl_p->basictime < 8
					&& Last_grpsyl_p->beamto == BT_SAME) {
				yyerror("beamed notes must be 8th or shorter");
			}
		}
	}
break;
case 287:
#line 2541 "gram.y"
{
		/* use the same as last time or default timeunit. */
		copy_timeunit(Curr_grpsyl_p, Last_grpsyl_p);
	}
break;
case 288:
#line 2548 "gram.y"
{
		if (Curr_grpsyl_p->is_meas == YES) {
			Curr_grpsyl_p->fulltime = Score.time;
		}
		else {
			Curr_grpsyl_p->fulltime = yyvsp[0].ratval;
		}
	}
break;
case 289:
#line 2558 "gram.y"
{
		yyval.intval = BT_SAME;
	}
break;
case 290:
#line 2564 "gram.y"
{
		switch (yyvsp[0].intval) {
		case PL_ABOVE:
			yyval.intval = BT_ABOVE;
			break;
		case PL_BELOW:
			yyval.intval = BT_BELOW;
			break;
		default:
			yyerror("bm with staff must be 'above' or 'below'");
			yyval.intval = BT_SAME;
		}
	}
break;
case 291:
#line 2579 "gram.y"
{
		yyval.ratval.n = yyvsp[0].intval;
		yyval.ratval.d = 1;

		/* if filling in a GRPSYL struct, need to fill in basic time,
		 * could also be here due to beamstyle, in which case the
		 * Curr_grpsyl_p will be NULL, or when getting tuplet duration,
		 * in which case flag will be set */
		if (Curr_grpsyl_p != (struct GRPSYL *) 0
					&& Getting_tup_dur == NO) {
			/* 1/2 is 0 internally. 1/4 is -1 internally */
			Curr_grpsyl_p->basictime = (yyvsp[0].intval == 2 ? 0 : -1);
		}
	}
break;
case 292:
#line 2596 "gram.y"
{
		/* check that a power of two from 1 to MAXBASICTIME */
		(void) power_of2check(yyvsp[0].intval, "note basic time value");
		/* can't use rangecheck here because the error message would
		 * say 0 and -1 are valid times, which is only true internally--
		 * the user has to use 1/2, 1/4 or m. */
		if (yyvsp[0].intval < MINBASICTIME || yyvsp[0].intval > MAXBASICTIME) {
			l_yyerror(Curr_filename, yylineno,
				"time value must be between 1 and %d, or 1/2 or 1/4 or m",
				MAXBASICTIME);
		}
		yyval.ratval.n = 1;
		/* avoid division by zero */
		if (yyvsp[0].intval == 0) {
			yyvsp[0].intval = 1;
		}
		yyval.ratval.d = yyvsp[0].intval;
		if (Curr_grpsyl_p != (struct GRPSYL *) 0
					&& Getting_tup_dur == NO) {
			Curr_grpsyl_p->basictime = yyvsp[0].intval;
		}
	}
break;
case 293:
#line 2620 "gram.y"
{
		User_meas_time = NO;
	}
break;
case 294:
#line 2626 "gram.y"
{
		if (Curr_grpsyl_p != (struct GRPSYL *) 0
					&& Getting_tup_dur == NO) {
			Curr_grpsyl_p->is_meas = YES;
			User_meas_time = YES;
		}
		else {
			yyerror("'m' is not valid here");
		}
	}
break;
case 295:
#line 2638 "gram.y"
{
		yyval.intval = 0;
	}
break;
case 296:
#line 2644 "gram.y"
{
		/* count up the number of dots */
		yyval.intval = yyvsp[-1].intval + 1;
	}
break;
case 297:
#line 2650 "gram.y"
{
		/* We can't distinguish between the default 0.0 and
		 * if user explicitly sets to 0.0, so if they set to 0.0
		 * then to something else, we won't catch that as setting
		 * twice, but that shouldn't be common. Besides, if they
		 * expect them to be additive, adding to zero will work
		 * as they expect... Because of roundoff, we can't do
		 * exact compare for specifying the same value more than once,
		 * so treat as identical if pretty close. */
		if (Curr_grpsyl_p->padding != 0.0 && 
				fabs(Curr_grpsyl_p->padding - (yyvsp[0].floatval * STEPSIZE))
				> 0.0001) {
			l_warning(Curr_filename, yylineno,
				"padding specified more than once; using last instance");
		}
		Curr_grpsyl_p->padding = yyvsp[0].floatval * STEPSIZE;
	}
break;
case 298:
#line 2670 "gram.y"
{
		yyval.floatval = (float) yyvsp[-1].intval * yyvsp[0].floatval;
	}
break;
case 299:
#line 2675 "gram.y"
{
		/* no sign--must be a positive number */
		yyval.intval = 1;
	}
break;
case 300:
#line 2682 "gram.y"
{
		/* user wants a negative number */
		yyval.intval = -1;
	}
break;
case 301:
#line 2688 "gram.y"
{
		Curr_grpsyl_p->grpcont = GC_NOTES;
		if (yyvsp[0].intval == 0) {
			/* no notes listed, use same as previous group */
			copy_notes(Curr_grpsyl_p, Last_grpsyl_p);
		}
		else {
			resize_notelist(Curr_grpsyl_p);
		}
	}
break;
case 305:
#line 2713 "gram.y"
{
		/* allocate GRPSYL for the case where everything is defaulted
		 * from the previous group, with just an extra attribute
		 * like ? or ~ specified. */
		if (Curr_grpsyl_p == (struct GRPSYL *) 0) {
			/* shouldn't ever happen, but just in case... */
			Curr_grpsyl_p = newGRPSYL(GS_GROUP);
		}
		if (Last_grpsyl_p != (struct GRPSYL *) 0 &&
					Last_grpsyl_p->nnotes >= 1) {
			copy_notes(Curr_grpsyl_p, Last_grpsyl_p);
		}
		else if (Last_grpsyl_p != (struct GRPSYL *) 0 && 
				Last_grpsyl_p->grpcont == GC_REST) {
			Curr_grpsyl_p->grpcont = GC_REST;
		}
	}
break;
case 306:
#line 2733 "gram.y"
{
		if (Curr_grpsyl_p->grpcont == GC_NOTES &&
					Curr_grpsyl_p->nnotes > 0) {
			Curr_grpsyl_p->notelist[Curr_grpsyl_p->nnotes - 1]
						.notesize = GS_SMALL;
		}
		else if (Curr_grpsyl_p->grpcont == GC_REST) {
			Curr_grpsyl_p->grpsize  = GS_SMALL;
		}
		else {
			yyerror("no note specified for '?'");
		}
	}
break;
case 307:
#line 2749 "gram.y"
{
		if (Curr_grpsyl_p->grpcont != GC_NOTES) {
			yyerror("can't tie a rest or space");
		}
		else if (Curr_grpsyl_p->nnotes > 0) {
			struct NOTE *n_p;
			n_p = &(Curr_grpsyl_p->notelist[Curr_grpsyl_p->nnotes - 1]);
			n_p->tie = YES;
			n_p->tiestyle = yyvsp[-1].intval;
			n_p->tiedir = yyvsp[0].intval;
		}
		else {
			yyerror("no note specified for '~'");
		}
	}
break;
case 308:
#line 2767 "gram.y"
{
		if (Curr_grpsyl_p->nnotes > 0) {
			addsym(yyvsp[0].stringval, Curr_grpsyl_p->notelist
					[Curr_grpsyl_p->nnotes - 1].c, CT_NOTE);
			var_valid();
		}
		else if (Curr_grpsyl_p->grpcont == GC_REST) {
			addsym(yyvsp[0].stringval, Curr_grpsyl_p->c, CT_GRPSYL);
			var_valid();
		}
		else {
			l_yyerror(Curr_filename, yylineno,
				"no note specified for location tag '%s'", yyvsp[0].stringval);
		}
	}
break;
case 309:
#line 2785 "gram.y"
{
		/* this is for bend on a non-tablature staff */
		if (Doing_tab_staff == YES) {
			yyerror("^ bend not allowed on tablature staff; use quoted bend string");
		}
		else {
			add_slurto(Curr_grpsyl_p, yyvsp[-1].intval, yyvsp[0].intval,
					Curr_grpsyl_p->nnotes - 1, L_NORMAL);
			if (Curr_grpsyl_p->nnotes > 0) {
				Curr_grpsyl_p->notelist
						[Curr_grpsyl_p->nnotes - 1]
						.is_bend = YES;
			}
		}
	}
break;
case 310:
#line 2803 "gram.y"
{
		/* this is for a small bend (1/4 step) on a non-tab staff */
		if (Doing_tab_staff == YES) {
			yyerror("^/ not allowed on tablature staff; use quoted bend string");
		}
		else {
			Curr_grpsyl_p->notelist[Curr_grpsyl_p->nnotes - 1]
					.smallbend = YES;
		}
	}
break;
case 311:
#line 2816 "gram.y"
{
		if (Curr_grpsyl_p->grpcont == GC_NOTES &&
					Curr_grpsyl_p->nnotes > 0) {
			if ((Curr_grpsyl_p->notelist[Curr_grpsyl_p->nnotes - 1]
					.headshape = get_shape_num(yyvsp[0].stringval + 2))
					== HS_UNKNOWN) {
				l_yyerror(Curr_filename, yylineno,
					"'%s' is not a valid head shape name",
					ascii_str(yyvsp[0].stringval, YES, NO, TM_NONE));
			}
		}
		else {
			yyerror("no note specified for headshape");
		}
		FREE(yyvsp[0].stringval);
	}
break;
case 315:
#line 2844 "gram.y"
{
		/* no notes -- use same as last time */
		yyval.intval = 0;
	}
break;
case 316:
#line 2851 "gram.y"
{
		/* return number of notes in notelist */
		yyval.intval = yyvsp[-2].intval + 1;
	}
break;
case 318:
#line 2860 "gram.y"
{
		add_note(Curr_grpsyl_p, (int) PP_REST, 0, USE_DFLT_OCTAVE,
							0, NO, (char *) 0);
	}
break;
case 319:
#line 2867 "gram.y"
{
		/* temporarily stash uncompressibility (NO in this case)
		 * in the octave field */
		add_note(Curr_grpsyl_p, (int) PP_SPACE, 0, NO,
							0, NO, (char *) 0);
		if (Curr_grpsyl_p->is_meas == YES && User_meas_time == YES) {
			yyerror("cannot specify time value on measure space");
		}
	}
break;
case 320:
#line 2879 "gram.y"
{
		/* temporarily stash uncompressibility (YES in this case)
		 * in the octave field */
		add_note(Curr_grpsyl_p, (int) PP_SPACE, 0, YES,
							0, NO, (char *) 0);
		if (Curr_grpsyl_p->is_meas == YES && User_meas_time == YES) {
			yyerror("cannot specify time value on measure uncompressible space");
		}
	}
break;
case 321:
#line 2891 "gram.y"
{
		if (Curr_grpsyl_p->is_meas == YES) {
			if (User_meas_time == YES) {
				yyerror("cannot specify time value on m rpt");
			}
		}
		else {
			yyerror("rpt can only be used with m");
		}
		add_note(Curr_grpsyl_p, (int) PP_RPT, 0, USE_DFLT_OCTAVE,
							0, NO, (char *) 0);
	}
break;
case 322:
#line 2906 "gram.y"
{
		if (yyvsp[0].intval == SINGLEBAR) {
			/* lexer thinks this is 'bar'
			 * but is really 'b' 'a' 'r' */
			keyword_notes(yytext);
		}
		else if (yyvsp[0].intval == ENDBAR) {
			/* lexer thinks this is 'endbar'
			 * but is really 'en' 'd' 'b' 'a' 'r' */
			add_note(Curr_grpsyl_p, (int) 'e', 'n',
				USE_DFLT_OCTAVE, 0, NO, (char *) 0);
			keyword_notes(yytext + 2);
		}
		else {
			yyerror("bar type not valid here");
		}
	}
break;
case 323:
#line 2926 "gram.y"
{
		/* lexer thinks this is 'grace' but really 'g' 'r' 'a' 'c' 'e' */
		keyword_notes(yytext);
	}
break;
case 324:
#line 2933 "gram.y"
{
		if (yyvsp[0].intval == ENDING_BARRED) {
			keyword_notes(yytext);
		}
		else {
			l_yyerror(Curr_filename, yylineno,
					"%s not valid here", yytext);
		}
	}
break;
case 325:
#line 2944 "gram.y"
{
		if (yyvsp[0].intval == J_RAGPARA) {
			keyword_notes(yytext);
		}
		else {
			l_yyerror(Curr_filename, yylineno,
					"%s not valid here", yytext);
		}
	}
break;
case 326:
#line 2956 "gram.y"
{
		if (yyvsp[0].intval == BRACELIST) {
			keyword_notes(yytext);
		}
		else {
			l_yyerror(Curr_filename, yylineno,
					"%s not valid here", yytext);
		}
	}
break;
case 327:
#line 2968 "gram.y"
{
		if (yyvsp[0].intval == BASS) {
			keyword_notes(yytext);
		}
		else {
			l_yyerror(Curr_filename, yylineno,
					"%s not valid here", yytext);
		}
	}
break;
case 328:
#line 2980 "gram.y"
{
		Curr_grpsyl_p->notelist[Curr_grpsyl_p->nnotes - 1].note_has_paren = YES;
	}
break;
case 329:
#line 2985 "gram.y"
{
		if (Doing_tab_staff == YES) {
			add_note(Curr_grpsyl_p, yyvsp[-1].intval, 0, NOFRET, 0, NO, yyvsp[0].stringval);
		}
		else {
			add_note(Curr_grpsyl_p, yyvsp[-1].intval, 0, USE_DFLT_OCTAVE, 0, NO, yyvsp[0].stringval);
		}
	}
break;
case 330:
#line 2996 "gram.y"
{
		if (Doing_tab_staff == YES) {
			yyerror("extraneous parentheses around accidental");
		}
		add_note(Curr_grpsyl_p, yyvsp[-4].intval, yyvsp[-2].intval, yyvsp[0].intval, 0, YES, (char *) 0);
	}
break;
case 331:
#line 3005 "gram.y"
{
		if (Doing_tab_staff == NO) {
			yyerror("extraneous parentheses around octave");
		}
		add_note(Curr_grpsyl_p, yyvsp[-4].intval, 0, yyvsp[-2].intval, 0, YES, yyvsp[0].stringval);
	}
break;
case 332:
#line 3014 "gram.y"
{
		/* Whoops! The left paren was really for a new note,
		 * which happens to be parenthesized. But beause yacc only
		 * looks ahead one token, it finds out too late. So we
		 * catch it here, push the parenthesis and pitch back into
		 * the input and return back to the parent grammar rule,
		 * since we now have complete note. */
		pushback(yytext[0]);
		pushback('(');
		add_note(Curr_grpsyl_p, yyvsp[-2].intval, 0,
			(Doing_tab_staff ? NOFRET : USE_DFLT_OCTAVE),
			0, NO, (char *) 0);
	}
break;
case 333:
#line 3030 "gram.y"
{
		add_note(Curr_grpsyl_p, yyvsp[-4].intval, yyvsp[-3].intval, yyvsp[-1].intval, yyvsp[-2].intval, NO, yyvsp[0].stringval);
	}
break;
case 334:
#line 3036 "gram.y"
{
		if (Doing_tab_staff == NO) {
			yyerror("extraneous parentheses around octave");
		}
		add_note(Curr_grpsyl_p, yyvsp[-6].intval, yyvsp[-5].intval, yyvsp[-2].intval, yyvsp[-4].intval, YES, yyvsp[0].stringval);
	}
break;
case 335:
#line 3045 "gram.y"
{
		/* Whoops! The left paren was really for a new note,
		 * which happens to be parenthesized. But beause yacc only
		 * looks ahead one token, it finds out too late. So we
		 * catch it here, push the parenthesis and pitch back into
		 * the input and return back to the parent grammar rule,
		 * since we now have complete note. */
		pushback(yytext[0]);
		pushback('(');
		add_note(Curr_grpsyl_p, yyvsp[-3].intval, yyvsp[-2].intval,
			(Doing_tab_staff ? NOFRET : USE_DFLT_OCTAVE),
			0, NO, (char *) 0);
	}
break;
case 336:
#line 3061 "gram.y"
{
		add_note(Curr_grpsyl_p, yyvsp[-3].intval, 0, yyvsp[-1].intval, yyvsp[-2].intval, NO, yyvsp[0].stringval);
	}
break;
case 337:
#line 3067 "gram.y"
{
		add_note(Curr_grpsyl_p, yyvsp[-5].intval, 0, yyvsp[-2].intval, yyvsp[-4].intval, YES, yyvsp[0].stringval);
	}
break;
case 338:
#line 3073 "gram.y"
{
		if (Doing_tab_staff == YES && yyvsp[-1].intval != '\0') {
			yyerror("accidental must be before fret number");
		}
		add_note(Curr_grpsyl_p, yyvsp[-3].intval, yyvsp[-1].intval, yyvsp[-2].intval, 0, NO, yyvsp[0].stringval);
	}
break;
case 339:
#line 3082 "gram.y"
{
		add_note(Curr_grpsyl_p, yyvsp[-4].intval, yyvsp[-1].intval, yyvsp[-3].intval, 0, YES, (char *) 0);
	}
break;
case 340:
#line 3088 "gram.y"
{
		/* Whoops! The left paren was really for a new note,
		 * which happens to be parenthesized. But beause yacc only
		 * looks ahead one token, it finds out too late. So we
		 * catch it here, push the parenthesis and pitch back into
		 * the input and return back to the parent grammar rule,
		 * since we now have complete note. */
		pushback(yytext[0]);
		pushback('(');
		add_note(Curr_grpsyl_p, yyvsp[-3].intval, 0, yyvsp[-2].intval, 0, NO, (char *) 0);
	}
break;
case 341:
#line 3101 "gram.y"
{
		yyval.stringval = (char *) 0;
	}
break;
case 342:
#line 3107 "gram.y"
{
		if (Doing_tab_staff == NO) {
			/* try to give helpful error message */
			if (get_shape_num(yyvsp[0].stringval + 2) != HS_UNKNOWN) {
				yyerror("missing 'hs' before headshape string");
			}
			else if (strcmp(yyvsp[0].stringval + 2, "full") == 0 || isdigit(yyvsp[0].stringval[2])) {
				yyerror("bend string not allowed on non-tablature staff; use ^");
			}
			else {
				yyerror("unexpected string");
			}
			yyval.stringval = (char *) 0;
		}
	}
break;
case 343:
#line 3124 "gram.y"
{
		/* return 'a' to 'g' value */
		yyval.intval = (int) *yytext;
	}
break;
case 344:
#line 3130 "gram.y"
{
		/* no accidental */
		yyval.intval = 0;
	}
break;
case 345:
#line 3137 "gram.y"
{
	}
break;
case 346:
#line 3141 "gram.y"
{
		/* no octave or fret designation */
		yyval.intval = (Doing_tab_staff ? NOFRET : USE_DFLT_OCTAVE);
	}
break;
case 347:
#line 3148 "gram.y"
{
	}
break;
case 348:
#line 3153 "gram.y"
{
		yyval.intval = '#';
	}
break;
case 349:
#line 3159 "gram.y"
{
		yyval.intval = '&';
	}
break;
case 350:
#line 3165 "gram.y"
{
		/* natural */
		yyval.intval = 'n';
	}
break;
case 351:
#line 3172 "gram.y"
{
		/* double sharp */
		yyval.intval = 'x';
	}
break;
case 352:
#line 3179 "gram.y"
{
		yyval.intval = 'B';
	}
break;
case 353:
#line 3185 "gram.y"
{
		yyval.intval = yyvsp[0].intval;
	}
break;
case 354:
#line 3191 "gram.y"
{
		/* we can't really fill in the actual octave yet, because
		 * it may be different on different staffs or voices, so
		 * we store the relative octave and fill in actual value
		 * later.
		 */
		/* this will be a negative number */
		yyval.intval = yyvsp[0].intval;
	}
break;
case 355:
#line 3203 "gram.y"
{
		/* we can't really fill in the actual octave yet, because
		 * it may be different on different staffs or voices, so
		 * we store the relative octave and fill in actual value
		 * later.
		 */
		yyval.intval = USE_DFLT_OCTAVE + yyvsp[0].intval;
	}
break;
case 356:
#line 3213 "gram.y"
{
		yyval.intval = -1;
	}
break;
case 357:
#line 3219 "gram.y"
{
		/* count up the number of minus signs */
		yyval.intval = yyvsp[-1].intval - 1;
	}
break;
case 358:
#line 3225 "gram.y"
{
		yyval.intval = 1;
	}
break;
case 359:
#line 3231 "gram.y"
{
		/* count up the number of plus signs */
		yyval.intval = yyvsp[-1].intval + 1;
	}
break;
case 360:
#line 3238 "gram.y"
{
		set_slurdir(Curr_grpsyl_p, Curr_grpsyl_p->nnotes - 1, yyvsp[0].intval);
	}
break;
case 361:
#line 3245 "gram.y"
{
		Slurstyle = yyvsp[0].intval;
		begin_slur(Curr_grpsyl_p, Curr_grpsyl_p->nnotes - 1);
	}
break;
case 362:
#line 3251 "gram.y"
{
		/* empty list. Only allowed if only one note in following
		 * group. However, we don't know that yet, so marked pitch
		 * as 'U' for unknown */
		add_slurto(Curr_grpsyl_p, 'U', USE_DFLT_OCTAVE,
				Curr_grpsyl_p->nnotes - 1, Slurstyle);
	}
break;
case 363:
#line 3261 "gram.y"
{
		add_slurto(Curr_grpsyl_p, 0, yyvsp[0].intval, Curr_grpsyl_p->nnotes - 1,
						Slurstyle);
	}
break;
case 367:
#line 3279 "gram.y"
{
		if (Doing_tab_staff == YES) {
			yyerror("cannot specify string inside <> on tab staffs");
		}
		else {
			add_slurto(Curr_grpsyl_p, yyvsp[-1].intval, yyvsp[0].intval,
					Curr_grpsyl_p->nnotes - 1, Slurstyle);
		}
	}
break;
case 368:
#line 3291 "gram.y"
{
		/* save address associated with entire group */
		if ( yyvsp[0].stringval != (char *) 0) {
			addsym(yyvsp[0].stringval, Curr_grpsyl_p->c, CT_GRPSYL);
			var_valid();
		}
	}
break;
case 369:
#line 3300 "gram.y"
{
		yyval.stringval = strcpy(Stringbuff, yytext);
	}
break;
case 370:
#line 3306 "gram.y"
{
		/* a-g are usually pitches, but in this context, they
		 * are 1-character variable names. */
		yyval.stringval = strcpy(Stringbuff, yytext);
	}
break;
case 371:
#line 3314 "gram.y"
{
		/* usually m means measure, but here it is the variable m */
		yyval.stringval = strcpy(Stringbuff, yytext);
	}
break;
case 372:
#line 3321 "gram.y"
{
		/* usually rest, here variable r */
		yyval.stringval = strcpy(Stringbuff, yytext);
	}
break;
case 373:
#line 3328 "gram.y"
{
		/* usually space, here variable s */
		yyval.stringval = strcpy(Stringbuff, yytext);
	}
break;
case 374:
#line 3335 "gram.y"
{
		/* usually uncompressible, here variable u */
		yyval.stringval = strcpy(Stringbuff, yytext);
	}
break;
case 375:
#line 3342 "gram.y"
{
		/* usually natural, here variable n */
		yyval.stringval = strcpy(Stringbuff, yytext);
	}
break;
case 376:
#line 3349 "gram.y"
{
		/* usually double sharp, here variable x */
		yyval.stringval = strcpy(Stringbuff, yytext);
	}
break;
case 377:
#line 3355 "gram.y"
{
		/* longer variable -- one with more than one character name */
		yyval.stringval = strcpy(Stringbuff, yytext);
	}
break;
case 378:
#line 3361 "gram.y"
{
		end_tuplet(yyvsp[-3].intval, yyvsp[-1].ratval, yyvsp[-2].intval, yyvsp[-4].intval);
		Getting_tup_dur = NO;
	}
break;
case 379:
#line 3367 "gram.y"
{
		/* nothing -- use default of when to print tuplet number/bracket */
		yyval.intval = PT_DEFAULT;
	}
break;
case 380:
#line 3374 "gram.y"
{
		/* don't print tuplet or bracket */
		yyval.intval = PT_NEITHER;
	}
break;
case 381:
#line 3381 "gram.y"
{
		if (*yytext == 'y') {
			yyval.intval = PT_BOTH;
		}
		else {
			yyerror("tuplet number/bracket qualifier must be y or n");
		}
	}
break;
case 382:
#line 3393 "gram.y"
{
		begin_tuplet();
	}
break;
case 383:
#line 3398 "gram.y"
{
		Getting_tup_dur = YES;
	}
break;
case 384:
#line 3403 "gram.y"
{
		yyval.intval = PL_UNKNOWN;
	}
break;
case 385:
#line 3409 "gram.y"
{
		if (yylval.intval == PL_BETWEEN) {
			yyerror("between not allowed for tuplet side");
			yyval.intval = PL_UNKNOWN;
		}
	}
break;
case 386:
#line 3417 "gram.y"
{
		/* optional time value is missing */
		yyval.ratval.n = 0;
		yyval.ratval.d = 1;
	}
break;
case 387:
#line 3425 "gram.y"
{
		yyval.ratval = yyvsp[0].ratval;
	}
break;
case 388:
#line 3430 "gram.y"
{
	}
break;
case 389:
#line 3434 "gram.y"
{
		Last_grpsyl_p = Curr_grpsyl_p = (struct GRPSYL *) 0;
	}
break;
case 390:
#line 3439 "gram.y"
{
		/* lyrics aren't STUFF, so they don't go through all the
		 * same code. They also have the oddity of a default place
		 * of below, whereas a lot of other things default to above.
		 * So if user didn't specify a place, fix that. */
		if (yyvsp[0].intval == PL_UNKNOWN) {
			Place = PL_BELOW;
		}
	}
break;
case 391:
#line 3450 "gram.y"
{
		/* empty, use default */
		begin_range(PL_UNKNOWN);
		yyval.intval = PL_UNKNOWN;
	}
break;
case 392:
#line 3458 "gram.y"
{
		begin_range(yyvsp[0].intval);
	}
break;
case 394:
#line 3466 "gram.y"
{
		/* means goes above or below all the staffs, regardless if
		 * some happen to be invisible at the moment, so find top
		 * or bottom visible staff as appropriate,and save that
		 * away as the staff range */
		all();
	}
break;
case 396:
#line 3479 "gram.y"
{
		/* if staff_range == 0, then only one staff specified, so
		 * use first staff number for both beginning and end of range */
		save_staff_range(yyvsp[-1].intval, (yyvsp[0].intval == 0 ? yyvsp[-1].intval : yyvsp[0].intval) );
	}
break;
case 397:
#line 3486 "gram.y"
{
		/* empty */
		chk_range_type(NO);
		yyval.intval = 0;
	}
break;
case 398:
#line 3494 "gram.y"
{
		chk_range_type(NO);
		yyval.intval = yyvsp[0].intval;
		if (yyvsp[0].intval == 0) {
			yyerror("staff of 0 is illegal");
		}
	}
break;
case 399:
#line 3504 "gram.y"
{
		chk_range_type(YES);
		yyval.intval = yyvsp[0].intval;
	}
break;
case 400:
#line 3511 "gram.y"
{
	}
break;
case 401:
#line 3516 "gram.y"
{
		if (*yytext != 'c') {
			yyerror("verse must be 'c' or number(s)");
		}
		else {
			save_vno_range(0, 0);
		}
	}
break;
case 402:
#line 3527 "gram.y"
{
	}
break;
case 403:
#line 3532 "gram.y"
{
	}
break;
case 404:
#line 3537 "gram.y"
{
		/* if end of range is 0, not really a range, use first number
		 * for both beginning and end */
		lyr_verse(yyvsp[-1].intval, (yyvsp[0].intval == 0 ? yyvsp[-1].intval : yyvsp[0].intval));
	}
break;
case 405:
#line 3544 "gram.y"
{
		/* empty */
		yyval.intval = 0;
	}
break;
case 406:
#line 3551 "gram.y"
{
		yyval.intval = yyvsp[0].intval;
	}
break;
case 407:
#line 3556 "gram.y"
{
		/* null token to allocate a GRPSYL for a lyric */
		Last_grpsyl_p = Curr_grpsyl_p;
		Curr_grpsyl_p = newGRPSYL(GS_SYLLABLE);
		if (Last_grpsyl_p == (struct GRPSYL *) 0) {
			Lyrics_p = Curr_grpsyl_p;
		}
	}
break;
case 408:
#line 3566 "gram.y"
{
		/* copies of the lyrics info has been made for all staffs/verses
		 * at this point, so get rid of master copy */
		free_grpsyls(Lyrics_p);
	}
break;
case 409:
#line 3573 "gram.y"
{
		/* empty -- need to derive times from music */
		Lyrics_p = derive_lyrtime();
	}
break;
case 414:
#line 3592 "gram.y"
{
		link_notegroup(Curr_grpsyl_p, Last_grpsyl_p);
	}
break;
case 415:
#line 3598 "gram.y"
{
	}
break;
case 416:
#line 3602 "gram.y"
{
		end_tuplet(yyvsp[-1].intval, yyvsp[0].ratval, NO, PL_UNKNOWN);
		Getting_tup_dur = NO;
	}
break;
case 417:
#line 3608 "gram.y"
{
		/* empty, will fill in an actual syllable later */
		Curr_grpsyl_p->syl = (char *) 0;
	}
break;
case 418:
#line 3615 "gram.y"
{
		/* space, not a lyric, so mark for later use */
		Curr_grpsyl_p->grpcont = GC_SPACE;
	}
break;
case 421:
#line 3628 "gram.y"
{
		/* If no [verseno] is specified, use -1 as special flag
		 * to be resolved later to "one more than previous",
		 * or if there wasn't a previous, to verse 1.
		 */
		lyr_verse(-1, -1);
		proc_lyrics(Lyrics_p, yyvsp[-1].stringval);
	}
break;
case 422:
#line 3639 "gram.y"
{
		proc_lyrics(Lyrics_p, yyvsp[-1].stringval);
	}
break;
case 423:
#line 3644 "gram.y"
{
		Currstruct_p = (struct MAINLL *) 0;
	}
break;
case 426:
#line 3656 "gram.y"
{
		struct MAINLL * mll_p;
		char *name;


		name = (yylval.intval == YES ? "newpage" : "newscore");

		(void) contextcheck(C_MUSIC | C_BLOCK, name);
		Currstruct_p = newMAINLLstruct(S_FEED, yylineno);
		Currstruct_p->u.feed_p->pagefeed = yyvsp[0].intval;

		/* make sure we're not in the middle of a measure. Go
		 * backwards in main list. If we hit STAFF before a BAR,
		 * then there is a problem.
		 */
		for (mll_p = Mainlltc_p; mll_p != (struct MAINLL *) 0;
					mll_p = mll_p->prev) {

			if (mll_p->str == S_BAR) {
				break;
			}
			else if (mll_p->str == S_STAFF) {
				l_yyerror(Curr_filename, yylineno,
					"%s not allowed in middle of measure",
					name);
				break;
			}
			else if (mll_p->str == S_FEED) {
				if (mll_p->prev != 0 &&
						mll_p->prev->str == S_BLOCKHEAD) {
					/* This is a feed following a block.
					 * If it was an implicit feed,
					 * we can get rid of it, because
					 * user now put an explicit one.
					 */
					if (mll_p->inputlineno == -1) {
						unlinkMAINLL(mll_p);
						FREE(mll_p);
					}
					/* We must be in block context,
					 * so no need to check farther back.
					 */
					break;
				}
				else {
					l_yyerror(Curr_filename, yylineno,
						"consecutive newscore/newpage not allowed");
				}
				break;
			}
		}
		insertMAINLL(Currstruct_p, Mainlltc_p);

		/* If this is inside a block, we need to create a new block
		 * and feed after it.
		 */
		if (Context == C_BLOCK) {
			CALLOC(BLOCKHEAD, Currblock_p, 1);
			set_win_coord(Currblock_p->c);
			Next_print_link_p_p = &(Currblock_p->printdata_p);
			insertMAINLL(newMAINLLstruct(S_BLOCKHEAD, yylineno),
						Mainlltc_p);
			Mainlltc_p->u.blockhead_p = Currblock_p;
			insertMAINLL(newMAINLLstruct(S_FEED, -1), Mainlltc_p);
		}

	}
break;
case 427:
#line 3724 "gram.y"
{
		Currstruct_p = (struct MAINLL *) 0;
	}
break;
case 431:
#line 3740 "gram.y"
{
		yyvsp[0].floatval = adjust2inches(yyvsp[0].floatval);	/* in case we are in centimeter mode */
		if (yyvsp[-2].intval == RIGHTMARGIN) {
			if (Currstruct_p->u.feed_p->rightmargin >= 0) {
				l_warning(Curr_filename, yylineno,
					"rightmargin specified more than once, using last instance");
				/* fall through to override the previous */
			}
			Currstruct_p->u.feed_p->rightmargin = yyvsp[0].floatval;
		}
		else if (yyvsp[-2].intval == LEFTMARGIN) {
			if (Currstruct_p->u.feed_p->leftmargin >= 0) {
				l_warning(Curr_filename, yylineno,
					"leftmargin specified more than once, using last instance");
				/* fall through to override the previous */
			}
			Currstruct_p->u.feed_p->leftmargin = yyvsp[0].floatval;
		}
		else {
			yyerror("unexpected parameter; only 'leftmargin' or 'rightmargin' allowed here");
		}
		chkmargin(Score.topmargin, Score.botmargin,
				Currstruct_p->u.feed_p->leftmargin,
				Currstruct_p->u.feed_p->rightmargin);
	}
break;
case 438:
#line 3786 "gram.y"
{

		/* build a BAR struct and add it to the main list */
		if (contextcheck(C_MUSIC, "bar") == YES) {

			Currstruct_p = newMAINLLstruct(S_BAR, yylineno);
			Currstruct_p->u.bar_p->bartype = (short) yylval.intval;
			Currstruct_p->u.bar_p->endingloc = Endingloc;
			if (yyvsp[-2].intval != L_NORMAL) {
				if (yyvsp[-2].intval != L_DASHED && yyvsp[-2].intval != L_DOTTED) {
					yyerror("bar modifier can only be 'dashed' or 'dotted'");
				}
				else if (yyvsp[0].intval != SINGLEBAR && yyvsp[0].intval != DOUBLEBAR) {
					yyerror("only 'bar' or 'dblbar' can be dashed or dotted");
				}
			}
			Currstruct_p->u.bar_p->linetype = yyvsp[-2].intval;
			insertMAINLL(Currstruct_p, Mainlltc_p);
			Currstruct_p->u.bar_p->timedssv_p = tssv_sort();

			/* re-order things in the  bar,
			 * make sure they are all consistent, etc */
			do_bar(yyvsp[0].intval);

			Got_ending = NO;
		}
		if (yyvsp[0].intval == RESTART && Endingloc != NOITEM) {
			yyerror("restart cannot be used inside an ending");
		}

		/* User can specify bar type to use on preceeding staff when
		 * a repeatstart gets moved to the next scores's pseudo-bar */
		if (yyvsp[-1].intval != -1) {
			if (yyvsp[0].intval != REPEATSTART) {
				yyerror("bar type for preceeding score only allowed on repeatstart");
			}
			if (Currstruct_p != 0 && Currstruct_p->u.bar_p != 0) {
				Currstruct_p->u.bar_p->precbartype = yyvsp[-1].intval;
			}
		}
		else {
			/* Default is single bar. It could be argued that in
			 * the case of a key change on this bar, it really
		 	 * should be a dblbar, but user can force that
			 * if they want it. */
			if (Currstruct_p != 0 && Currstruct_p->u.bar_p != 0) {
				Currstruct_p->u.bar_p->precbartype = SINGLEBAR;
			}
		}
	}
break;
case 439:
#line 3839 "gram.y"
{
		yyval.intval = -1;	/* special value to mean "none specified" */
	}
break;
case 440:
#line 3845 "gram.y"
{
		/* Parens wouldn't be strictly necessary to be able to parse,
		 * but seem useful semantically to make it clear this is
		 * optional, only applying when repeatstart is moved to
		 * next score's pseudo bar. */
		yyval.intval = yyvsp[-1].intval;
	}
break;
case 443:
#line 3861 "gram.y"
{
		if (Currstruct_p != (struct MAINLL *) 0) {
			Currstruct_p->u.bar_p->padding += yyvsp[0].floatval * STEPSIZE;
		}
	}
break;
case 444:
#line 3869 "gram.y"
{
		if (Currstruct_p != (struct MAINLL *) 0) {
			/* fill in location info */
			if (yyvsp[0].stringval != (char *) 0) {
				addsym(yyvsp[0].stringval, Currstruct_p->u.bar_p->c, CT_BAR);
			}
		}
	}
break;
case 445:
#line 3880 "gram.y"
{
		if (Got_ending == YES) {
			yyerror("Only one ending allowed per bar");
		}
		Got_ending = YES;

		if (Currstruct_p != (struct MAINLL *) 0) {
			if (Currstruct_p->u.bar_p->bartype == RESTART) {
				yyerror("ending not allowed on restart");
			}
			/* fill in ending label if any */
			Currstruct_p->u.bar_p->endinglabel = yyvsp[0].stringval;
			Currstruct_p->u.bar_p->endingloc = Endingloc;
			ped_endings(Endingloc);

			/* for next time around, figure out what endingloc will
			 * be if user doesn't specify something different */
			switch (Endingloc) {
			case STARTITEM:
				Endingloc = INITEM;
				break;
			case ENDITEM:
				Endingloc = NOITEM;
				break;
			default:
				break;
			}
		}
	}
break;
case 446:
#line 3912 "gram.y"
{
		if (Currstruct_p != 0) {
			if (Currstruct_p->u.bar_p->bartype == RESTART) {
				yyerror("hidechanges not allowed on restart");
			}
			Currstruct_p->u.bar_p->hidechanges = YES;
		}
	}
break;
case 447:
#line 3922 "gram.y"
{
		if (Currstruct_p != 0) {
			set_mnum(Currstruct_p->u.bar_p, yyvsp[0].intval);
		}
	}
break;
case 448:
#line 3929 "gram.y"
{
		/* +2 to skip font/size */
		init_reh(-1, yyvsp[0].stringval + 2, Currstruct_p);
	}
break;
case 449:
#line 3935 "gram.y"
{
		init_reh(yyvsp[0].intval, (char *)0, Currstruct_p);
	}
break;
case 450:
#line 3941 "gram.y"
{
		if (Currstruct_p != (struct MAINLL *) 0) {
			set_reh_string(Currstruct_p->u.bar_p, yyvsp[-4].intval, yyvsp[-3].intval, yyvsp[-2].intval, yyvsp[-1].stringval);
			Currstruct_p->u.bar_p->dist = Dist;
			Currstruct_p->u.bar_p->dist_usage = Dist_usage;
		}
	}
break;
case 451:
#line 3951 "gram.y"
{
		yyval.stringval = yyvsp[0].stringval;
		(void) fix_string(yyvsp[0].stringval, FONT_TR, DFLT_SIZE, Curr_filename, yylineno);
		Endingloc = STARTITEM;
	}
break;
case 452:
#line 3959 "gram.y"
{
		if (Endingloc == NOITEM) {
			yyerror("no ending in progress");
		}
		else {
			Endingloc = ENDITEM;
		}
		yyval.stringval = (char *) 0;
	}
break;
case 453:
#line 3970 "gram.y"
{
		/* null token to check for more than one rehearsal mark
		 * on one BAR */
		if (Currstruct_p != (struct MAINLL *) 0 &&
				Currstruct_p->u.bar_p->reh_type != REH_NONE) {
			yyerror("only one rehearsal mark allowed per bar");
		}
	}
break;
case 454:
#line 3980 "gram.y"
{
		if (Currstruct_p != (struct MAINLL *) 0) {
			Currstruct_p->u.bar_p->reh_type = REH_STRING;
		}
	}
break;
case 455:
#line 3988 "gram.y"
{
		if (Currstruct_p != (struct MAINLL *) 0) {
			Currstruct_p->u.bar_p->reh_type = REH_MNUM;
		}
		yyval.stringval = (char *) 0;
	}
break;
case 456:
#line 3997 "gram.y"
{
		if (Currstruct_p != (struct MAINLL *) 0) {
			Currstruct_p->u.bar_p->reh_type = REH_NUM;
		}
		yyval.stringval = (char *) 0;
	}
break;
case 457:
#line 4006 "gram.y"
{
		if (Currstruct_p != (struct MAINLL *) 0) {
			Currstruct_p->u.bar_p->reh_type = REH_LET;
		}
		yyval.stringval = (char *) 0;
	}
break;
case 458:
#line 4014 "gram.y"
{
		yyval.inpcoord_p = yyvsp[-4].inpcoord_p;
	}
break;
case 459:
#line 4019 "gram.y"
{
		/* null token to allocate an INPCOORD */
		CALLOC(INPCOORD, Curr_loc_info_p, 1);
		yyval.inpcoord_p = Curr_loc_info_p;
		/* set to positive in case input starts with an absolute
		 * x coordinate */
		Plus_minus = 1;
	}
break;
case 460:
#line 4029 "gram.y"
{
		/* reset sign to positive in case y coordinate is absolute */
		Plus_minus = 1;
	}
break;
case 462:
#line 4038 "gram.y"
{
		/* only 1 of these allowed per location */
		if (Curr_loc_info_p->hor_p != (float *) 0) {
			pfatal("only one horizontal location variable allowed per coordinate");
		}
		/* fill in the coord address and type */
		Curr_loc_info_p->hor_p = symval(yyvsp[-2].stringval);
		Curr_loc_info_p->htype = yyvsp[0].intval;
	}
break;
case 464:
#line 4054 "gram.y"
{
		/* only valid letters is w */
		if (*yytext != 'w') {
			yyerror("invalid direction: must be x, w, or e");
		}
		yyval.intval = AW;
	}
break;
case 465:
#line 4063 "gram.y"
{
		yyval.intval = AX;
	}
break;
case 466:
#line 4069 "gram.y"
{
		/* only valid value is e */
		if (*yytext != 'e') {
			yyerror("invalid direction: must be x, w, or e");
		}
		yyval.intval = AE;
	}
break;
case 470:
#line 4086 "gram.y"
{
		yyval.intval = Plus_minus = 1;
	}
break;
case 471:
#line 4092 "gram.y"
{
		yyval.intval = Plus_minus = -1;
	}
break;
case 472:
#line 4097 "gram.y"
{
		Curr_loc_info_p->hsteps += yyvsp[0].floatval * (float) Plus_minus;
	}
break;
case 473:
#line 4103 "gram.y"
{
		if (Curr_loc_info_p->hor_p == (float *) 0) {
			yyerror("must specify location before giving time offset");
		}
		/* add or subtract time value from horizontal */
		if (Plus_minus == 1) {
			Curr_loc_info_p->counts += yyvsp[0].floatval;
		}
		else {
			Curr_loc_info_p->counts -= yyvsp[0].floatval;
		}
	}
break;
case 476:
#line 4123 "gram.y"
{
		/* only one of these allowed per location */
		if (Curr_loc_info_p->vert_p != (float *) 0) {
			pfatal("only one vertical location variable allowed per coordinate");
		}
		/* fill in the coord address and type */
		Curr_loc_info_p->vert_p = symval(yyvsp[-2].stringval);
		Curr_loc_info_p->vtype = yyvsp[0].intval;

	}
break;
case 477:
#line 4135 "gram.y"
{
		yyval.intval = AN;
	}
break;
case 478:
#line 4141 "gram.y"
{
		yyval.intval = AS;
	}
break;
case 479:
#line 4147 "gram.y"
{
		/* only valid value here is y */
		if (*yytext != 'y') {
			yyerror("invalid direction: must be y, n, or s");
		}
		yyval.intval = AY;
	}
break;
case 482:
#line 4162 "gram.y"
{
	}
break;
case 483:
#line 4167 "gram.y"
{
		/* this is number of steps up (or down if negative) */
		Curr_loc_info_p->vsteps += yyvsp[0].floatval * (float) Plus_minus;
	}
break;
case 484:
#line 4173 "gram.y"
{
		yyval.floatval = (float) yyvsp[0].intval;
	}
break;
case 485:
#line 4178 "gram.y"
{
		sprintf(Tmpbuff,"%d.%s", yyvsp[-2].intval, yyvsp[0].stringval);
		yyval.floatval = (float) atof(Tmpbuff);
	}
break;
case 486:
#line 4184 "gram.y"
{
		sprintf(Tmpbuff,"0.%s", yyvsp[0].stringval);
		yyval.floatval = (float) atof(Tmpbuff);
	}
break;
case 487:
#line 4190 "gram.y"
{
		/* no decimal fraction part of float number */
		yyval.stringval = "";
	}
break;
case 489:
#line 4200 "gram.y"
{
		yyval.stringval = yytext;
	}
break;
case 491:
#line 4208 "gram.y"
{
		Currstruct_p->u.line_p->linetype = yyvsp[-6].intval;
		Currstruct_p->u.line_p->start = *(yyvsp[-3].inpcoord_p);
		Currstruct_p->u.line_p->end = *(yyvsp[-1].inpcoord_p);
		Currstruct_p->u.line_p->string = yyvsp[0].stringval;

		/* copies of the location info went into the LINE struct,
		 * so we can free the original copy of the information */
		if (yyvsp[-3].inpcoord_p) {
			FREE(yyvsp[-3].inpcoord_p);
		}
		if (yyvsp[-1].inpcoord_p) {
			FREE(yyvsp[-1].inpcoord_p);
		}
		insertMAINLL(Currstruct_p, Mainlltc_p);
		Currstruct_p = (struct MAINLL *) 0;
	}
break;
case 492:
#line 4227 "gram.y"
{
		/* null token to cause allocation of LINE struct */
		(void) contextcheck(C_MUSIC, "line");
		Currstruct_p = newMAINLLstruct(S_LINE, yylineno);
	}
break;
case 493:
#line 4234 "gram.y"
{
		yyval.intval = L_NORMAL;
	}
break;
case 495:
#line 4244 "gram.y"
{
		yyval.stringval = 0;
	}
break;
case 496:
#line 4249 "gram.y"
{
		yyval.stringval = fix_string(yyvsp[0].stringval, Titlefont, Titlesize, Curr_filename, yylineno);
	}
break;
case 497:
#line 4255 "gram.y"
{
		if (yyvsp[-6].intval == L_WAVY) {
			l_yyerror(Curr_filename, yylineno,
						"wavy curve not allowed");
		}
		Currstruct_p->u.curve_p->curvetype = yyvsp[-6].intval;
		insertMAINLL(Currstruct_p, Mainlltc_p);
		Currstruct_p = (struct MAINLL *) 0;
	}
break;
case 498:
#line 4266 "gram.y"
{
		(void) contextcheck(C_MUSIC, "curve");
		Currstruct_p = newMAINLLstruct(S_CURVE, yylineno);

		/* get space for a list of locations and mark it as empty */
		Item_count = 0;
		Currstruct_p->u.curve_p->ncoord = (short) Item_count;
		Currstruct_p->u.curve_p->nbulge = (short) Item_count;
		MALLOC(INPCOORD, Currstruct_p->u.curve_p->coordlist, ITEMS);
		Max_items = ITEMS;
	}
break;
case 505:
#line 4300 "gram.y"
{
		/* get enough space to store some values if don't have enough */
		if (Currstruct_p->u.curve_p->nbulge == 0) {
			MALLOCA(float, Currstruct_p->u.curve_p->bulgelist, ITEMS);
		}
		else if ( (Currstruct_p->u.curve_p->nbulge % ITEMS) == 0) {
			REALLOCA(float, Currstruct_p->u.curve_p->bulgelist,
				Currstruct_p->u.curve_p->nbulge + ITEMS);
		}

		/* fill in the value and increment the count of how many */
		Currstruct_p->u.curve_p->bulgelist[Currstruct_p->u.curve_p->nbulge] = yyvsp[-1].intval * yyvsp[0].floatval;
		(Currstruct_p->u.curve_p->nbulge)++;
	}
break;
case 506:
#line 4316 "gram.y"
{
		/* if ran off end of list, make list bigger */
		if (Currstruct_p->u.curve_p->ncoord >= Max_items) {
			Max_items += ITEMS;
			REALLOC(INPCOORD, Currstruct_p->u.curve_p->coordlist,
							Max_items);
		}

		/* put in current place in list and increment count */
		Currstruct_p->u.curve_p->coordlist [ (Currstruct_p->u.curve_p->ncoord)++ ] = *(yyvsp[0].inpcoord_p);
		FREE(yyvsp[0].inpcoord_p);
	}
break;
case 507:
#line 4330 "gram.y"
{
		attach_stuff();
		/* so reset flag */
		Defining_multiple = NO;
	}
break;
case 508:
#line 4338 "gram.y"
{
		chk_stuff_header(yyvsp[-4].intval, yyvsp[-3].intval, yyvsp[-2].intval, Dist_usage);
		add_to_sv_list();
	}
break;
case 509:
#line 4345 "gram.y"
{
		/* Separate rule for phrase because it has the linetype
		 * modifier in front, unlike any other STUFF. The size,
		 * and dist are not really allowed, but we match them
		 * if they are there in order to give more clear error
		 * message than just "syntax error," since there was
		 * already code elsewhere to check for these errors.
		 */
		set_stuff_type(ST_PHRASE);
		chk_stuff_header(yyvsp[-3].intval, yyvsp[-5].intval, yyvsp[-2].intval, Dist_usage);
		add_to_sv_list();
	}
break;
case 510:
#line 4359 "gram.y"
{
		/* need a separate rule for midi, because it can include
		 * a voice as well as staff. It also cannot have the chord,
		 * size or place, although those are checked elsewhere anyway */
		chk_stuff_header(-1, NO, PL_UNKNOWN, SD_NONE);
	}
break;
case 511:
#line 4368 "gram.y"
{
		set_stuff_type(ST_MIDI);
		begin_range(PL_UNKNOWN);
	}
break;
case 513:
#line 4378 "gram.y"
{
		all();
		add_to_sv_list();
	}
break;
case 514:
#line 4385 "gram.y"
{
		/* dummy token for sole purpose of saving away current
		 * stuff type. We have to do this in order to have it in
		 * time to figure out the default place for "all" */
		set_stuff_type(yyvsp[0].intval);
	}
break;
case 515:
#line 4394 "gram.y"
{
		Curr_family = yyvsp[-1].intval;

		switch (yyvsp[0].intval) {
		case FONT_TR:
			yyval.intval = ST_ROM;
			break;
		case FONT_TI:
			yyval.intval = ST_ITAL;
			break;
		case FONT_TB:
			yyval.intval = ST_BOLD;
			break;
		case FONT_TX:
			yyval.intval = ST_BOLDITAL;
			break;
		default:
			pfatal("unknown font");
			break;
		}
	}
break;
case 516:
#line 4418 "gram.y"
{
		if (yyvsp[0].intval != L_NORMAL) {
			/* user used 'dotted' or 'dashed' which applies only
			 * to use of T_L_ANGLE for slurs. Use generic error
			 * message so looks like any other similar error. */
			yyerror("syntax error");
		}
		yyval.intval = ST_CRESC;
	}
break;
case 517:
#line 4430 "gram.y"
{
		yyval.intval = ST_DECRESC;
	}
break;
case 518:
#line 4436 "gram.y"
{
		yyval.intval = ST_PEDAL;
	}
break;
case 519:
#line 4442 "gram.y"
{
		yyval.intval = ST_MUSSYM;
	}
break;
case 520:
#line 4448 "gram.y"
{
		yyval.intval = ST_OCTAVE;
	}
break;
case 521:
#line 4454 "gram.y"
{
		/* use -1 to indicate user didn't specify so must use default */
		yyval.intval = -1;
	}
break;
case 522:
#line 4461 "gram.y"
{
		yyval.intval = yyvsp[-1].intval;
	}
break;
case 523:
#line 4466 "gram.y"
{
		yyval.intval = TM_NONE;
	}
break;
case 524:
#line 4472 "gram.y"
{
		yyval.intval = yyvsp[0].intval;
	}
break;
case 525:
#line 4477 "gram.y"
{
		Dist = 0;
		Dist_usage = SD_NONE;
	}
break;
case 526:
#line 4484 "gram.y"
{
		if (yyvsp[-1].intval != DIST) {
			yyerror("unexpected parameter name instead of 'dist'");
		}
		else {
			if (rangecheck(yyvsp[0].intval, -MAXDIST, MAXDIST, "dist") == YES) {
				Dist = yyvsp[0].intval;
			}
		}
	}
break;
case 527:
#line 4497 "gram.y"
{
		yyval.intval = yyvsp[-1].intval;
	}
break;
case 528:
#line 4503 "gram.y"
{
		if (Dist_usage != SD_FORCE) {
			l_warning(Curr_filename, yylineno,
				"negative dist requires !, setting to 0");
			yyval.intval = 0;
		}
		else {
			yyval.intval = - yyvsp[-1].intval;
		}
	}
break;
case 529:
#line 4515 "gram.y"
{
		Dist_usage = SD_MIN;
	}
break;
case 530:
#line 4521 "gram.y"
{
		Dist_usage = SD_FORCE;
	}
break;
case 533:
#line 4534 "gram.y"
{
		add_stuff_item(yyvsp[-4].floatval, yyvsp[-3].floatval, Gracebackup, yyvsp[-2].stringval, Til_bars, Til_offset,
						Dist, Dist_usage);
	}
break;
case 534:
#line 4541 "gram.y"
{
		Til_bars = 0;
		Til_offset = 0.0;
		yyval.floatval = yyvsp[-1].floatval;
	}
break;
case 535:
#line 4548 "gram.y"
{
		Gracebackup = 0;
	}
break;
case 536:
#line 4554 "gram.y"
{
		Gracebackup = yyvsp[-1].intval;
	}
break;
case 537:
#line 4559 "gram.y"
{
		yyval.floatval = 0.0;
	}
break;
case 538:
#line 4565 "gram.y"
{
		yyval.floatval = yyvsp[-1].floatval * (float) yyvsp[-2].intval;
	}
break;
case 539:
#line 4570 "gram.y"
{
		yyval.stringval = (char *) 0;
	}
break;
case 540:
#line 4576 "gram.y"
{
		if (get_stuff_type() != ST_PEDAL) {
			yyerror("'*' only valid with pedal");
		}
		MALLOCA(char, yyval.stringval, 12);
		yyval.stringval[0] = FONT_MUSIC;
		yyval.stringval[1] = DFLT_SIZE;
		sprintf(yyval.stringval + 2, "\\(endped)");
	}
break;
case 541:
#line 4588 "gram.y"
{
		yyval.stringval = yyvsp[0].stringval;
	}
break;
case 542:
#line 4593 "gram.y"
{
	}
break;
case 543:
#line 4598 "gram.y"
{
	}
break;
case 544:
#line 4603 "gram.y"
{
		/** if there is a non-empty til_suffix, the floatnum better
		 ** have been an integer, and is really the number
		 * of bars. **/
		if (yyvsp[0].intval == YES) {
			Til_bars = (int) yyvsp[-1].floatval;
			if (yyvsp[-1].floatval - Til_bars != 0.0) {
				yyerror("number of measures in til clause must be a whole number");
			}
		}
		else {
			/* only a number, that means it was really a beat offset */
			Til_offset = yyvsp[-1].floatval;
		}
	}
break;
case 545:
#line 4620 "gram.y"
{
		yyval.intval = NO;
	}
break;
case 546:
#line 4626 "gram.y"
{
		yyval.intval = YES;
	}
break;
case 547:
#line 4631 "gram.y"
{
		Til_offset = 0.0;
	}
break;
case 548:
#line 4637 "gram.y"
{
		Til_offset = yyvsp[0].floatval;
	}
break;
case 550:
#line 4645 "gram.y"
{
		newROLLINFO();
	}
break;
case 551:
#line 4651 "gram.y"
{
		newROLLINFO();
		rolldown();
	}
break;
case 553:
#line 4662 "gram.y"
{
		/* save staffnum and voice 1 for start of roll */
		rollparam(yyvsp[0].intval, 1, -1, -1);
	}
break;
case 554:
#line 4669 "gram.y"
{
		/* save staff and voice for start of roll */
		rollparam(yyvsp[-1].intval, yyvsp[0].intval, -1, -1);
	}
break;
case 556:
#line 4678 "gram.y"
{
		/* save staffnum, voice 1 for end of roll */
		rollparam(-1, -1, yyvsp[0].intval, 1);
	}
break;
case 557:
#line 4685 "gram.y"
{
		/* save staffnum and voice for end of roll */
		rollparam(-1, -1, yyvsp[-1].intval, yyvsp[0].intval);
	}
break;
case 558:
#line 4692 "gram.y"
{
		rolloffset( (double) yyvsp[-1].floatval);
	}
break;
case 559:
#line 4698 "gram.y"
{	
		rolloffset( (double) yyvsp[-1].floatval);
	}
break;
case 560:
#line 4703 "gram.y"
{
		if (Curr_family == FAMILY_DFLT) {
			Curr_family = Score.fontfamily;
		}
		Extra = extra_needed(Curr_family + Curr_font, Curr_size, yyvsp[0].stringval);
		proc_printcmd(yyvsp[-2].intval, Curr_loc_info_p, yyvsp[0].stringval, Curr_family + Curr_font,
			Curr_size, yyvsp[-1].intval, NO, (yyvsp[-1].intval == YES ? Extra : (double) 0.0));
	}
break;
case 561:
#line 4714 "gram.y"
{
		if (yyvsp[-1].intval == YES) {
			warning("nl ignored for postscript");
		}
		proc_printcmd(J_NONE, Curr_loc_info_p, yyvsp[0].stringval, FONT_UNKNOWN,
			DFLT_SIZE, NO, YES, (double) 0.0);
		end_raw();
	}
break;
case 562:
#line 4725 "gram.y"
{
		begin_raw();
	}
break;
case 564:
#line 4733 "gram.y"
{
		/* this means print at default location */
		Curr_loc_info_p = (struct INPCOORD *) 0;
		yyval.intval = NO;
	}
break;
case 565:
#line 4741 "gram.y"
{
		Curr_loc_info_p = (struct INPCOORD *) 0;
		yyval.intval = YES;
	}
break;
case 566:
#line 4748 "gram.y"
{
		yyval.intval = NO;
	}
break;
#line 6559 "y.tab.c"
    }
    yyssp -= yym;
    yystate = *yyssp;
    yyvsp -= yym;
    yym = yylhs[yyn];
    if (yystate == 0 && yym == 0)
    {
#if YYDEBUG
        if (yydebug)
            fprintf(stderr, "%sdebug: after reduction, shifting from state 0 to\
 state %d\n", YYPREFIX, YYFINAL);
#endif
        yystate = YYFINAL;
        *++yyssp = YYFINAL;
        *++yyvsp = yyval;
        if (yychar < 0)
        {
            if ((yychar = yylex()) < 0) yychar = 0;
#if YYDEBUG
            if (yydebug)
            {
                yys = 0;
                if (yychar <= YYMAXTOKEN) yys = yyname[yychar];
                if (!yys) yys = "illegal-symbol";
                fprintf(stderr, "%sdebug: state %d, reading %d (%s)\n",
                        YYPREFIX, YYFINAL, yychar, yys);
            }
#endif
        }
        if (yychar == 0) goto yyaccept;
        goto yyloop;
    }
    if ((yyn = yygindex[yym]) && (yyn += yystate) >= 0 &&
            yyn <= YYTABLESIZE && yycheck[yyn] == yystate)
        yystate = yytable[yyn];
    else
        yystate = yydgoto[yym];
#if YYDEBUG
    if (yydebug)
        fprintf(stderr, "%sdebug: after reduction, shifting from state %d \
to state %d\n", YYPREFIX, *yyssp, yystate);
#endif
    if (yyssp >= yyss + yystacksize - 1)
    {
        goto yyoverflow;
    }
    *++yyssp = yystate;
    *++yyvsp = yyval;
    goto yyloop;
yyoverflow:
    yyerror("yacc stack overflow");
yyabort:
    return (1);
yyaccept:
    return (0);
}
